/*
 * resourced
 *
 * Copyright (c) 2000 - 2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/*
 * Cgroup creation implementation
 */

#include "cgroup.h"
#include "const.h"
#include "macro.h"
#include "util.h"
#include "resourced.h"
#include "trace.h"
#include "file-helper.h"

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>		/*mkdirat */
#include <glib.h>
#include <limits.h>
#include <sys/stat.h>		/*mkdirat */
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>		/*time function */
#include <unistd.h>
#include <sys/mount.h>

#define RELEASE_AGENT	"release_agent"
#define NOTIFY_ON_RELEASE  "notify_on_release"

static bool cgroup_is_exists(const char *cgroup_full_path)
{
	struct stat stat_buf;
	return stat(cgroup_full_path, &stat_buf) == 0;
}

static int cgroup_create(const char *cgroup_full_path)
{
	if (mkdir(cgroup_full_path,
		S_IRUSR | S_IWUSR | S_IRGRP) < 0)
		return -errno;

	return 0;
}

/*
 * @desc place pid to cgroup.procs file
 * @return 0 in case of success, errno value in case of failure
 */
resourced_ret_c cgroup_write_pid_fullpath(const char *cgroup_full_path,
	const int pid)
{
	int ret;

	if (pid <= 0) {
		_E("try to write empty pid to %s", cgroup_full_path);
		return RESOURCED_ERROR_NO_DATA;
	}

	ret = cgroup_write_node_uint32(cgroup_full_path, CGROUP_FILE_NAME,
		(u_int32_t)pid);

	ret_value_msg_if(ret < 0, RESOURCED_ERROR_FAIL,
		"Failed place all pid to cgroup %s : %m",
			 cgroup_full_path);
	return RESOURCED_ERROR_NONE;
}

resourced_ret_c cgroup_write_pid(const char *cgroup_subsystem,
	const char *cgroup_name, const int pid)
{
	char buf[MAX_PATH_LENGTH];
	snprintf(buf, sizeof(buf), "%s/%s", cgroup_subsystem, cgroup_name);
	return cgroup_write_pid_fullpath(buf, pid);
}

resourced_ret_c cgroup_write_pidtree(const char *cgroup_subsystem,
	const char *cgroup_name, const int pid)
{
	char buf[MAX_PATH_LENGTH];

	/*/proc/%d/task/%d/children */
	char child_buf[21 + MAX_DEC_SIZE(int) + MAX_DEC_SIZE(int)];
	char pidbuf[MAX_DEC_SIZE(int)];
	resourced_ret_c ret;

	FILE *f;

	snprintf(buf, sizeof(buf), "%s/%s", cgroup_subsystem, cgroup_name);
	/* place parent */
	ret = cgroup_write_pid_fullpath(buf, pid);
	ret_value_msg_if(ret != RESOURCED_ERROR_NONE, ret,
	  "Failed to put parent process %d into %s cgroup", pid, cgroup_name);

	snprintf(child_buf, sizeof(child_buf), PROC_TASK_CHILDREN,
		 pid, pid);
	f = fopen(child_buf, "r");
	ret_value_msg_if(!f, RESOURCED_ERROR_FAIL, "Failed to get child pids!");
	while (fgets(pidbuf, sizeof(pidbuf), f) != NULL) {
		int child_pid = atoi(pidbuf);
		if (child_pid < 0) {
			_E("Invalid child pid!");
			fclose(f);
			return RESOURCED_ERROR_FAIL;
		}
		resourced_ret_c ret = cgroup_write_pid_fullpath(buf, child_pid);
		if (ret != RESOURCED_ERROR_NONE) {
			_E("Failed to put parent process %d into %s cgroup", pid, cgroup_name);
			fclose(f);
			return ret;
		}
	}
	fclose(f);
	return RESOURCED_ERROR_NONE;
}

int cgroup_write_node_uint32(const char *cgroup_name,
		const char *file_name, uint32_t value)
{
	char buf[MAX_PATH_LENGTH];
	snprintf(buf, sizeof(buf), "%s/%s", cgroup_name, file_name);
	_SD("cgroup_buf %s, value %d\n", buf, value);
	return fwrite_uint(buf, value);
}

int cgroup_write_node_int32(const char *cgroup_name,
		const char *file_name, int32_t value)
{
	char buf[MAX_PATH_LENGTH];
	snprintf(buf, sizeof(buf), "%s/%s", cgroup_name, file_name);
	_SD("cgroup_buf %s, value %d\n", buf, value);
	return fwrite_int(buf, value);
}

int cgroup_write_node_str(const char *cgroup_name,
			const char *file_name, const char *string)
{
	char buf[MAX_PATH_LENGTH];
	snprintf(buf, sizeof(buf), "%s/%s", cgroup_name, file_name);
	_SD("cgroup_buf %s, string %s\n", buf, string);
	return fwrite_str(buf, string);
}

int cgroup_read_node_uint32(const char *cgroup_name,
		const char *file_name, uint32_t *value)
{
	char buf[MAX_PATH_LENGTH];
	int ret;
	snprintf(buf, sizeof(buf), "%s/%s", cgroup_name, file_name);
	ret = fread_uint(buf, value);
	_SD("cgroup_buf %s, value %d\n", buf, *value);
	return ret;
}

int cgroup_read_node_int32(const char *cgroup_name,
		const char *file_name, int32_t *value)
{
	char buf[MAX_PATH_LENGTH];
	int ret;
	snprintf(buf, sizeof(buf), "%s/%s", cgroup_name, file_name);
	ret = fread_int(buf, value);
	_SD("cgroup_buf %s, value %d\n", buf, *value);
	return ret;
}

int cgroup_make_subdir(const char* parentdir, const char* cgroup_name, bool *already)
{
	char buf[MAX_PATH_LENGTH];
	bool cgroup_exists;
	bool cgroup_remount = false;
	int ret = 0;

	if (!parentdir || !cgroup_name) {
		_E("You should input parent and cgroup name");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	ret = snprintf(buf, sizeof(buf), "%s/%s", parentdir, cgroup_name);

	ret_value_msg_if(ret > sizeof(buf), RESOURCED_ERROR_FAIL,
		"Not enought buffer size for %s%s", parentdir, cgroup_name);

	cgroup_exists = cgroup_is_exists(buf);
	if (!cgroup_exists) {
		/* FIXME: do something about this mounting, then replace
		 * `is_exists` with checking for -EEXIST in `create`,
		 * as currently this is both duplication AND a race condition */
		if (!strncmp(parentdir, DEFAULT_CGROUP, sizeof(DEFAULT_CGROUP))) {
			ret = mount("tmpfs", DEFAULT_CGROUP, "tmpfs",
					MS_REMOUNT|MS_NOSUID|MS_NOEXEC|MS_NODEV|MS_STRICTATIME, "mode=755");
			if (ret < 0) {
				_E("Fail to RW mount cgroup directory. Can't make %s cgroup", cgroup_name);
				return RESOURCED_ERROR_FAIL;
			}
			cgroup_remount = true;
		}

		ret = cgroup_create(buf);
		ret_value_msg_if(ret < 0, RESOURCED_ERROR_FAIL,
			"Fail to create cgroup %s : err %d", cgroup_name, errno);

		if (cgroup_remount) {
			ret = mount("tmpfs", DEFAULT_CGROUP, "tmpfs",
					MS_REMOUNT|MS_NOSUID|MS_NOEXEC|MS_NODEV|MS_STRICTATIME|MS_RDONLY, "mode=755");
			if (ret < 0)
				_D("Fail to RO mount");
		}
	}

	if (already)
		*already = cgroup_exists;

	return RESOURCED_ERROR_NONE;
}

int cgroup_mount_subsystem(char* source, char* mount_point, char* opts)
{
	return mount(source, mount_point, "cgroup",
		    MS_NODEV | MS_NOSUID | MS_NOEXEC, opts);
}

int cgroup_set_release_agent(const char *cgroup_subsys, const char *release_agent)
{
	_cleanup_free_ char *buf = NULL;
	int r;

	r = asprintf(&buf, "%s/%s", DEFAULT_CGROUP, cgroup_subsys);
	if (r < 0)
		return -ENOMEM;

	r = cgroup_write_node_str(buf, RELEASE_AGENT, release_agent);
	if (r < 0)
		return r;

	return cgroup_write_node_str(buf, NOTIFY_ON_RELEASE, "1");
}

int cgroup_pid_get_path(const char *controller, pid_t pid, char **path)
{
	_cleanup_free_ char *pid_cgroup = NULL;
	_cleanup_fclose_ FILE *f = NULL;
	char buf[LINE_MAX];
	int r;

	assert(controller);
	assert(pid >= 0);
	assert(path);

	if (pid)
		r = asprintf(&pid_cgroup, "/proc/%d/cgroup", pid);
	else
		r = asprintf(&pid_cgroup, "/proc/self/cgroup");
	if (r < 0)
		return -ENOMEM;

	f = fopen(pid_cgroup, "re");
	if (!f)
		return -errno;

	for (;;) {
		_cleanup_free_ char *subsys = NULL, *p = NULL;
		int n, id;

		if (!fgets(buf, sizeof(buf), f)) {
			if (ferror(f))
				return -errno;

			return -ENXIO;
		}

		n = sscanf(buf, "%d:%m[^:]:%ms", &id, &subsys, &p);
		if (n != 3)
			continue;

		if ((streq(subsys, controller)) ||
		    (strneq(subsys, "name=", strlen("name=")) &&
		     streq(subsys + strlen("name="), controller))) {
			*path = strdup(p);

			if (!*path)
				return -ENOMEM;

			break;
		}
	}

	return 0;
}

static void cgroup_memory_stat_init(struct cgroup_memory_stat *mem_stat, long long val)
{
	enum cgroup_memory_stat_id id;

	assert(mem_stat);

	for (id = 0; id < CGROUP_MEMORY_STAT_MAX; id++)
		mem_stat->value[id] = val;
}

int cgroup_get_memory_stat(const char *name, struct cgroup_memory_stat **mem_stat)
{
	_cleanup_fclose_ FILE *f = NULL;
	struct cgroup_memory_stat *st;
	char p[PATH_MAX] = "";
	char buf[LINE_MAX];
	const char *memory_stat = "memory.stat";
	const int memory_stat_len = strlen(memory_stat);

	if (name) {
		int l;
		int name_len = strlen(name);

		if (strneq(name, CGROUP_MEMORY, strlen(CGROUP_MEMORY)))
			l = snprintf(p, PATH_MAX, "%s", name);
		else
			l = snprintf(p, PATH_MAX, "%s%s%s",
				     CGROUP_MEMORY,
				     name[0] != '/' ? "/" : "",
				     name);

		if (name_len >= memory_stat_len &&
			memcmp(name + name_len - memory_stat_len, memory_stat, memory_stat_len))
			snprintf(p + l, PATH_MAX - l, "%s%s",
				 p[l - 1] != '/' ? "/" : "",
				 memory_stat);
	} else
		snprintf(p, PATH_MAX, "%s/%s", CGROUP_MEMORY, memory_stat);

	f = fopen(p, "re");
	if (!f)
		return -errno;

	st = (struct cgroup_memory_stat *)malloc(sizeof(struct cgroup_memory_stat));
	if (!st)
		return -ENOMEM;

	cgroup_memory_stat_init(st, -1);

	for (;;) {
		enum cgroup_memory_stat_id id;
		size_t l;

		if (!fgets(buf, sizeof(buf), f)) {
			if (ferror(f)) {
				free(st);
				return -errno;
			}
			break;
		}

		l = strcspn(buf, " ");
		if (!l)
			break;

		buf[l] = 0;

		id = cgroup_memory_stat_string_to_id(buf);
		if (id < 0 || id >= CGROUP_MEMORY_STAT_MAX)
			continue;

		st->value[id] = atoll(buf + l + 1);
	}

	*mem_stat = st;

	return 0;
}

/*
 * Usage example:
 *	int i;
 *	pid_t pid;
 *	GArray *pids_array = NULL;
 *
 *	cgroup_get_pids(name, &pids_array);
 *
 *	for (i=0; i < pids_array->len; i++)
 *		_D("pid_t: %d", g_array_index(pids_array, pid_t, i));
 *	g_array_free(pids_array, TRUE);
 *
 */
int cgroup_get_pids(const char *name, GArray **pids)
{
	_cleanup_free_ char *cg_procs = NULL;
	_cleanup_fclose_ FILE *f = NULL;
	char buf[MAX_PATH_LENGTH] = "";
	GArray *array;
	pid_t pid;

	assert(name);
	assert(pids);
	assert(!*pids);

	if (asprintf(&cg_procs, "%s/cgroup.procs", name) < 0)
		return -ENOMEM;

	f = fopen(cg_procs, "r");
	if (!f) {
		_E("Failed to open '%s': %m", name);
		return -errno;
	}

	array = g_array_new(false, false, sizeof(pid_t));

	while (fgets(buf, 32, f)) {
		pid = atoi(buf);
		g_array_append_val(array, pid);
	}

	*pids = array;

	return 0;
}
