/*
 * resourced
 *
 * Copyright (c) 2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <assert.h>
#include <string.h>
#include <resourced.h>
#include <sys/types.h>
#include <stdint.h>
#include <glib.h>
#include "macro.h"

/*
 * Cgroup creation interface
 */

#ifndef _CGROUP_LIBRARY_CGROUP_H_
#define _CGROUP_LIBRARY_CGROUP_H_

#define DEFAULT_CGROUP       "/sys/fs/cgroup"
#define CGROUP_MEMORY        DEFAULT_CGROUP "/memory"
#define PROC_TASK_CHILDREN   "/proc/%d/task/%d/children"

/**
 * @desc Get one unsigned int32 value from cgroup
 * @param cgroup_name - cgroup path
 * @param file_name - cgroup content to write
 * @param value - out parameter, value to fill
 * @return negative value if error
*/
int cgroup_read_node_uint32(const char *cgroup_name,
		const char *file_name, uint32_t *value);

/**
 * @desc Get one signed int32 value from cgroup
 * @param cgroup_name - cgroup path
 * @param file_name - cgroup content to write
 * @param value - out parameter, value to fill
 * @return negative value if error
 */
int cgroup_read_node_int32(const char *cgroup_name,
		const char *file_name, int32_t *value);

/**
 * @desc Put unsigned int32 value to cgroup,
 * @param cgroup_name - cgroup path
 * @param file_name - cgroup content to write
 * @param value - unsigned int32 data to write
 * @return negative value if error
 */
int cgroup_write_node_uint32(const char *cgroup_name,  const char *file_name, uint32_t value);

/**
 * @desc Put signed int32 value to cgroup,
 * @param cgroup_name - cgroup path
 * @param file_name - cgroup content to write
 * @param value - signed int32 data to write
 * @return negative value if error
 */
int cgroup_write_node_int32(const char *cgroup_name,  const char *file_name, int32_t value);

/**
 * @desc Put value to cgroup,
 * @param cgroup_name - cgroup path
 * @param file_name - cgroup content to write
 * @param string -string to write
 * @return negative value if error
 */
int cgroup_write_node_str(const char *cgroup_name,
		const char *file_name, const char *string);

/**
 * @desc make cgroup,
 * @param parentdir - parent cgroup path
 * @param cgroup_name - cgroup subdirectory to write
 * @param already - true if subdir already exists, NULL pointer is possible
 * as formal argument, in this case it will not be filled
 * @return negative value if error
 */
int cgroup_make_subdir(const char* parentdir, const char* cgroup_name, bool *already);

/**
 * @desc mount cgroup,
 * @param source -cgroup name
 * @param mount_point - cgroup path
 * @param opts - mount options
 * @return negative value if error
 */
int cgroup_mount_subsystem(char* source, char* mount_point, char* opts);

/**
 * @desc write pid into cgroup_subsystem/cgroup_name file,
 * @param cgroup_subsystem path to /sys/fs/cgroup/subsystem
 * @param cgroup_name - name in /sys/fs/cgroup/subsystem/
 * @return negative value if error
 */
resourced_ret_c cgroup_write_pid(const char *cgroup_subsystem,
	const char *cgroup_name, const int pid);

resourced_ret_c cgroup_write_pid_fullpath(const char *cgroup_full_path,
	const int pid);

/**
 * @desc doing the same as @see cgroup_write_pid,
 * but also put into cgroup first level child processes
 */
resourced_ret_c cgroup_write_pidtree(const char *cgroup_subsystem,
	const char *cgroup_name, const int pid);

/**
 * @desc this function sets release agent path into cgroup subsystem
 * and enables this mechanism
 * @param cgroup_sussys - cgroup subsystem name, it's relative path to cgroup,
 * relativelly default cgroup path (DEFAULT_CGROUP)
 * @param release_agent full path to release agent executable
 * @return negative value if error
 */
int cgroup_set_release_agent(const char *cgroup_subsys, const char *release_agent);

/**
 * @desc get controller path of pid
 * @param controller the name of control group such like
 * "cpuacct,cpu", memory, freezer or debug. Especially supporting
 * "name=" prefix. For example, if "systemd" is given as controller
 * then finding also "name=systemd".
 * @param pid pid to get the controller path
 * @param path contoller path is filled. This value has to be free-ed
 * by caller function.
 * @return negative value if error
 */
int cgroup_pid_get_path(const char *controller, pid_t pid, char **path);

/**
 * @desc get PIDs of processes in a certain cgroup, an allocated array must be provided
 * @return 0 if pids were read and array filled
 */
int cgroup_get_pids(const char *name, GArray **pids);

enum cgroup_memory_stat_id {
	CGROUP_MEMORY_STAT_CACHE = 0,
	CGROUP_MEMORY_STAT_RSS,
	CGROUP_MEMORY_STAT_RSS_HUGE,
	CGROUP_MEMORY_STAT_MAPPED_FILE,
	CGROUP_MEMORY_STAT_DIRTY,
	CGROUP_MEMORY_STAT_WRITEBACK,
	CGROUP_MEMORY_STAT_SWAP,
	CGROUP_MEMORY_STAT_PGPGIN,
	CGROUP_MEMORY_STAT_PGPGOUT,
	CGROUP_MEMORY_STAT_PGFAULT,
	CGROUP_MEMORY_STAT_PGMAJFAULT,
	CGROUP_MEMORY_STAT_INACTIVE_ANON,
	CGROUP_MEMORY_STAT_ACTIVE_ANON,
	CGROUP_MEMORY_STAT_INACTIVE_FILE,
	CGROUP_MEMORY_STAT_ACTIVE_FILE,
	CGROUP_MEMORY_STAT_UNEVICTABLE,
	CGROUP_MEMORY_STAT_HIERARCHICAL_MEMORY_LIMIT,
	CGROUP_MEMORY_STAT_HIERARCHICAL_MEMSW_LIMIT,
	CGROUP_MEMORY_STAT_TOTAL_CACHE,
	CGROUP_MEMORY_STAT_TOTAL_RSS,
	CGROUP_MEMORY_STAT_TOTAL_RSS_HUGE,
	CGROUP_MEMORY_STAT_TOTAL_MAPPED_FILE,
	CGROUP_MEMORY_STAT_TOTAL_DIRTY,
	CGROUP_MEMORY_STAT_TOTAL_WRITEBACK,
	CGROUP_MEMORY_STAT_TOTAL_SWAP,
	CGROUP_MEMORY_STAT_TOTAL_PGPGIN,
	CGROUP_MEMORY_STAT_TOTAL_PGPGOUT,
	CGROUP_MEMORY_STAT_TOTAL_PGFAULT,
	CGROUP_MEMORY_STAT_TOTAL_PGMAJFAULT,
	CGROUP_MEMORY_STAT_TOTAL_INACTIVE_ANON,
	CGROUP_MEMORY_STAT_TOTAL_ACTIVE_ANON,
	CGROUP_MEMORY_STAT_TOTAL_INACTIVE_FILE,
	CGROUP_MEMORY_STAT_TOTAL_ACTIVE_FILE,
	CGROUP_MEMORY_STAT_TOTAL_UNEVICTABLE,
	CGROUP_MEMORY_STAT_RECENT_ROTATED_ANON,
	CGROUP_MEMORY_STAT_RECENT_ROTATED_FILE,
	CGROUP_MEMORY_STAT_RECENT_SCANNED_ANON,
	CGROUP_MEMORY_STAT_RECENT_SCANNED_FILE,
	CGROUP_MEMORY_STAT_MAX,
	CGROUP_MEMORY_STAT_INVALID = -1,
};

struct cgroup_memory_stat {
	long long value[CGROUP_MEMORY_STAT_MAX];
};

const char *cgroup_memory_stat_id_to_string(enum cgroup_memory_stat_id id);
enum cgroup_memory_stat_id cgroup_memory_stat_string_to_id(const char *str);

int cgroup_get_memory_stat(const char *name, struct cgroup_memory_stat **mem_stat);

#endif /*_CGROUP_LIBRARY_CGROUP_H_*/
