/*
 * resourced
 *
 * Copyright (c) 2000 - 2019 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/*
 * @file dbus-handler.c
 *
 * @desc dbus handler using libdbus
 *
 */

#include <glib.h>
#include <gio/gio.h>
#include <stdint.h>

#include "trace.h"
#include "dbus-handler.h"
#include "macro.h"
#include "resourced.h"
#include "util.h"
#include "fd-handler.h"

static guint owner_id;
static GList *dbus_signal_list;
static GList *dbus_method_handle_list;

static GDBusConnection *d_bus_get_connection(void)
{
	static GDBusConnection *g_dbus_conn;

	if (g_dbus_conn)
		return g_dbus_conn;

	g_dbus_conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, NULL);
	g_assert(g_dbus_conn);

	return g_dbus_conn;
}

static struct d_bus_signal *d_bus_find_signal(const char *path, const char *interface, const char *name)
{
	size_t path_len, interface_len, name_len;
	GList *iter;
	struct d_bus_signal *signal;

	path_len = strlen(path) + 1;
	interface_len = strlen(interface) + 1;
	name_len = strlen(name) + 1;

	for (iter = dbus_signal_list; iter; iter = g_list_next(iter)) {
		signal = (struct d_bus_signal *)iter->data;
		if (strncmp(signal->path, path, path_len))
			continue;
		if (strncmp(signal->interface, interface, interface_len))
			continue;
		if (strncmp(signal->name, name, name_len))
			continue;
		return signal;
	}

	return NULL;
}

/* Find appropriate method and call its callback */
static void d_bus_method_handler(GDBusConnection *connection,
		const gchar *sender, const gchar *object_path,
		const gchar *interface_name, const gchar *method_name,
		GVariant *parameters, GDBusMethodInvocation *invocation,
		gpointer user_data)
{
	int i;
	struct d_bus_method_handle *handle;
	int name_len;

	if (!user_data) {
		_E("There is no handle about received gdbus method");
		return;
	}
	handle = (struct d_bus_method_handle *)user_data;

	name_len = strlen(method_name) + 1;
	for (i = 0; i < handle->num_methods; i++) {
		if (strncmp(handle->methods[i].name, method_name, name_len))
			continue;

		handle->methods[i].callback(invocation, parameters);
		break;
	}
}

static const GDBusInterfaceVTable vtable = { d_bus_method_handler, NULL, NULL };

bool dbus_get_pid(const gchar *name, pid_t *outpid)
{
	GError *error = NULL;
	GVariant *const reply = g_dbus_connection_call_sync(d_bus_get_connection(),
			DBUS_BUS_NAME, DBUS_OBJECT_PATH,
			DBUS_INTERFACE_NAME, METHOD_DBUS_GET_PID,
			g_variant_new("(s)", name), NULL, G_DBUS_CALL_FLAGS_NONE,
			-1, NULL, &error);
	if (error || !reply) {
		_E("Failed to call D-Bus name %s, object path %s, interface %s, method %s: %s", DBUS_BUS_NAME, DBUS_OBJECT_PATH, DBUS_INTERFACE_NAME, METHOD_DBUS_GET_PID, error ? error->message : "reply signature mismatch");
		goto error;
	}

	do_expr_unless_g_variant_consume_typechecked(goto error, reply, "(u)", outpid);
	return true;

error:
	if (error)
		g_error_free(error);
	return false;
}

static void d_bus_signal_handler(GDBusConnection *connection,
		const gchar *sender, const gchar *object_path,
		const gchar *interface_name, const gchar *signal_name,
		GVariant *parameters, gpointer user_data)
{
	struct d_bus_signal *signal = (struct d_bus_signal *)user_data;
	if (!signal || !signal->callback)
		return;

	signal->callback(parameters);
}

struct dbus_int {
	int *list;
	int size;
};

static GVariant *append_variant(const char *sig, char *param[])
{
	GVariantBuilder builder;
	GVariantBuilder *sub_builder;
	GVariant *var;
	struct dbus_int *array_int;
	char *ch;
	int i, j;

	if (!sig || !param)
		return NULL;

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);

	for (ch = (char *)sig, i = 0; *ch != '\0'; ++i, ++ch) {
		switch (*ch) {
		case 'i':
			g_variant_builder_add(&builder, "i", atoi(param[i]));
			break;
		case 'u':
			g_variant_builder_add(&builder, "u", strtoul(param[i], NULL, 10));
			break;
		case 't':
			g_variant_builder_add(&builder, "t", atoll(param[i]));
			break;
		case 's':
			g_variant_builder_add(&builder, "s", param[i]);
			break;
		case 'a':
			++ch;
			switch (*ch) {
			case 'i':
				sub_builder = g_variant_builder_new(G_VARIANT_TYPE("ai"));
				array_int = (struct dbus_int *)param[i];
				for (j = 0; j < array_int->size; j++)
					g_variant_builder_add(sub_builder, "i", array_int->list[j]);
				var = g_variant_new("ai", sub_builder);
				g_variant_builder_unref(sub_builder);
				g_variant_builder_add_value(&builder, var);
				break;
			default:
				break;
			}
			break;
		default:
			return NULL;
		}
	}

	return g_variant_builder_end(&builder);
}

void serialize_params(char *params[], size_t n, ...)
{
	va_list va;
	int i = 0;
	va_start(va, n);
	for (i = 0; i < n; ++i)
		params[i] = va_arg(va, char *);
	va_end(va);
}

int d_bus_call_method_sync(const char *dest, const char *path,
		const char *interface, const char *method,
		const char *sig, char *param[])
{
	GVariant *gv = NULL;

	if (sig || param) {
		gv = append_variant(sig, param);
		if (!gv) {
			_E("Failed to build g_variant");
			return -EPERM;
		}
	}

	return d_bus_call_method_sync_gvariant(dest, path, interface, method, gv);
}

int d_bus_call_method_sync_gvariant(const char *dest, const char *path,
		const char *interface, const char *method,
		GVariant *gv)
{
	int ret;
	GVariant *reply = NULL;
	GError *err = NULL;

	reply = g_dbus_connection_call_sync(d_bus_get_connection(), dest, path,
			interface, method, gv, NULL, G_DBUS_CALL_FLAGS_NONE, -1, NULL, &err);
	if (err || !reply) {
		/* The case where !err && !reply (i.e. reply type mismatch) can't currently happen
		 * because we don't specify the desired reply type, but it's good to be future-proof */
		_E("Failed to call D-Bus name %s, objpath %s, iface %s, method %s: %s",
			dest, path, interface, method, err ? err->message : "reply signature mismatch");
		if (err)
			g_error_free(err);
		return -ECOMM;
	}

	do_expr_unless_g_variant_consume_typechecked(return -EINVAL, reply, "(i)", &ret);

	return ret;
}

int d_bus_call_method_async(const char *dest, const char *path,
		const char *interface, const char *method,
		const char *sig, char *param[])
{
	GVariant *gv = append_variant(sig, param);
	if (!gv) {
		_E("Failed to build g_variant");
		return -EPERM;
	}
	return d_bus_call_method_async_gvariant(dest, path, interface, method, gv);
}

int d_bus_call_method_async_gvariant(const char *dest, const char *path,
		const char *interface, const char *method, GVariant *gv)
{
	g_dbus_connection_call(d_bus_get_connection(), dest, path, interface, method,
			gv, NULL, G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);

	return 0;
}

resourced_ret_c d_bus_register_signal(const char *path, const char *interface,
		const char *name, d_bus_signal_callback callback, void *user_data)
{
	struct d_bus_signal *signal;

	if (d_bus_find_signal(path, interface, name)) {
		_E("Same D-Bus signal is already registered");
		return RESOURCED_ERROR_FAIL;
	}

	signal = calloc(1, sizeof(struct d_bus_signal));
	if (!signal) {
		_E("Not enough memory!");
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}

	signal->path = path;
	signal->interface = interface;
	signal->name = name;
	signal->callback = callback;
	signal->user_data = user_data;

	signal->subscription_id = g_dbus_connection_signal_subscribe(d_bus_get_connection(),
			NULL, interface, name, path, NULL, G_DBUS_SIGNAL_FLAGS_NONE,
			d_bus_signal_handler, signal, NULL);
	if (!(signal->subscription_id)) {
		_E("Failed to subscribe signal %s", name);
		free(signal);
		return RESOURCED_ERROR_FAIL;
	}

	dbus_signal_list = g_list_append(dbus_signal_list, signal);

	return RESOURCED_ERROR_NONE;
}

int d_bus_broadcast_signal(const char *path, const char *interface,
		const char *name, const char *sig, char *param[])
{
	gboolean ret;
	GVariant *gv;

	gv = append_variant(sig, param);
	if (!gv) {
		_E("Failed to build g_variant");
		return -EPERM;
	}

	ret = g_dbus_connection_emit_signal(d_bus_get_connection(),
						NULL, path, interface, name, gv, NULL);
	if (ret != TRUE) {
		_E("Failed to emit gdbus signal(%s:%s-%s)", path, interface, name);
		return -ECOMM;
	}

	return RESOURCED_ERROR_NONE;
}

resourced_ret_c d_bus_register_methods(const char *path, const char *xml,
		const struct d_bus_method *methods, const size_t num_methods)
{
	struct d_bus_method_handle *handle = NULL;
	GError *err = NULL;

	if (!path || !xml || !methods) {
		_E("You must input path, introspection xml and method list");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	handle = (struct d_bus_method_handle *)calloc(1, sizeof(struct d_bus_method_handle));
	if (!handle) {
		_E("Not enough memory!");
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}

	handle->path = path;
	handle->methods = methods;
	handle->num_methods = num_methods;

	handle->node_info = g_dbus_node_info_new_for_xml(xml, &err);
	if (!handle->node_info) {
		_E("Failed to make node_info : %s", err->message);
		goto on_error;
	}

	handle->subscription_id = g_dbus_connection_register_object(d_bus_get_connection(),
			path, handle->node_info->interfaces[0], &vtable, (gpointer)handle, NULL, &err);
	if (!handle->subscription_id) {
		_E("Failed to register gdbus methods with path %s : %s", path, err->message);
		g_dbus_node_info_unref(handle->node_info);
		goto on_error;
	} else
		dbus_method_handle_list = g_list_append(dbus_method_handle_list, (gpointer)handle);

	return RESOURCED_ERROR_NONE;

on_error:
	g_clear_error(&err);
	free(handle);

	return RESOURCED_ERROR_FAIL;
}

resourced_ret_c d_bus_register_signals(const struct d_bus_signal *signals, const size_t size)
{
	int i;
	int ret;

	for (i = 0; i < size; i++) {
		if (!signals[i].path || !signals[i].interface || !signals[i].name || !signals[i].callback)
			continue;

		ret = d_bus_register_signal(
				signals[i].path,
				signals[i].interface,
				signals[i].name,
				signals[i].callback,
				signals[i].user_data);
		if (ret != RESOURCED_ERROR_NONE) {
			_E("Fail to add signal %s, %s!\n", signals[i].path, signals[i].name);
			return RESOURCED_ERROR_FAIL;
		}
	}

	return RESOURCED_ERROR_NONE;
}

resourced_ret_c d_bus_reply_message(GDBusMessage *msg)
{
	GError *err = NULL;

	if (!g_dbus_connection_send_message(d_bus_get_connection(), msg,
				G_DBUS_SEND_MESSAGE_FLAGS_NONE, NULL, &err)) {
		_E("Fail to reply gdbus message");
		g_clear_error(&err);
		return RESOURCED_ERROR_FAIL;
	}

	g_clear_error(&err);
	return RESOURCED_ERROR_NONE;
}

void d_bus_init(void)
{
	owner_id = g_bus_own_name(G_BUS_TYPE_SYSTEM, RESOURCED_DBUS_BUS_NAME,
			G_BUS_NAME_OWNER_FLAGS_NONE, NULL, NULL, NULL, NULL, NULL);
	if (owner_id == 0) {
		_E("Failed to get gdbus own name");
	}
}

void d_bus_exit(void)
{
	GList *iter;
	struct d_bus_signal *signal;
	struct d_bus_method_handle *method_handle;
	GDBusConnection *conn = d_bus_get_connection();

	if (owner_id != 0)
		g_bus_unown_name(owner_id);

	/* Remove D-Bus signal list */
	for (iter = dbus_signal_list; iter; iter = g_list_next(iter)) {
		signal = (struct d_bus_signal *)iter->data;
		g_dbus_connection_signal_unsubscribe(conn, signal->subscription_id);
		free(signal);
		dbus_signal_list = g_list_remove(dbus_signal_list, iter);
	}

	/* Remove D-Bus method list */
	for (iter = dbus_method_handle_list; iter; iter = g_list_next(iter)) {
		method_handle = (struct d_bus_method_handle *)iter->data;
		g_dbus_connection_unregister_object(conn, method_handle->subscription_id);
		g_dbus_node_info_unref(method_handle->node_info);
		free(method_handle);
		dbus_method_handle_list = g_list_remove(dbus_method_handle_list, iter);
	}

	g_object_unref(conn);
}
