/*
 * resourced
 *
 * Copyright (c) 2013 - 2019 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/**
 * @file dbus-handler.h
 * @desc  dbus handler using libdbus
 **/

#ifndef __DBUS_HANDLER_H__
#define __DBUS_HANDLER_H__

#include <dbus-names-local.h>
#include <dbus-names-external.h>
#include <gio/gio.h>
#include <resourced.h>

typedef void (*d_bus_method_callback)(GDBusMethodInvocation *invocation, GVariant *params);
typedef void (*d_bus_signal_callback)(GVariant *params);

struct d_bus_method {
	const char *name;
	d_bus_method_callback callback;
};

struct d_bus_method_handle {
	const char *path;
	const struct d_bus_method *methods;
	size_t num_methods;
	GDBusNodeInfo *node_info;
	guint subscription_id;
};

struct d_bus_signal {
	const char *path;
	const char *interface;
	const char *name;
	d_bus_signal_callback callback;
	void *user_data;
	guint subscription_id;
};

/*
 * If reply signature is existed, GDBus allows the return type
 * to declared type or error.
 *
 * However, before GDBus is applied, resourced returns NULL
 * when error is occurred.
 *
 * Therefore, for backward compatibility, we don't use the function
 * g_dbus_method_invocation_return_error and return NULL with sending new message
 *
 * cf) Process resource usage functions use return_error because
 *     it is only for Runtime-info (we can control this package)
 */
#define D_BUS_REPLY_NULL(ivc)	\
{	\
	GDBusMessage *re = g_dbus_message_new_method_reply(g_dbus_method_invocation_get_message(ivc));	\
	d_bus_reply_message(re);	\
	g_object_unref(re);	\
	g_object_unref(ivc);	\
}

/*
 * @desc helper function for filling params array
 * That params array is used in d_bus_method_sync/d_bus_method_async
 * */
void serialize_params(char *params[], size_t n, ...);


int d_bus_call_method_sync_gvariant(const char *dest, const char *path,
		const char *interface, const char *method,
		GVariant *gv);

int d_bus_call_method_sync(const char *dest, const char *path,
		const char *interface, const char *method,
		const char *sig, char *param[]);

int d_bus_call_method_async_gvariant(const char *dest, const char *path,
		const char *interface, const char *method,
		GVariant *gv);

int d_bus_call_method_async(const char *dest, const char *path,
		const char *interface, const char *method,
		const char *sig, char *param[]);

resourced_ret_c d_bus_register_signal(const char *path, const char *interface,
		const char *name, d_bus_signal_callback callback, void *user_data);

int d_bus_broadcast_signal(const char *path, const char *interface,
		const char *name, const char *sig, char *param[]);

resourced_ret_c d_bus_register_methods(const char *path, const char *xml,
		const struct d_bus_method *methods, const size_t num_methods);

resourced_ret_c d_bus_register_signals(const struct d_bus_signal *signals,
									   const size_t size);

resourced_ret_c d_bus_reply_message(GDBusMessage *msg);

int launch_system_app_by_dbus(const char *dest, const char *path,
		const char *iface, const char *method, const char *arg_type, ...);

bool dbus_get_pid(const gchar *name, pid_t *outpid);

void d_bus_init(void);
void d_bus_exit(void);

#endif /* __DBUS_HANDLER_H__ */
