/*
 * resourced
 *
 * Copyright (c) 2000 - 2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */


/*
 * @file file-helper.c
 * @desc Helper functions for working with files
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>

#include "file-helper.h"
#include "trace.h"
#include "macro.h"
#include "util.h"

int fwrite_str(const char *path, const char *str)
{
	_cleanup_fclose_ FILE *f = NULL;
	int ret;

	assert(path);
	assert(str);

	f = fopen(path, "w");
	ret_value_errno_msg_if(!f, -errno,
			       "Fail to open file %s", path);

	ret = fputs(str, f);
	ret_value_errno_msg_if(ret == EOF, errno ? -errno : -EIO,
			       "Fail to write file");

	return RESOURCED_ERROR_NONE;
}

int fwrite_int(const char *path, const int number)
{
	_cleanup_free_ char *digit_buf = NULL;
	int ret;

	ret = asprintf(&digit_buf, "%d", number);
	ret_value_errno_msg_if(ret < 0, -ENOMEM,
			       "sprintf failed\n");

	return fwrite_str(path, digit_buf);
}

int fwrite_uint(const char *path, const u_int32_t number)
{
	_cleanup_free_ char *digit_buf = NULL;
	int ret;

	ret = asprintf(&digit_buf, "%d", number);
	ret_value_errno_msg_if(ret < 0, -ENOMEM,
			       "sprintf failed\n");

	return fwrite_str(path, digit_buf);
}

int fwrite_ulong(const char *path, const unsigned long number)
{
	_cleanup_free_ char *digit_buf = NULL;
	int ret;

	ret = asprintf(&digit_buf, "%lu", number);
	ret_value_errno_msg_if(ret < 0, -ENOMEM,
			       "sprintf failed\n");

	return fwrite_str(path, digit_buf);
}

int fwrite_array(const char *path, const void *array,
		 const size_t size_of_elem,
		 const size_t numb_of_elem)
{
	_cleanup_fclose_ FILE *f = NULL;
	int ret;

	assert(path);
	assert(array);

	f = fopen(path, "w");
	ret_value_errno_msg_if(!f, -errno,
			       "Failed open %s file", path);

	ret = fwrite(array, size_of_elem, numb_of_elem, f);
	ret_value_errno_msg_if(ret != numb_of_elem, -errno,
			       "Failed write array into %s file", path);

	return RESOURCED_ERROR_NONE;
}

int fread_str(const char *path, char **str)
{
	_cleanup_fclose_ FILE *f = NULL;

	f = fopen(path, "r");
	ret_value_errno_msg_if(!f, -errno,
			"Fail to open %s file.", path);

	*str = fgets(*str, LINE_MAX, f);
	ret_value_errno_msg_if(!(*str), -ENODATA,
			"Fail to read file\n");

	return RESOURCED_ERROR_NONE;
}

int fread_nth_int(const char *path, size_t n, int32_t *number)
{
	_cleanup_fclose_ FILE *f = NULL;
	size_t i;
	int32_t t;
	int ret;

	f = fopen(path, "r");
	ret_value_errno_msg_if(!f, -errno,
			       "Fail to open  %s file.", path);

	for (i = 0; i <= n; i++) {
		ret = fscanf(f, "%d", &t);
		ret_value_errno_msg_if(ret == EOF, -errno,
				       "Fail to read file\n");
	}

	*number = t;
	return RESOURCED_ERROR_NONE;
}

int fread_nth_uint(const char *path, size_t n, u_int32_t *number)
{
	_cleanup_fclose_ FILE *f = NULL;
	size_t i;
	u_int32_t t;
	int ret;

	f = fopen(path, "r");
	ret_value_errno_msg_if(!f, -errno,
			       "Fail to open %s file.", path);

	for (i = 0; i <= n; i++) {
		ret = fscanf(f, "%u", &t);
		ret_value_errno_msg_if(ret == EOF, -errno,
				       "Fail to read file\n");
	}

	*number = t;
	return RESOURCED_ERROR_NONE;
}

int fread_nth_ulong(const char *path, size_t n, unsigned long *number)
{
	_cleanup_fclose_ FILE *f = NULL;
	size_t i;
	unsigned long t;
	int ret;

	f = fopen(path, "r");
	ret_value_errno_msg_if(!f, -errno,
			       "Fail to open %s file.", path);

	for (i = 0; i <= n; i++) {
		ret = fscanf(f, "%lu", &t);
		ret_value_errno_msg_if(ret == EOF, -errno,
				       "Fail to read file\n");
	}

	*number = t;
	return RESOURCED_ERROR_NONE;
}
