/*
 * resourced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/*
 *  @file: macro.h
 *
 *  @desc general macros
 */


#ifndef _RESOURCED_MACRO_H_
#define _RESOURCED_MACRO_H_

#include <assert.h>
#include <stdio.h>
#include <config.h>
#include <glib.h>
#include <stdlib.h>
#include <string.h>

#ifndef API
#define API __attribute__((visibility("default")))
#endif

#define UNUSED __attribute__((__unused__))

#define ARRAY_SIZE(x) (sizeof(x)/sizeof(x[0]))

/*
 * One byte digit has 3 position in decimal representation
 * 2 - 5
 * 4 - 10
 * 8 - 20
 * >8 - compile time error
 * plus 1 null termination byte
 * plus 1 for negative prefix
 */
#define MAX_DEC_SIZE(type)				\
	(2 + (sizeof(type) <= 1 ? 3 :			\
	      sizeof(type) <= 2 ? 5 :			\
	      sizeof(type) <= 4 ? 10 :			\
	      sizeof(type) <= 8 ? 20 :			\
	      sizeof(int[-2*(sizeof(type) > 8)])))

#define CHECK_BIT(a, bit)			\
	((a) & (bit))

#define ret_msg_if(expr, fmt, arg...) do {	\
	if (expr) {				\
		_E(fmt, ##arg);			\
		return;				\
	}					\
} while (0)

#define ret_unless(val) \
	do { \
		if (!(val)) { \
			_E("Invalid parameter value (`" #val "` check failed)"); \
			return; \
		} \
	} while (0)

#define ret_value_if(expr, val) do {					\
	if (expr) {							\
		_E("(%s) -> %s():%d return", #expr, __FUNCTION__, __LINE__); \
		return (val);						\
	}								\
} while (0)

#define ret_value_msg_if(expr, val, fmt, arg...) do {	\
	if (expr) {					\
		_E(fmt, ##arg);				\
		return val;				\
	}						\
} while (0)

#define ret_value_errno_msg_if(expr, val, fmt, arg...) do {	\
	if (expr) {						\
		ETRACE_ERRNO_MSG(fmt, ##arg);			\
		return val;					\
	}							\
} while (0)

#define do_expr_if_gvariant_type_mismatch(expr, gvariant, type) \
	do { \
		char _filtered_type [strlen(type) + 1]; \
		int _index_copy = 0; \
		for (int _index_original = 0; _index_original < sizeof _filtered_type - 1; ++_index_original) \
			if (type[_index_original] != '&') \
				_filtered_type[_index_copy++] = type[_index_original]; \
		_filtered_type[_index_copy] = '\0'; \
		if (!g_variant_is_of_type(gvariant, G_VARIANT_TYPE(_filtered_type))) { \
			_E("Invalid parameter type (expected: %s, got: %s)", \
			   _filtered_type, g_variant_get_type_string(gvariant)); \
			expr; \
		} \
	} while (0)

#define ret_if_gvariant_type_mismatch(gvariant, type, args...) \
	do_expr_if_gvariant_type_mismatch(return ##args, gvariant, type)

#define do_expr_unless_g_variant_get_typechecked(expr, gvariant, type, args...) \
	do { \
		do_expr_if_gvariant_type_mismatch(expr, gvariant, type); \
		g_variant_get(gvariant, type, ##args); \
	} while (0)

#define do_expr_unless_g_variant_consume_typechecked(expr, gvariant, type, args...) \
	do { \
		do_expr_unless_g_variant_get_typechecked(g_variant_unref(gvariant) ; expr, gvariant, type, ##args); \
		g_variant_unref(gvariant); \
	} while (0)

#define gslist_for_each_item(item, list)				\
	for (item = list; item != NULL; item = g_slist_next(item))

#define gslist_for_each(head, elem, node)				\
	for (elem = head, node = NULL; elem && ((node = elem->data) != NULL); elem = elem->next, node = NULL)

#define gslist_for_each_safe(head, elem, elem_next, node)		\
	for (elem = head, elem_next = g_slist_next(elem), node = NULL;	\
	     elem && ((node = elem->data) != NULL);			\
	     elem = elem_next, elem_next = g_slist_next(elem), node = NULL)

#define MODULE_REGISTER(module)						\
	static void __attribute__ ((constructor)) module_init(void)	\
	{								\
		add_module(module);					\
	}								\
	static void __attribute__ ((destructor)) module_exit(void)	\
	{								\
		remove_module(module);					\
	}

/* Overlap definition to consider three strings as a string */
#define STRING_FORMAT_SPECIFIER_WITH_MACRO(macro) "%"#macro"s"
#define STR_FS(macro) STRING_FORMAT_SPECIFIER_WITH_MACRO(macro)

#endif	/* _RESOURCED_MACRO_H_ */
