/*
 * resourced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/**
 * @file userinfo-list.c
 * @desc define helper functions to get user info.
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gum/gum-user.h>
#include <gum/gum-user-service.h>

#include "trace.h"
#include "userinfo-list.h"

static GArray *user_list = NULL;

static void userinfo_free_list(void)
{
	int i;
	struct user_info *elem;

	if (user_list) {
		for (i = 0; i < user_list->len; i++) {
			elem = g_array_index(user_list, struct user_info*, i);
			free(elem);
		}
		g_array_free(user_list, TRUE);
		user_list = NULL;
	}
}

static resourced_ret_c userinfo_update_list(void)
{
	GumUserService *gus = NULL;
	GumUserList *users = NULL;
	GumUserList *iter = NULL;
	GumUser *user = NULL;
	gchar **query;

	struct user_info *elem;
	uid_t uid;
	gchar *home_dir = NULL;

	/* Get user data by using libgum */
	gus = gum_user_service_create_sync(TRUE);
	if (!gus) {
		_E("Failed to create gum user service");
		return RESOURCED_ERROR_FAIL;
	}

	query = g_strsplit("admin,normal", ",", -1);

	users = gum_user_service_get_user_list_sync(gus, (const gchar *const *)query);
	g_strfreev(query);

	if (!users) {
		_E("Failed to get gum user list");
		g_object_unref(gus);
		return RESOURCED_ERROR_FAIL;
	}

	/* Remove old user list */
	userinfo_free_list();

	/* Make new user list */
	user_list = g_array_new(FALSE, FALSE, sizeof(struct user_info*));
	assert(user_list);

	iter = users;
	for (; iter != NULL; iter = g_list_next(iter)) {
		user = (GumUser*) iter->data;
		g_object_get(G_OBJECT(user), "uid", &uid, NULL);
		g_object_get(G_OBJECT(user), "homedir", &home_dir, NULL);

		elem = malloc(sizeof(struct user_info));
		assert(elem);

		elem->uid = uid;
		snprintf(elem->home_dir, PATH_MAX, "%s", home_dir);

		g_array_append_val(user_list, elem);

		if (home_dir)
			free(home_dir);
	}

	gum_user_service_list_free(users);
	g_object_unref(gus);

	return RESOURCED_ERROR_NONE;
}

const char *userinfo_find_home_dir(uid_t uid)
{
	int i;
	struct user_info *elem;

	if (userinfo_update_list() != RESOURCED_ERROR_NONE)
		return NULL;

	for (i = 0; i < user_list->len; i++) {
		elem = g_array_index(user_list, struct user_info*, i);

		if (elem->uid == uid)
			return elem->home_dir;
	}
	return NULL;
}

const GArray *userinfo_get_list(void)
{
	if (!user_list)
		if (userinfo_update_list() != RESOURCED_ERROR_NONE)
			return NULL;

	return user_list;
}

static void __attribute__ ((destructor)) userinfo_exit(void)
{
	userinfo_free_list();
}
