/*
 * resourced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * @file heart-battery.c
 *
 * @desc heart battery module
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 */

#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <glib.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <math.h>

#include "proc-common.h"
#include "notifier.h"
#include "resourced.h"
#include "dbus-handler.h"
#include "heart.h"
#include "logging.h"
#include "heart-common.h"
#include "config-parser.h"
#include "trace.h"
#include "vconf.h"
#include "module.h"
#include "macro.h"
#include "util.h"
#include "file-helper.h"

#define TIZEN_SYSTEM_APPID			"org.tizen.system"
#define TIZEN_SYSTEM_BATTERY_APPID		"org.tizen.system.battery.capacity"
#define BATTERY_NAME				"battery"
#define BATTERY_DATA_MAX			1024
#define BATTERY_CAPACITY_MAX			512
#define BATTERY_LINE_MAX			128
#define BATTERY_CLEAN_MAX			100
#define BATTERY_HISTORY_DAY_MAX			7
#define BATTERY_HISTORY_RESET_MAX		5
#define BATTERY_HISTORY_RESET_CURRENT		(BATTERY_HISTORY_RESET_MAX - 1)
#define BATTERY_HISTORY_COUNT_MAX		1000
#define HEART_BATTERY_UPDATE_INTERVAL		HALF_HOUR
#define HEART_BATTERY_SAVE_INTERVAL		HALF_HOUR
#define HEART_BATTERY_CAPACITY_DATA_FILE	HEART_FILE_PATH"/.battery_capacity.dat"
#define HEART_BATTERY_CONF_SECTION              "BATTERY_POWER_MODE"
#define GET_CHARGER_STATUS			"ChargerStatus"
#define GET_BATTERY_CAPACITY			"GetPercent"
#define CALCULATE_DAY_BASE_TIME(x)		((x / DAY_TO_SEC(1)) * (DAY_TO_SEC(1)))
#define REMAIN_CAPACITY(x)			(100 - x)
#define BATTERY_PREDICTION_DATA_MIN		5
#define BATTERY_USAGE_LEARNING			-1
#define CUMUL_WEIGHT				(0.8)
#define TREND_WEIGHT				(1 - CUMUL_WEIGHT)
#define DISCHARGE_BASE_RATE			(0.2)

/*
 * BATTERY_PREDICTION_LATEST_COUNT must be >= BATTERY_PREDICTION_DATA_MIN
 */
#define BATTERY_PREDICTION_LATEST_COUNT		5
/*
 * BATTERY_PREDICTION_PERIOD possible values:
 * DATA_LATEST, DATA_3HOUR, DATA_6HOUR, DATA_12HOUR, DATA_1DAY
 */
#define BATTERY_PREDICTION_PERIOD		DATA_3HOUR

#define BATTERY_USED_TIME                       "BATTERY_USED_TIME"
#define BATTERY_STATUS                          "BATTERY_STATUS"
#define BATTERY_RESET_USAGE                     "BATTERY_RESET_USAGE"
#define BATTERY_WEEK_DAY_USAGE                  "BATTERY_WEEK_DAY_USAGE"
#define BATTERY_LEVEL_USAGE                     "BATTERY_LEVEL_USAGE"
#define BATTERY_PREDICTION                      "BATTERY_PREDICTION"

#define INDEX_WINDOW_SIZE		10  /*data storing index*/
#define BATTERY_LEVEL_GAP		5  /*previous 5 battery data*/
#define LONG_TIME_WEIGHT		(0.7) /*weightage given for longer time discharge*/
#define SHORT_TIME_WEIGHT		(1-LONG_TIME_WEIGHT) /*weightage given for short time discharge*/
#define MIN_TIME_FOR_LVL_CHANGE		50
#define BATT_CHG_FLUSH_DATA		10

#define UPS_FACTOR				(2.88)
#define DOUBLE_ZERO				(0.000000)
#define FULL_CAPACITY			100
/* BATTERY_C_RATE is defined as the rate of which battery capacity is changing
 * If total battery capacity is XmAh, then in one second battery capacity change
 * will be considered as 90
 */
#define BATTERY_C_RATE			90
#define SEC_TO_MIN(x) ((x)/60)
#define CAL_MIN(x, y) ((x < y) ? x : y)
#define BATTERY_WINDOW_INDEX(x)		(((x) +  INDEX_WINDOW_SIZE) % INDEX_WINDOW_SIZE)
#define HEART_BATTERY_DATA_FILE		HEART_FILE_PATH"/.battery_data.dat"
#define VCONFKEY_HEART_BATTERY_DEVICE_MODE			"db/setting/psmode"

enum {
	TA     = 0,	/* prediction based on total data average */
	PCB    = 1,	/* prediction with physiological behaviors */
	WEEK   = 2,	/* prediction based on weekly data */
	COUNT  = 3,	/* prediction based on last BATTERY_PREDICTION_COUNT number of items */
	PERIOD = 4,	/* prediction based on data from last BATTERY_PREDICTION_PERIOD time */
	MAX_STRATEGY = 5,
};

enum {
	POWER_NORMAL_MODE  = 0,
	POWER_SAVING_MODE  = 1,
	ULTRA_SAVING_MODE  = 2,
	POWER_MODE_MAX     = 3,
};

enum {
	BATTERY_LEVEL_LOW = 0, /* 15 ~ 0 */
	BATTERY_LEVEL_MID = 1, /* 49 ~ 16 */
	BATTERY_LEVEL_HIGH = 2, /* 50 ~ 100 */
	BATTERY_LEVEL_MAX = 3,
};

enum {
	DEFAULT_MIN = 0,
	DEFAULT_AVG = 1,
	DEFAULT_MAX = 2,
	DEFAULT_VALUE_MAX = 3,
};

struct battery_used {
	time_t used_time_sec; /* seconds on battery */
	time_t last_update_time;
	int last_charger_status;
};

struct battery_usage {
	time_t start_time; /* timestamp when event started */
	long sec_per_cap[MAX_CHARGER_STATE]; /* seconds per capacity level change */
	long cap_counter[MAX_CHARGER_STATE]; /* number of capacity level changes */
};

struct battery_prediction {
	long sec_per_cap[MAX_STRATEGY]; /* seconds per capacity level change */
	long cap_counter[MAX_STRATEGY]; /* number of capacity level changes */
	long time_pred_min[MAX_STRATEGY]; /* time prediction in minutes */
};

struct battery_status {
	/* current battery status */
	enum charger_status_type curr_charger_status;
	int curr_capacity;

	enum discharge_rate_level_type discharge_rate_level;
	/* current runtime statistics */
	long curr_run_time_sec[MAX_CHARGER_STATE]; /* seconds since reset */
	long curr_cap_counter[MAX_CHARGER_STATE]; /* capacity level changes */

	long last_wall_time[INDEX_WINDOW_SIZE];  /* previous battery level discharge time */
	/* previous battery level change time in charging mode */
	long last_wall_time_chg;
	/* charging data index*/
	int index_chg;
	/* store last 5 battery level change time diff */
	int last_wall_time_chg_diff[BATTERY_LEVEL_GAP];
	 /* indicates whether do we have enough data to estimate the battery time */
	int data_available;
	int curr_index;  /* store current index value */
	int last_capacity;  /* last battery capacity value */
	int last_capacity_chg; /* last battery capacity value in charge mode */
	int remaining_time; /* previous remaining time */
	int remaining_time_chg; /* previous remaining time in charging mode */
	int remaining_time_ups; /* previous remaining time in UPS */
	double last_volt_intg[INDEX_WINDOW_SIZE]; /* store previous calculated voltage integral(energy) value */
	double last_pwr_bchg;  /* store the battery power available just before charging start */

	/* wall clock time stamp when last event happened in seconds */
	time_t last_event_wall_time;
	time_t last_clock_tick;

	/*
	 * reset mark is set when battery is charged in over 90% and
	 * charger was disconnected from the device.
	 * We consider then the device as "charged"
	 *
	 * The possible values are 0 and 1 they're swapped to opposite on change.
	 */
	int reset_mark;
	time_t reset_mark_timestamp;

	/* usage time from last reset_mark change*/
	struct battery_usage batt_reset_usage[BATTERY_HISTORY_RESET_MAX];

	/* usage time by week day */
	struct battery_usage week_day_usage[BATTERY_HISTORY_DAY_MAX];

	/* usage time by user behavior & battery level */
	struct battery_usage batt_lvl_usage[BATTERY_LEVEL_MAX];

	/* calculated battery prediction */
	struct battery_prediction prediction[MAX_CHARGER_STATE];
};

static int default_sec_per_cap[MAX_CHARGER_STATE][DEFAULT_VALUE_MAX] = {
	{ 70, 670, 3600 }, /* DISCHARGING MIN: 70s, AVG: 670s, MAX: 1 hour */
	{ 30, 80, 3600 }    /* CHARGING MIN: 30s, AVG: 80s,  MAX: 1 hour */
};

static double default_mode_spc[POWER_MODE_MAX] = {
	670,	/* POWER_NORMAL_MODE */
	750,	/* POWER_SAVING_MODE */
	1947	/* ULTRA_SAVING_MODE */
};

static double default_mode_factor[POWER_MODE_MAX] = {
	1,	/* POWER_NORMAL_MODE */
	1.1,	/* POWER_SAVING_MODE */
	2.88	/* ULTRA_SAVING_MODE */
};

static struct battery_used batt_used;
static struct battery_status batt_stat;
static GSList *capacity_history_list = NULL;
static pthread_mutex_t heart_battery_mutex = PTHREAD_MUTEX_INITIALIZER;
static time_t last_file_commit_time;
static int battery_learning_mode;

static int ocv_degree;	/* Degree of the OCV and SOC polynomial */
static double *intg;	/* Co-efficients of the OCV and SOC polynomial */
static double pivot_nor;  /* time pivot for normal mode */
static double pivot_ups; /* time pivot for UPS mode */
static double volt_intg_full; /* 100% battery voltage integral(engery) value */
static bool data_avail_chg = false; /* 5 readings in charging mode */
 /*
 * variable to hold the current device mode(NORMAL, PS & UPS)
 */
static int device_mode = POWER_NORMAL_MODE;
/*
 * Given the 'SOC', remaining energy of the cell can be predicted. New battery
 * remaining time estimation logic will calculating the remaining energy as
 * 'curr_volt_intg'.
 *
 * Based on estimated average power(P), a time estimate can be calcluated.
 * Power can be estimated based on constant SOC interval Ps and constant
 * time interval Pt.
 *
 * At high usage Ps will be short term average power and Pt will be long term
 * average power.
 *
 * At low usage Ps will be long term average power and Pt will be short term
 * average power.
 *
 * The new logic will calculate the average power 'batt_pwr' from Ps and Pt.
 *
 * Under-relaxation is a popular and effective technique of stabilizing prediction.
 * The predicted time is stabilized around a pivot, which is calculated from the
 * high-usage time and low-usage time
 *
 * The new logic will calculated 'pivot_nor' and 'pivot_ups' for normal and UPS
 * mode respectively
 *
 * Finally, using current available energy(curr_volt_intg), average power(batt_pwr)
 * and pivot time (pivot_nor or pivot_ups) this new logic will estimate the Battery
 * Remaining Time for normal mode and UPS mode
 */
/*
 * new battery remaining time estimation logic
 */
static int logic_v2;
/*
 * Total time taken in minutes when device is discharging at fastest rate
 */
static int discharge_fast;
/*
 * Total time taken in minutes when device is discharging at slowest rate
 */
static int discharge_slow;
/*
 * Avarage power consumed in UPS mode
 */
static double average_pwr;
/*
 * Total battery capacity in minutes(mAM)
 */
static int total_battery_capacity;
/*
 * This variable indicates first battery level update after charger insert
 */
static bool first_level_change = FALSE;
/*
 * This variable indicates the battery header used in heart.conf file
 */
static char battery_header[BATTERY_LINE_MAX] = "BATTERY";

int heart_battery_get_capacity(void)
{
	return batt_stat.curr_capacity;
}

enum charger_status_type heart_battery_get_charger_status(void)
{
	return batt_stat.curr_charger_status;
}

enum discharge_rate_level_type heart_battery_get_discharge_rate_level(void)
{
	return batt_stat.discharge_rate_level;
}

static inline void heart_battery_set_usage_reset_stime(int history, time_t start_time)
{
	batt_stat.batt_reset_usage[history].start_time = start_time;
}

static inline time_t heart_battery_get_usage_reset_stime(int history)
{
	return batt_stat.batt_reset_usage[history].start_time;
}

static inline void heart_battery_set_usage_reset(int history, enum charger_status_type status, long sec_per_cap, long cap_counter)
{
	batt_stat.batt_reset_usage[history].sec_per_cap[status] = sec_per_cap;
	batt_stat.batt_reset_usage[history].cap_counter[status] = cap_counter;
}

static inline long heart_battery_get_usage_reset_total_time(int history, enum charger_status_type status)
{
	return batt_stat.batt_reset_usage[history].sec_per_cap[status] * batt_stat.batt_reset_usage[history].cap_counter[status];
}

static inline long heart_battery_get_usage_reset_count(int history, enum charger_status_type status)
{
	return batt_stat.batt_reset_usage[history].cap_counter[status];
}

static inline void heart_battery_set_usage_level_stime(int level, time_t start_time)
{
	batt_stat.batt_lvl_usage[level].start_time = start_time;
}

static inline time_t heart_battery_get_usage_level_stime(int level)
{
	return batt_stat.batt_lvl_usage[level].start_time;
}

static inline void heart_battery_set_usage_level(int level, enum charger_status_type status, long sec_per_cap, long cap_counter)
{
	batt_stat.batt_lvl_usage[level].sec_per_cap[status] = sec_per_cap;
	batt_stat.batt_lvl_usage[level].cap_counter[status] = cap_counter;
}

static inline long heart_battery_get_usage_level_total_time(int level, enum charger_status_type status)
{
	return batt_stat.batt_lvl_usage[level].sec_per_cap[status] * batt_stat.batt_lvl_usage[level].cap_counter[status];
}

static inline long heart_battery_get_usage_level_spc(int level, enum charger_status_type status)
{
	return batt_stat.batt_lvl_usage[level].sec_per_cap[status];
}

static inline long heart_battery_get_usage_level_count(int level, enum charger_status_type status)
{
	return batt_stat.batt_lvl_usage[level].cap_counter[status];
}

static inline long heart_battery_get_usage_week_total_time(int day, enum charger_status_type status)
{
	return batt_stat.week_day_usage[day].sec_per_cap[status] * batt_stat.week_day_usage[day].cap_counter[status];
}

static inline long heart_battery_get_usage_week_count(int day, enum charger_status_type status)
{
	return batt_stat.week_day_usage[day].cap_counter[status];
}

static inline void heart_battery_set_usage_week_stime(int day, time_t start_time)
{
	batt_stat.week_day_usage[day].start_time = start_time;
}

static inline time_t heart_battery_get_usage_week_stime(int day)
{
	return batt_stat.week_day_usage[day].start_time;
}

static inline int heart_battery_get_learning_mode(void)
{
	int i, count = 0;

	if (logic_v2) { /* new logic */
		/*wait till enough data gets collected to estimated battery remaining time */
		if (batt_stat.data_available == TRUE)
			return 1;
		return 0;
	}
	 /* old logic */
	for (i = 0; i < BATTERY_HISTORY_DAY_MAX; i++) {
		if (heart_battery_get_usage_week_stime(i))
			count++;
		if (count > 1)
			return 1;
	}
	return 0;
}

static inline void heart_battery_set_usage_week(int day, enum charger_status_type status, long sec_per_cap, long cap_counter)
{
	batt_stat.week_day_usage[day].sec_per_cap[status] = sec_per_cap;
	batt_stat.week_day_usage[day].cap_counter[status] = cap_counter;
}

static inline void heart_battery_set_prediction(int strategy, enum charger_status_type status, long sec_per_cap, long cap_counter, long pred_min)
{
	batt_stat.prediction[status].sec_per_cap[strategy] = sec_per_cap;
	batt_stat.prediction[status].cap_counter[strategy] = cap_counter;
	batt_stat.prediction[status].time_pred_min[strategy] = pred_min;
}

static inline long heart_battery_get_prediction_time(int strategy, enum charger_status_type status)
{
	return batt_stat.prediction[status].time_pred_min[strategy];
}

static inline time_t heart_battery_get_file_commit_timestamp()
{
	return last_file_commit_time;
}

static inline void heart_battery_set_file_commit_timestamp(time_t timestamp)
{
	last_file_commit_time = timestamp;
}

static int heart_battery_calculate_discharge_rate_level(long spc)
{
	long gap, total_spc;

	total_spc = batt_stat.prediction[DISCHARGING].sec_per_cap[TA];
	if (!total_spc)
		return BATTERY_DISCHARGE_NONE;
	gap = (total_spc * DISCHARGE_BASE_RATE);

	_I("total: %ld, spc: %ld, gap: %ld", total_spc, spc, gap);

	if (total_spc + gap < spc)
		return BATTERY_DISCHARGE_LOW;
	else if (total_spc - gap > spc)
		return BATTERY_DISCHARGE_HIGH;

	return BATTERY_DISCHARGE_AVG;
}

static void heart_battery_save_used_time(char *key, struct battery_used *used)
{
	assert(key);
	assert(used);

	logging_leveldb_putv(key, strlen(key), "%d %d ",
			used->used_time_sec, used->last_update_time);
};

static void heart_battery_save_status(char *key, struct battery_status *status)
{
	assert(key);
	assert(status);

	logging_leveldb_putv(key, strlen(key), "%d %ld %ld %ld %ld %d %d ",
			status->curr_capacity,
			status->curr_run_time_sec[DISCHARGING],
			status->curr_cap_counter[DISCHARGING],
			status->curr_run_time_sec[CHARGING],
			status->curr_cap_counter[CHARGING],
			status->curr_charger_status,
			status->reset_mark);
};

static void heart_battery_save_usage(char *key, struct battery_usage *usage, int total_size)
{
	int i, len, num;
	char buf[BATTERY_DATA_MAX] = {0, };

	assert(key);
	assert(usage);

	len = 0;
	num = total_size/sizeof(struct battery_usage);
	for (i = 0; i < num; i++) {
		len += snprintf(buf + len, BATTERY_DATA_MAX - len, "%ld %ld %ld %ld %ld ",
				usage[i].start_time,
				usage[i].sec_per_cap[DISCHARGING],
				usage[i].cap_counter[DISCHARGING],
				usage[i].sec_per_cap[CHARGING],
				usage[i].cap_counter[CHARGING]);
	}
	logging_leveldb_put(key, strlen(key), buf, len);
};

static void heart_battery_save_prediction(char *key, struct battery_prediction *prediction)
{
	int i, len;
	char buf[BATTERY_DATA_MAX] = {0, };

	assert(key);
	assert(prediction);

	len = 0;
	for (i = 0; i < MAX_STRATEGY; i++) {
		len += snprintf(buf + len, BATTERY_DATA_MAX - len, "%ld %ld %ld %ld %ld %ld ",
				prediction[DISCHARGING].sec_per_cap[i],
				prediction[DISCHARGING].cap_counter[i],
				prediction[DISCHARGING].time_pred_min[i],
				prediction[CHARGING].sec_per_cap[i],
				prediction[CHARGING].cap_counter[i],
				prediction[CHARGING].time_pred_min[i]);
	}
	logging_leveldb_put(key, strlen(key), buf, len);
};


static int heart_battery_load_used_time(char *key, struct battery_used *used)
{
	int ret;
	char *token;
	char buf[BATTERY_DATA_MAX] = {0, };
	char *saveptr;

	if (!key || !used)
		return RESOURCED_ERROR_FAIL;

	ret = logging_leveldb_read(key, strlen(key), buf, sizeof(buf));
	if (ret != RESOURCED_ERROR_NONE) {
		_E("Failed to read leveldb key: %s", key);
		return RESOURCED_ERROR_FAIL;
	}
	token = strtok_r(buf, " ", &saveptr);
	if (!token) {
		_E("Failed to token value");
		return RESOURCED_ERROR_FAIL;
	}
	used->used_time_sec = atoi(token);
	token = strtok_r(NULL, " ", &saveptr);
	if (!token) {
		_E("Failed to token value");
		return RESOURCED_ERROR_FAIL;
	}
	used->last_update_time = atoi(token);
	return RESOURCED_ERROR_NONE;
};

static int heart_battery_load_status(char *key, struct battery_status *status)
{
	int ret;
	char *token;
	char buf[BATTERY_DATA_MAX] = {0, };
	char *saveptr;

	if (!key || !status)
		return RESOURCED_ERROR_FAIL;

	ret = logging_leveldb_read(key, strlen(key), buf, sizeof(buf));
	if (ret != RESOURCED_ERROR_NONE) {
		_E("Failed to read leveldb key: %s", key);
		return RESOURCED_ERROR_FAIL;
	}
	token = strtok_r(buf, " ", &saveptr);
	if (!token) {
		_E("Failed to token value");
		return RESOURCED_ERROR_FAIL;
	}
	status->curr_capacity = atoi(token);
	token = strtok_r(NULL, " ", &saveptr);
	if (!token) {
		_E("Failed to token value");
		return RESOURCED_ERROR_FAIL;
	}
	status->curr_run_time_sec[DISCHARGING] = atol(token);
	token = strtok_r(NULL, " ", &saveptr);
	if (!token) {
		_E("Failed to token value");
		return RESOURCED_ERROR_FAIL;
	}
	status->curr_cap_counter[DISCHARGING] = atol(token);
	token = strtok_r(NULL, " ", &saveptr);
	if (!token) {
		_E("Failed to token value");
		return RESOURCED_ERROR_FAIL;
	}
	status->curr_run_time_sec[CHARGING] = atol(token);
	token = strtok_r(NULL, " ", &saveptr);
	if (!token) {
		_E("Failed to token value");
		return RESOURCED_ERROR_FAIL;
	}
	status->curr_cap_counter[CHARGING] = atol(token);
	token = strtok_r(NULL, " ", &saveptr);
	if (!token) {
		_E("Failed to token value");
		return RESOURCED_ERROR_FAIL;
	}
	status->curr_charger_status = atoi(token);
	token = strtok_r(NULL, " ", &saveptr);
	if (!token) {
		_E("Failed to token value");
		return RESOURCED_ERROR_FAIL;
	}
	status->reset_mark = atoi(token);
	return RESOURCED_ERROR_NONE;
};

static int heart_battery_load_usage(char *key, struct battery_usage *usage, int total_size)
{
	int i, num, ret;
	char *token;
	char buf[BATTERY_DATA_MAX] = {0, };
	char *saveptr;

	if (!key || !usage)
		return RESOURCED_ERROR_FAIL;

	ret = logging_leveldb_read(key, strlen(key), buf, sizeof(buf));
	if (ret != RESOURCED_ERROR_NONE) {
		_E("Failed to read leveldb key");
		return RESOURCED_ERROR_FAIL;
	}

	if (buf[0] == '\0') {
		_D("There is no history about %s", key);
		return RESOURCED_ERROR_NONE;
	}

	i = 0;
	num = total_size/sizeof(struct battery_usage);
	token = strtok_r(buf, " ", &saveptr);
	if (!token) {
		_E("Failed to token value");
		return RESOURCED_ERROR_FAIL;
	}
	while (token && i++ < num) {
		usage[i].start_time = atol(token);
		token = strtok_r(NULL, " ", &saveptr);
		if (!token) {
			_E("Failed to token value");
			return RESOURCED_ERROR_FAIL;
		}
		usage[i].sec_per_cap[DISCHARGING] = atol(token);
		token = strtok_r(NULL, " ", &saveptr);
		if (!token) {
			_E("Failed to token value");
			return RESOURCED_ERROR_FAIL;
		}
		usage[i].cap_counter[DISCHARGING] = atol(token);
		token = strtok_r(NULL, " ", &saveptr);
		if (!token) {
			_E("Failed to token value");
			return RESOURCED_ERROR_FAIL;
		}
		usage[i].sec_per_cap[CHARGING] = atol(token);
		token = strtok_r(NULL, " ", &saveptr);
		if (!token) {
			_E("Failed to token value");
			return RESOURCED_ERROR_FAIL;
		}
		usage[i].cap_counter[CHARGING] = atol(token);
		token = strtok_r(NULL, " ", &saveptr);
		_D("load [%d] stime: %ld, spc: %ld, count: %ld, spc: %ld, count: %ld",
				i, usage[i].start_time, usage[i].sec_per_cap[DISCHARGING],
				usage[i].cap_counter[DISCHARGING], usage[i].sec_per_cap[CHARGING],
				usage[i].cap_counter[CHARGING]);
	}
	return RESOURCED_ERROR_NONE;
};

static int heart_battery_load_prediction(char *key, struct battery_prediction *prediction)
{
	int ret, i;
	char *token;
	char buf[BATTERY_DATA_MAX] = {0, };
	char *saveptr;

	if (!key || !prediction)
		return RESOURCED_ERROR_FAIL;

	ret = logging_leveldb_read(key, strlen(key), buf, sizeof(buf));
	if (ret != RESOURCED_ERROR_NONE) {
		_E("Failed to read leveldb key: %s", key);
		return RESOURCED_ERROR_FAIL;
	}
	token = strtok_r(buf, " ", &saveptr);
	if (!token) {
		_E("Failed to token value");
		return RESOURCED_ERROR_FAIL;
	}
	for (i = 0; i < MAX_STRATEGY && token; i++) {
		prediction[DISCHARGING].sec_per_cap[i] = atol(token);
		token = strtok_r(NULL, " ", &saveptr);
		if (!token) {
			_E("Failed to token value");
			return RESOURCED_ERROR_FAIL;
		}
		prediction[DISCHARGING].cap_counter[i] = atol(token);
		token = strtok_r(NULL, " ", &saveptr);
		if (!token) {
			_E("Failed to token value");
			return RESOURCED_ERROR_FAIL;
		}
		prediction[DISCHARGING].time_pred_min[i] = atol(token);
		token = strtok_r(NULL, " ", &saveptr);
		if (!token) {
			_E("Failed to token value");
			return RESOURCED_ERROR_FAIL;
		}
		prediction[CHARGING].sec_per_cap[i] = atol(token);
		token = strtok_r(NULL, " ", &saveptr);
		if (!token) {
			_E("Failed to token value");
			return RESOURCED_ERROR_FAIL;
		}
		prediction[CHARGING].cap_counter[i] = atol(token);
		token = strtok_r(NULL, " ", &saveptr);
		if (!token) {
			_E("Failed to token value");
			return RESOURCED_ERROR_FAIL;
		}
		prediction[CHARGING].time_pred_min[i] = atol(token);
		token = strtok_r(NULL, " ", &saveptr);
	}
	return RESOURCED_ERROR_NONE;
};

static void heart_battery_update_used_time(time_t now, enum charger_status_type status)
{
	if (batt_used.last_charger_status == DISCHARGING)
		batt_used.used_time_sec +=
			now - batt_used.last_update_time;
	batt_used.last_charger_status = status;
	batt_used.last_update_time = now;
	heart_battery_save_used_time(BATTERY_USED_TIME, &batt_used);
}

static int heart_battery_get_capacity_history_size(void)
{
	int size, ret;

	ret = pthread_mutex_lock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	size = g_slist_length(capacity_history_list);
	if (!size) {
		_I("capacity history is empty");
		ret = pthread_mutex_unlock(&heart_battery_mutex);
		if (ret) {
			_E("pthread_mutex_unlock() failed, %d", ret);
			return RESOURCED_ERROR_FAIL;
		}
		return RESOURCED_ERROR_NONE;
	}
	ret = pthread_mutex_unlock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	return size;
}

static void heart_battery_insert_capacity(GSList **history_list, int capacity,
		int diff_capacity, time_t timestamp, long used_time, long charging_time,
		enum charger_status_type charger_status, int reset_mark, int clear)
{
	static int old_reset_mark = 0;
	GSList *iter, *next;
	int ret, count;
	struct heart_battery_capacity *lbc, *tlbc;

	lbc = malloc(sizeof(struct heart_battery_capacity));
	if (!lbc) {
		_E("malloc failed");
		return;
	}
	lbc->capacity = capacity;
	lbc->diff_capacity = diff_capacity;
	lbc->used_time = used_time;
	lbc->charging_time = charging_time;
	lbc->charger_status = charger_status;
	lbc->reset_mark = reset_mark;
	lbc->timestamp = timestamp;

	ret = pthread_mutex_lock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		free(lbc);
		return;
	}
	/* clean all history when reset event */
	if (clear && *history_list && lbc->reset_mark != old_reset_mark) {
		g_slist_free_full(*history_list, free);
		*history_list = NULL;
	}

	/* history reached maximum limitation number */
	if (*history_list && g_slist_length(*history_list) > BATTERY_CAPACITY_MAX) {
		count = 0;
		gslist_for_each_safe(*history_list, iter, next, tlbc) {
			*history_list = g_slist_remove(*history_list, (gpointer)tlbc);
			free(tlbc);
			if (BATTERY_CLEAN_MAX < count++)
				break;
		}
	}
	old_reset_mark = lbc->reset_mark;
	*history_list = g_slist_append(*history_list, (gpointer)lbc);
	ret = pthread_mutex_unlock(&heart_battery_mutex);
	if (ret)
		_E("pthread_mutex_unlock() failed, %d", ret);
}

/* ======================== Serialization/Deserialization ==================== */

static void heart_battery_status_save_to_db(void)
{
	heart_battery_save_used_time(BATTERY_USED_TIME, &batt_used);
	heart_battery_save_status(BATTERY_STATUS, &batt_stat);

	heart_battery_save_usage(BATTERY_RESET_USAGE, batt_stat.batt_reset_usage, sizeof(batt_stat.batt_reset_usage));
	heart_battery_save_usage(BATTERY_WEEK_DAY_USAGE, batt_stat.week_day_usage, sizeof(batt_stat.week_day_usage));
	heart_battery_save_usage(BATTERY_LEVEL_USAGE, batt_stat.batt_lvl_usage, sizeof(batt_stat.batt_lvl_usage));

	heart_battery_save_prediction(BATTERY_PREDICTION, batt_stat.prediction);
}

static int heart_battery_status_read_from_db(void)
{
	bool ok = true;

	ok &= (RESOURCED_ERROR_NONE == heart_battery_load_used_time(BATTERY_USED_TIME, &batt_used));
	ok &= (RESOURCED_ERROR_NONE == heart_battery_load_status(BATTERY_STATUS, &batt_stat));

	ok &= (RESOURCED_ERROR_NONE == heart_battery_load_usage(BATTERY_RESET_USAGE, batt_stat.batt_reset_usage, sizeof(batt_stat.batt_reset_usage)));
	ok &= (RESOURCED_ERROR_NONE == heart_battery_load_usage(BATTERY_WEEK_DAY_USAGE, batt_stat.week_day_usage, sizeof(batt_stat.week_day_usage)));
	ok &= (RESOURCED_ERROR_NONE == heart_battery_load_usage(BATTERY_LEVEL_USAGE, batt_stat.batt_lvl_usage, sizeof(batt_stat.batt_lvl_usage)));

	ok &= (RESOURCED_ERROR_NONE == heart_battery_load_prediction(BATTERY_PREDICTION, batt_stat.prediction));

	return ok ? RESOURCED_ERROR_NONE : RESOURCED_ERROR_FAIL;
}

static int heart_battery_capacity_save_to_file(char *filename)
{
	int size, ret, count, len = 0;
	struct heart_battery_capacity *lbc;
	GSList *iter, *next;
	FILE *fp;
	char buf[BATTERY_DATA_MAX] = {0, };

	if (!capacity_history_list) {
		_E("capacity history is NULL!");
		return RESOURCED_ERROR_NONE;
	}
	ret = pthread_mutex_lock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	size = g_slist_length(capacity_history_list);
	if (!size) {
		_I("capacity history is empty");
		ret = pthread_mutex_unlock(&heart_battery_mutex);
		if (ret) {
			_E("pthread_mutex_unlock() failed, %d", ret);
			return RESOURCED_ERROR_FAIL;
		}
		return RESOURCED_ERROR_NONE;
	}
	fp = fopen(filename, "w");
	if (!fp) {
		_E("%s fopen failed %d", filename, errno);
		ret = pthread_mutex_unlock(&heart_battery_mutex);
		if (ret) {
			_E("pthread_mutex_unlock() failed, %d", ret);
			return RESOURCED_ERROR_FAIL;
		}
		return RESOURCED_ERROR_FAIL;
	}
	gslist_for_each_item(iter, capacity_history_list) {
		lbc = (struct heart_battery_capacity *)iter->data;
		if (!lbc)
			break;
		len += snprintf(buf + len, BATTERY_DATA_MAX - len, "%d %d %ld %ld %ld %d %d\n",
				lbc->capacity, lbc->diff_capacity, lbc->timestamp, lbc->used_time,
				lbc->charging_time, lbc->charger_status,
				lbc->reset_mark);
		if (BATTERY_DATA_MAX < len + BATTERY_LINE_MAX) {
			fputs(buf, fp);
			len = 0;
		}
	}
	fputs(buf, fp);
	fclose(fp);
	if (BATTERY_CAPACITY_MAX < size) {
		count = 0;
		gslist_for_each_safe(capacity_history_list, iter, next, lbc) {
			capacity_history_list = g_slist_remove(capacity_history_list, (gpointer)lbc);
			free(lbc);
			if (BATTERY_CLEAN_MAX < count++)
				break;
		}
	}
	ret = pthread_mutex_unlock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	return RESOURCED_ERROR_NONE;
}

static int heart_battery_capacity_read_from_file(char *filename)
{
	int len;
	int capacity, diff_capacity, charger_status, reset_mark;
	long used_time, charging_time;
	time_t timestamp;
	FILE *fp;
	char buf[BATTERY_DATA_MAX] = {0, };
	int ret;

	fp = fopen(filename, "r");
	if (!fp) {
		if (errno != ENOENT) {
			_E("Fail to open %s (%d)", buf, errno);
			return RESOURCED_ERROR_FAIL;
		}

		_D("%s doesn't exist. Make new one.", buf);
		ret = mkdir(HEART_FILE_PATH, S_IRWXU | S_IRWXG | S_IROTH);
		if (ret != 0 && errno != EEXIST) {
			_E("Fail to create %s (%d)", buf, errno);
			return RESOURCED_ERROR_FAIL;
		}

		fp = fopen(filename, "w+");
		if (!fp) {
			_E("Fail to create %s (%d)", buf, errno);
			return RESOURCED_ERROR_FAIL;
		}
		fclose(fp);
		return RESOURCED_ERROR_NONE;
	}

	while (fgets(buf, BATTERY_DATA_MAX, fp)) {
		len = sscanf(buf, "%d %d %ld %ld %ld %d %d", &capacity, &diff_capacity,
				&timestamp, &used_time, &charging_time,
				&charger_status, &reset_mark);
		if (len < 0) {
			_E("sscanf failed");
			fclose(fp);
			return RESOURCED_ERROR_FAIL;
		}
		heart_battery_insert_capacity(&capacity_history_list, capacity, diff_capacity,
				timestamp, used_time, charging_time,
				charger_status, reset_mark, true);
	}
	fclose(fp);
	return RESOURCED_ERROR_NONE;
}

/* ==================== Serialization/Deserialization END ==================== */

static void heart_battery_save_to_file(bool force)
{
	int ret;
	time_t now = logging_get_time(CLOCK_BOOTTIME);

	heart_battery_update_used_time(now, batt_stat.curr_charger_status);

	if (!force &&
	    heart_battery_get_file_commit_timestamp() + HEART_BATTERY_SAVE_INTERVAL >= now)
		return;

	heart_battery_status_save_to_db();

	ret = heart_battery_capacity_save_to_file(HEART_BATTERY_CAPACITY_DATA_FILE);
	if (ret)
		_E("failed to save capacity file");
	heart_battery_set_file_commit_timestamp(now);
}

/*
 * This function will open the '.battery_data.dat' file and write all battery
 * related data into it.
 */
static void heart_battery_write_data_to_file(void)
{
	int i;
	int len = 0;
	char buff[BATTERY_DATA_MAX] = {0, 0};
	_cleanup_fclose_ FILE *fp = NULL;

	fp = fopen(HEART_BATTERY_DATA_FILE, "w");
	if (fp == NULL) {
		_E("Could not open file battery_data file\n");
		return;
	}

	for (i = 0; i < INDEX_WINDOW_SIZE; i++) {
		len += snprintf(buff + len, BATTERY_DATA_MAX - len, "%ld %lf\n", \
			batt_stat.last_wall_time[i], batt_stat.last_volt_intg[i]);

		if (BATTERY_DATA_MAX < len + BATTERY_LINE_MAX) {
			fputs(buff, fp);
			len = 0;
		}
	}
	len += snprintf(buff + len, BATTERY_DATA_MAX - len, "%d %d %d %d %lf %d", \
		batt_stat.curr_index, batt_stat.last_capacity, batt_stat.remaining_time, \
		batt_stat.remaining_time_ups, batt_stat.last_pwr_bchg, batt_stat.data_available);

	fputs(buff, fp);
}

/*
 * This function will open the '.battery_data.dat' file and read all battery
 * related data from it and store in variable 'batt_stat'
 */
static void heart_battery_read_data_from_file(void)
{
	int i;
	int len = 0;
	char buff[BATTERY_DATA_MAX] = {0, 0};
	char *ch;
	_cleanup_fclose_ FILE *fp = NULL;

	fp = fopen(HEART_BATTERY_DATA_FILE, "r");
	if (fp == NULL) {
		_E("Could not open battery_data file\n");
		return;
	}

	for (i = 0; i < INDEX_WINDOW_SIZE; i++) {
		ch = fgets(buff, BATTERY_DATA_MAX, fp);
		if (ch == NULL)
			/*if file is empty then return from here */
			return;

		len += sscanf(buff, "%ld %lf", \
				&batt_stat.last_wall_time[i], &batt_stat.last_volt_intg[i]);
	}

	ch = fgets(buff, BATTERY_DATA_MAX, fp);
	if (ch != NULL)
		len += sscanf(buff, "%d %d %d %d %lf %d", \
			&batt_stat.curr_index, &batt_stat.last_capacity, &batt_stat.remaining_time, \
			&batt_stat.remaining_time_ups, &batt_stat.last_pwr_bchg, &batt_stat.data_available);
}

/*
 * This function will reject all the battery related data which were stored earlier
 */
static void heart_battery_reject_data(void)
{
	int i;

	for (i = 0; i < INDEX_WINDOW_SIZE; i++) {
		batt_stat.last_wall_time[i] = 0;
		batt_stat.last_volt_intg[i] = 0.0;
	}

	/* Do not reject this, once data available, do not change untill next binary flash */
	/*batt_stat.data_available = 0;   */
	batt_stat.last_capacity = 0;
	batt_stat.curr_index = 0;
	batt_stat.remaining_time = 0;
	batt_stat.remaining_time_ups = 0;
	/* do not flush this value. When proper power value is not present, previously
	calculated power only will be used. */
	/*batt_stat.last_pwr_bchg = 0.0;  */

	heart_battery_write_data_to_file();
}

/*
 * This function will calculate the voltage integral(energy) at battery level 100.
 * We will calculate Pivot using fast discharge and slow discharge rate for
 * estimating the battery time in normal mode
 */
static void heart_battery_calculate_pivot(void)
{
	int i;

	for (i = 0; i < ocv_degree; i++)
		volt_intg_full += (intg[i] /(i+1)) * pow(100 , i+1);
	volt_intg_full = volt_intg_full/100000.0;

	pivot_nor = ((sqrt(discharge_fast) + sqrt(discharge_slow))/2);
	pivot_nor = pivot_nor * pivot_nor; // taking square of Tpivot

	pivot_ups = ((sqrt(discharge_fast*UPS_FACTOR) + sqrt(discharge_slow*UPS_FACTOR))/2);
	pivot_ups = pivot_ups * pivot_ups; // taking square of Tpivot
}

/* function to return the time(in seconds) since the  Epoch*/
static long heart_battery_logging_get_time_sec_new(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);

	return (tv.tv_sec + tv.tv_usec / 1000000);
}

/*
 * This function will read the current time and will compare with previously stored time.
 * If time difference is more then it is assumed that device was powered off and it will
 * adjust all stored time values
 */
static void heart_battery_power_off_time_adjustment(void)
{
	int i;
	int index;
	int clock_tick_diff = 0;
	long int curr_time;
	long int time_diff = 0;

	curr_time = heart_battery_logging_get_time_sec_new();

	index = BATTERY_WINDOW_INDEX(batt_stat.curr_index - 1);
	if (batt_stat.last_wall_time[index] == 0)
		index = 0;

	clock_tick_diff = logging_get_time(CLOCK_BOOTTIME) - batt_stat.last_clock_tick;

	/* check if we have store last time value or not */
	if (batt_stat.last_wall_time[index] != 0) {
		time_diff = curr_time - batt_stat.last_wall_time[index];
		time_diff -= clock_tick_diff;
	}

	_I("All store time value will be scaled up by %ld secs\n", time_diff);

	/*scale up all stored time value */
	for (i = 0; i < INDEX_WINDOW_SIZE; i++) {
		if (batt_stat.last_wall_time[i] == 0) {
			/* we do not have stored time value from this point onwards, stop updating*/
			break;
		} else
			batt_stat.last_wall_time[i] += time_diff;
	}

	/* update same value in battery data file*/
	heart_battery_write_data_to_file();

	/*
	 * during charge time
	 */
	time_diff = curr_time - batt_stat.last_wall_time_chg;
	time_diff -= clock_tick_diff;
	/*
	  * update last_wall_time_chg as per previous and new time diff
	  * it will take care of positive of negative difference
	  */
	batt_stat.last_wall_time_chg += time_diff;
	_I("last_wall_time_chg got updated by %ld\n", time_diff);
	/*
	  * make first_level_change as TRUE so that charge remaining time
	  * should not vary much
	  */
	first_level_change = TRUE;
}

/*
 * In case of any time change in the device, this callback get called and it
 * will do the time adjustment
 */
static void time_change_notify_cb(keynode_t *key, void *data)
{
	heart_battery_power_off_time_adjustment();
}

static int get_battery_remaining_time_new(void)
{
	return batt_stat.remaining_time;
}

static int get_battery_remaining_time_ups()
{
	return batt_stat.remaining_time_ups;
}

/*
 * This function will return the average of stored time diff values
 */
static int heart_battery_get_time_diff_avg(int time_diff_avg)
{
	int temp_index;
	int i;

	if (data_avail_chg == true) {
		/*
		 * calculate the average of last 5(including current) time diff values
		 */
		temp_index = ((batt_stat.index_chg-1) + BATTERY_LEVEL_GAP) % BATTERY_LEVEL_GAP;
		for (i = 1; i < BATTERY_LEVEL_GAP; i++) {
			time_diff_avg += batt_stat.last_wall_time_chg_diff[temp_index];
			temp_index = ((temp_index-1) + BATTERY_LEVEL_GAP) % BATTERY_LEVEL_GAP;
		}
		time_diff_avg = time_diff_avg/BATTERY_LEVEL_GAP;
	} else {
		/*
		 * take average of only available data. Do not include 1st value so
		 * i stared from 1.
		 */
		for (i = 1; i < batt_stat.index_chg; i++)
			time_diff_avg += batt_stat.last_wall_time_chg_diff[i];

		time_diff_avg = time_diff_avg/i;
	}
	/*
	 * to minimize the time diff fluctuation
	 */
	time_diff_avg = (time_diff_avg+BATTERY_C_RATE)/2;
	_I("time_diff after averaging= %d seconds\n", time_diff_avg);

	return time_diff_avg;
}

/*
 * This function will return the time diff between last and current battery
 * level change
 */
int heart_battery_get_time_diff(long curr_wall_time)
{
	if (batt_stat.last_wall_time_chg == 0)
		/*
		 * First time estimation
		 * In this case charge rate will be considered as time diff
		 */
		return BATTERY_C_RATE;

	return (curr_wall_time - batt_stat.last_wall_time_chg);
}

static void heart_battery_print_prev_charge_data(void)
{
	int i;

	_I("Charging stored data, current index = %d", batt_stat.index_chg);

	for (i = 0; i < BATTERY_LEVEL_GAP; i++)
		_I("Index = %d, time_diff = %d", i, batt_stat.last_wall_time_chg_diff[i]);
}

/*
 * This function will flush the stored charge data
 */
static void heart_battery_flush_charge_data(void)
{
	int i;

	 _I("Prev and curr battery lvl gap = %d, flush charge data",
		batt_stat.last_capacity_chg - batt_stat.curr_capacity);

	 for (i = 0; i < BATTERY_LEVEL_GAP; i++)
		batt_stat.last_wall_time_chg_diff[i] = 0;

	 batt_stat.last_wall_time_chg = 0;
	 batt_stat.index_chg = 0;
	 data_avail_chg = false;
}

/*
 * This function will calculate the battery charging remaining time
 */
static void heart_battery_cal_charging_rem_time(void)
{
	long curr_wall_time = 0;
	int rem_time;
	int time_diff;
	int time_diff_avg;
	double temp_cv;

	if (batt_stat.last_capacity_chg == batt_stat.curr_capacity) {
		_I("Last capacity %d is same as current capacity %d\n",
			batt_stat.last_capacity_chg, batt_stat.curr_capacity);
		return;
	}

	curr_wall_time = heart_battery_logging_get_time_sec_new();

	time_diff = heart_battery_get_time_diff(curr_wall_time);

	/*
	 * just after inserting the charger, if time diff between two level
	 * is less than MIN_TIME_FOR_LVL_CHANGE value then consider time diff
	 * as BATTERY_C_RATE so that there will not be sudden fall in charge
	 * remaining time
	 */
	if (time_diff < MIN_TIME_FOR_LVL_CHANGE && first_level_change == TRUE) {
		/*
		 * first_level_change will be set to TRUE in function
		 * heart_battery_charger_status() and set to FALSE in function
		 * heart_battery_capacity_status()
		 */
		_I("time diff %d is less than min value of lvl change %d\n",
			time_diff, MIN_TIME_FOR_LVL_CHANGE);
		time_diff = BATTERY_C_RATE;
	}

	_I("data_avail_chg = %d, index_chg = %d\n",
		data_avail_chg, batt_stat.index_chg);
	_I("last_wall_time_chg = %ld, curr_wall_time = %ld\n",
		batt_stat.last_wall_time_chg, curr_wall_time);
	_I("time diff before averaging = %d seconds\n", time_diff);

	/*
	 * get the average of all time diff values stored including current
	 */
	time_diff_avg = heart_battery_get_time_diff_avg(time_diff);

	if (time_diff_avg > BATTERY_C_RATE && batt_stat.curr_capacity <= 30)
		time_diff_avg = BATTERY_C_RATE;

	if (batt_stat.curr_capacity == FULL_CAPACITY) {
		batt_stat.remaining_time_chg = 0;
	} else {
		/*
		 * calculate the remaining charge time based on constant current
		 * and constant votlage(CCCV) logic. 'time_diff_avg' is for CC and
		 * 'temp_cv' is for CV part.
		 */
		temp_cv = ((double)time_diff_avg)/((double)BATTERY_C_RATE);
		temp_cv = temp_cv * temp_cv;
		rem_time = (time_diff_avg + temp_cv)*(FULL_CAPACITY - batt_stat.curr_capacity);
		batt_stat.remaining_time_chg = SEC_TO_MIN(rem_time);
	}

	batt_stat.last_wall_time_chg = curr_wall_time;
	batt_stat.last_wall_time_chg_diff[batt_stat.index_chg] = time_diff;
	batt_stat.index_chg = (batt_stat.index_chg+1)%BATTERY_LEVEL_GAP;
	batt_stat.last_capacity_chg = batt_stat.curr_capacity;
	batt_stat.last_clock_tick =  logging_get_time(CLOCK_BOOTTIME);

	_I("Capacity = %d, charging remaining time = %d minutes\n",
		batt_stat.curr_capacity, batt_stat.remaining_time_chg);

	if (batt_stat.index_chg+1 == BATTERY_LEVEL_GAP)
		data_avail_chg = true;
}

/*
 * This function will calculate the battery estimation time for every battery
 * level change
 */
static void heart_battery_cal_discharge_rem_time(void)
{
	double curr_volt_intg = 0;
	double volt_intg_diff = 0;
	double ps, pt;
	double batt_pwr;
	int rem_time = 0;
	int index = 0;
	int temp_index = 0;
	int index_count;
	int i;
	int update_time;
	long curr_wall_time = 0;
	long time_diff1, time_diff2;

	_I("last capacity = %d, current capacity = %d\n", batt_stat.last_capacity, batt_stat.curr_capacity);

	/* if battery level change happen then only calculate the new time*/
	if (batt_stat.last_capacity != batt_stat.curr_capacity) {
		/* calculate the voltage integral(energy) for current capacity */
		for (i = 0; i < ocv_degree; i++)
			curr_volt_intg += (intg[i] /(i+1)) * pow(batt_stat.curr_capacity , i+1);
		curr_volt_intg = curr_volt_intg /100000.0;

		/*current time*/
		curr_wall_time = heart_battery_logging_get_time_sec_new();

		_I("curr_volt_intg = %lf, curr_wall_time = %ld\n", curr_volt_intg, curr_wall_time);

		if (batt_stat.last_capacity < batt_stat.curr_capacity) {
			/*
			 *this indicates that device was put for charging so need to reject
			 *all the previous data except 'batt_stat.data_available' and
			 *'batt_stat.last_pwr_bchg'. Use last available battery power
			 *before connecting the charger as current average power.
			 */
			batt_pwr = batt_stat.last_pwr_bchg;
			heart_battery_reject_data();
		} else {
			index = BATTERY_WINDOW_INDEX(batt_stat.curr_index - BATTERY_LEVEL_GAP);
			if (batt_stat.last_wall_time[index] == 0 &&
				batt_stat.last_volt_intg[index] == 0) {
				/*
				 *do not have enough previous data so take 1st stored data
				 *as reference
				 */
				time_diff1 = curr_wall_time - batt_stat.last_wall_time[0];
				volt_intg_diff = batt_stat.last_volt_intg[0] - curr_volt_intg;
				if (volt_intg_diff < DOUBLE_ZERO)
					volt_intg_diff = volt_intg_full - curr_volt_intg;
			} else {
				/* time differenc since last battery level change*/
				time_diff1 = curr_wall_time - batt_stat.last_wall_time[index];
				/* voltage integral(energy) change since last battery level change */
				volt_intg_diff = batt_stat.last_volt_intg[index] - curr_volt_intg;
			}
			_I("time_diff1 = %ld, volt_intg_diff = %lf\n", time_diff1, volt_intg_diff);

			time_diff1 = SEC_TO_MIN(time_diff1); /*convert second to minute */

			/* calculate short term average power component*/
			if (time_diff1 != 0)
				ps = (total_battery_capacity *volt_intg_diff) / time_diff1;
			else
				ps = 0;

			_I("ps = %lf\n", ps);

			/* find the voltage integral(energy) of 1 hr back battery level.
			1. Place the index at proper location
			2. get the time stored at that index
			3. if difference between last time and current time is >= 60 then take that voltage integral
			*/
			index_count = INDEX_WINDOW_SIZE;
			time_diff2 = 0;
			temp_index = BATTERY_WINDOW_INDEX(batt_stat.curr_index - 1);
			while ((index_count > 0) && (time_diff2 < 60)) {
				if  (batt_stat.last_wall_time[temp_index] == 0)
					/* data not available so break the loop */
					break;

				time_diff2 = SEC_TO_MIN(curr_wall_time - batt_stat.last_wall_time[temp_index]);
				temp_index = BATTERY_WINDOW_INDEX(temp_index - 1);
				index_count--;
			}

			/* calculate long term average power component*/
			if (time_diff2 < 60) {
				pt = 0;
			} else {
				temp_index = BATTERY_WINDOW_INDEX(temp_index + 1);
				volt_intg_diff = batt_stat.last_volt_intg[temp_index] - curr_volt_intg ;
				if (volt_intg_diff < DOUBLE_ZERO)
					pt = 0;
				else
					pt = (total_battery_capacity  * volt_intg_diff) / time_diff2;
			}

			_I("time_diff2 = %ld, volt_intg_diff = %lf, pt = %lf\n", time_diff2, volt_intg_diff, pt);

			/* calculate the average power from the short component and long component*/
			if ((pt + ps) != 0)
				batt_pwr = (((pt  * pt) + (ps * ps)) / (pt + ps));
			else
				batt_pwr = batt_stat.last_pwr_bchg;
		}

		/* total remaining time based on current battery level*/
		rem_time = (total_battery_capacity * curr_volt_intg) / batt_pwr;

		_I("pivot_nor = %lf, curr_volt_intg = %lf , volt_intg_full %lf, batt_pwr = %lf", pivot_nor, curr_volt_intg, volt_intg_full, batt_pwr);

		/* calculate remaining time for normal mode */
		update_time = (pivot_nor * curr_volt_intg) / volt_intg_full;
		batt_stat.remaining_time = (LONG_TIME_WEIGHT * update_time + SHORT_TIME_WEIGHT * CAL_MIN(update_time, rem_time));

		 /* calculate remaining time for UPS mode */
		update_time = (pivot_ups * curr_volt_intg) / volt_intg_full;
		batt_stat.remaining_time_ups = (LONG_TIME_WEIGHT * update_time + SHORT_TIME_WEIGHT * CAL_MIN(update_time, rem_time));

		_I("normal mode remaining time = %d minutes, ups remaining time = %d minutes\n",
			batt_stat.remaining_time, batt_stat.remaining_time_ups);

		batt_stat.last_volt_intg[batt_stat.curr_index] = curr_volt_intg;
		batt_stat.last_wall_time[batt_stat.curr_index] = curr_wall_time;
		batt_stat.last_capacity = batt_stat.curr_capacity;
		batt_stat.curr_index = (batt_stat.curr_index + 1) % INDEX_WINDOW_SIZE;
		batt_stat.last_clock_tick =  logging_get_time(CLOCK_BOOTTIME);

		if (batt_pwr > DOUBLE_ZERO) {
			/*
			 *when battery percentage is 0 then available power will be 0 so
			 *do not store this value. keep previous value only.
			 */
			batt_stat.last_pwr_bchg = batt_pwr;
		}

		if (batt_stat.curr_index > BATTERY_LEVEL_GAP) {
			/*
			 *When enough data to estimate the available battery time are
			 *stored, no need to update the status of batt_stat. data_available
			 *untill we are flashing new binary.
			 */
			batt_stat.data_available = TRUE;
		}

		/* save all calculated battery data to file*/
		heart_battery_write_data_to_file();
	}
}

void heart_battery_update(struct logging_table_form *data, void *user_data)
{
	heart_battery_save_to_file(false);
}

static int heart_battery_get_level_usage_index(int capacity)
{
	return (capacity > 49) ? BATTERY_LEVEL_HIGH :
		(capacity < 16) ? BATTERY_LEVEL_LOW : BATTERY_LEVEL_MID;
}

static int heart_battery_get_week_day_usage_index(time_t timestamp)
{
	int i;

	for (i = 0; i < BATTERY_HISTORY_DAY_MAX; i++) {
		if (!heart_battery_get_usage_week_stime(i))
			return i;
		else if (abs(timestamp - heart_battery_get_usage_week_stime(i)) < DAY_TO_SEC(1))
			return i;
	}
	for (i = 0; i < BATTERY_HISTORY_DAY_MAX - 1; i++) {
		batt_stat.week_day_usage[i].start_time =
			batt_stat.week_day_usage[i + 1].start_time;
		batt_stat.week_day_usage[i].sec_per_cap[DISCHARGING] =
			batt_stat.week_day_usage[i + 1].sec_per_cap[DISCHARGING];
		batt_stat.week_day_usage[i].sec_per_cap[CHARGING] =
			batt_stat.week_day_usage[i + 1].sec_per_cap[CHARGING];
		batt_stat.week_day_usage[i].cap_counter[DISCHARGING] =
			batt_stat.week_day_usage[i + 1].cap_counter[DISCHARGING];
		batt_stat.week_day_usage[i].cap_counter[CHARGING] =
			batt_stat.week_day_usage[i + 1].cap_counter[CHARGING];
	}
	return BATTERY_HISTORY_DAY_MAX - 1;
}

static int heart_battery_get_batt_reset_usage_index(void)
{
	int i;

	for (i = 0; i < BATTERY_HISTORY_RESET_MAX; i++) {
		if (heart_battery_get_usage_reset_count(i, DISCHARGING) < BATTERY_HISTORY_COUNT_MAX
			&& heart_battery_get_usage_reset_count(i, CHARGING) < BATTERY_HISTORY_COUNT_MAX)
			return i;
	}
	for (i = 0; i < BATTERY_HISTORY_RESET_MAX - 1; i++) {
		batt_stat.batt_reset_usage[i].start_time =
			batt_stat.batt_reset_usage[i + 1].start_time;
		batt_stat.batt_reset_usage[i].sec_per_cap[DISCHARGING] =
			batt_stat.batt_reset_usage[i + 1].sec_per_cap[DISCHARGING];
		batt_stat.batt_reset_usage[i].sec_per_cap[CHARGING] =
			batt_stat.batt_reset_usage[i + 1].sec_per_cap[CHARGING];
		batt_stat.batt_reset_usage[i].cap_counter[DISCHARGING] =
			batt_stat.batt_reset_usage[i + 1].cap_counter[DISCHARGING];
		batt_stat.batt_reset_usage[i].cap_counter[CHARGING] =
			batt_stat.batt_reset_usage[i + 1].cap_counter[CHARGING];
	}
	return BATTERY_HISTORY_RESET_CURRENT;
}

static int heart_battery_reset(void *data)
{
	int idx;
	long total_time, total_count, sec_per_cap;

	idx = heart_battery_get_batt_reset_usage_index();

	/* DISCHARGING */
	total_time = 0; total_count = 0;
	total_time = heart_battery_get_usage_reset_total_time(idx, DISCHARGING) + batt_stat.curr_run_time_sec[DISCHARGING];
	total_count = heart_battery_get_usage_reset_count(idx, DISCHARGING) + batt_stat.curr_cap_counter[DISCHARGING];

	if (total_time && total_count) {
		sec_per_cap = total_time / total_count;
		if (sec_per_cap < default_sec_per_cap[DISCHARGING][DEFAULT_MIN])
			sec_per_cap = default_sec_per_cap[DISCHARGING][DEFAULT_MIN];
		else if (sec_per_cap > default_sec_per_cap[DISCHARGING][DEFAULT_MAX])
			sec_per_cap = default_sec_per_cap[DISCHARGING][DEFAULT_MAX];
		heart_battery_set_usage_reset(idx, DISCHARGING, sec_per_cap, total_count);
	}
	/* CHARGING */
	total_time = 0; total_count = 0;
	total_time = heart_battery_get_usage_reset_total_time(idx, CHARGING)
		+ batt_stat.curr_run_time_sec[CHARGING];
	total_count = heart_battery_get_usage_reset_count(idx, CHARGING) + batt_stat.curr_cap_counter[CHARGING];

	if (total_time && total_count) {
		sec_per_cap = total_time / total_count;
		if (sec_per_cap < default_sec_per_cap[CHARGING][DEFAULT_MIN])
			sec_per_cap = default_sec_per_cap[CHARGING][DEFAULT_MIN];
		else if (sec_per_cap > default_sec_per_cap[CHARGING][DEFAULT_MAX])
			sec_per_cap = default_sec_per_cap[CHARGING][DEFAULT_MAX];
		heart_battery_set_usage_reset(idx, CHARGING, sec_per_cap, total_count);
	}

	batt_stat.reset_mark = batt_stat.reset_mark ? 0 : 1; /* Swap reset_mark */
	batt_stat.reset_mark_timestamp = time(NULL);
	batt_stat.curr_run_time_sec[DISCHARGING] = 0;
	batt_stat.curr_run_time_sec[CHARGING] = 0;
	batt_stat.curr_cap_counter[DISCHARGING] = 0;
	batt_stat.curr_cap_counter[CHARGING] = 0;
	batt_used.used_time_sec = 0;
	batt_used.last_update_time = logging_get_time(CLOCK_BOOTTIME);

	return RESOURCED_ERROR_NONE;
}

static long heart_battery_compute_remaining_time_in_min(int capacity_count, long sec_per_cap)
{
	/*
	 * Calculates and returns remaining time in minutes based on number
	 * of capacity changes and time needed for one change.
	 */
	long time;

	time = (capacity_count * sec_per_cap); /* seconds */
	time = time + 30; /* add 30s margin */
	time = time / 60; /* change to minutes */
	return time;
}

static void heart_battery_calculate_prediction(enum charger_status_type goal)
{
	int i, capacity, level;
	long total_time, total_count, sec_per_cap, pred_min;
	long low_count, mid_count, high_count;
	struct heart_battery_capacity *lbc = NULL;
	GArray *arrays = NULL;

	if (goal == CHARGING)
		capacity = REMAIN_CAPACITY(batt_stat.curr_capacity);
	else {
		capacity = batt_stat.curr_capacity;

		if (goal != DISCHARGING) {
			_E("Wrong charging status is written. Suppose discharging");
			goal = DISCHARGING;
		}
	}

	/* PREDICTION METHOD: total average */
	total_time = 0;
	total_count = 0;
	for (i = 0; i < BATTERY_HISTORY_RESET_MAX; i++) {
		total_time += heart_battery_get_usage_reset_total_time(i, goal);
		total_count += heart_battery_get_usage_reset_count(i, goal);
	}
	total_time += batt_stat.curr_run_time_sec[goal];
	total_count += batt_stat.curr_cap_counter[goal];

	if (total_time && total_count >= BATTERY_PREDICTION_DATA_MIN) {
		sec_per_cap = total_time / total_count;
		if (sec_per_cap < default_sec_per_cap[goal][DEFAULT_MIN])
			sec_per_cap = default_sec_per_cap[goal][DEFAULT_MIN];
		else if (sec_per_cap > default_sec_per_cap[goal][DEFAULT_MAX])
			sec_per_cap = default_sec_per_cap[goal][DEFAULT_MAX];
		pred_min = heart_battery_compute_remaining_time_in_min(capacity, sec_per_cap);
		heart_battery_set_prediction(TA, goal,
				sec_per_cap, total_count,
				pred_min);
	} else {
		heart_battery_set_prediction(TA, goal, 0, 0, 0);
	}


	/* PREDICTION METHOD:
	 * Prediction of battery remaining usage time
	 * considering users' psychological usage patterns
	 * by batt_lvl_usage of battery charge
	 * */
	pred_min = 0;
	sec_per_cap = 0;
	level = heart_battery_get_level_usage_index(capacity);
	low_count = heart_battery_get_usage_level_count(BATTERY_LEVEL_LOW, goal);
	mid_count = heart_battery_get_usage_level_count(BATTERY_LEVEL_MID, goal);
	high_count = heart_battery_get_usage_level_count(BATTERY_LEVEL_HIGH, goal);

	if (level == BATTERY_LEVEL_LOW && low_count) {
		sec_per_cap = heart_battery_get_usage_level_spc(BATTERY_LEVEL_LOW, goal);
		pred_min = heart_battery_compute_remaining_time_in_min(capacity, sec_per_cap);
	} else if (level == BATTERY_LEVEL_MID && low_count && mid_count) {
		sec_per_cap = heart_battery_get_usage_level_spc(BATTERY_LEVEL_LOW, goal);
		pred_min = heart_battery_compute_remaining_time_in_min(15, sec_per_cap);
		sec_per_cap = heart_battery_get_usage_level_spc(BATTERY_LEVEL_MID, goal);
		pred_min +=
		heart_battery_compute_remaining_time_in_min(capacity - 15, sec_per_cap);
	} else if (level == BATTERY_LEVEL_HIGH && low_count && mid_count && high_count) {
		sec_per_cap = heart_battery_get_usage_level_spc(BATTERY_LEVEL_LOW, goal);
		pred_min = heart_battery_compute_remaining_time_in_min(15, sec_per_cap);
		sec_per_cap = heart_battery_get_usage_level_spc(BATTERY_LEVEL_MID, goal);
		pred_min +=
			heart_battery_compute_remaining_time_in_min(35, sec_per_cap);
		sec_per_cap = heart_battery_get_usage_level_spc(BATTERY_LEVEL_HIGH, goal);
		pred_min +=
			heart_battery_compute_remaining_time_in_min(capacity - 50, sec_per_cap);
	}
	heart_battery_set_prediction(PCB, goal, 0, 0, pred_min);


	/* PREDICTION METHOD: week average */
	total_time = 0;
	total_count = 0;
	for (i = 0; i < BATTERY_HISTORY_DAY_MAX; i++) {
		total_time += heart_battery_get_usage_week_total_time(i, goal);
		total_count += heart_battery_get_usage_week_count(i, goal);
	}
	if (total_time && total_count >= BATTERY_PREDICTION_DATA_MIN) {
		sec_per_cap = total_time / total_count;
		if (sec_per_cap < default_sec_per_cap[goal][DEFAULT_MIN])
			sec_per_cap = default_sec_per_cap[goal][DEFAULT_MIN];
		else if (sec_per_cap > default_sec_per_cap[goal][DEFAULT_MAX])
			sec_per_cap = default_sec_per_cap[goal][DEFAULT_MAX];
		pred_min =
			heart_battery_compute_remaining_time_in_min(capacity, sec_per_cap);
		heart_battery_set_prediction(WEEK, goal, sec_per_cap, total_count, pred_min);
	} else
		heart_battery_set_prediction(WEEK, goal, 0, 0, 0);


	/* PREDICTION METHOD:  last BATTERY_PREDICTION_COUNT data average */
	arrays = g_array_new(FALSE, FALSE, sizeof(struct heart_battery_capacity *));
	if (!arrays) {
		_E("Failed to alloc array");
		return;
	}
	if (heart_battery_get_capacity_history_latest(arrays, goal, BATTERY_PREDICTION_LATEST_COUNT)
			!= RESOURCED_ERROR_NONE) {
		_E("Failed to get battery capacity history");
		g_array_free(arrays, TRUE);
		arrays = NULL;
		return;
	}
	total_time = 0;
	total_count = 0;
	for (i = 0; i < arrays->len; i++) {
		lbc = g_array_index(arrays, struct heart_battery_capacity *, i);
		if (!lbc)
			break;
		total_count += lbc->diff_capacity;
		if (goal == CHARGING)
			total_time += lbc->charging_time;
		else
			total_time += lbc->used_time;
		free(lbc);
	}
	if (total_time && total_count >= BATTERY_PREDICTION_DATA_MIN) {
		sec_per_cap = total_time / total_count;
		if (sec_per_cap < default_sec_per_cap[goal][DEFAULT_MIN])
			sec_per_cap = default_sec_per_cap[goal][DEFAULT_MIN];
		else if (sec_per_cap > default_sec_per_cap[goal][DEFAULT_MAX])
			sec_per_cap = default_sec_per_cap[goal][DEFAULT_MAX];

		pred_min =
			heart_battery_compute_remaining_time_in_min(capacity, sec_per_cap);
		heart_battery_set_prediction(COUNT, goal, sec_per_cap, total_count, pred_min);
		/* check current discharge rate */
		if (sec_per_cap && goal == DISCHARGING) {
			batt_stat.discharge_rate_level =
				heart_battery_calculate_discharge_rate_level(sec_per_cap);
			_I("discharge rate level : %d", batt_stat.discharge_rate_level);
		}
	} else
		heart_battery_set_prediction(COUNT, goal, 0, 0, 0);
	g_array_free(arrays, TRUE);
	arrays = NULL;


	/* PREDICTION METHOD: last BATTERY_PREDICTION_PERIOD hours average */
	arrays = g_array_new(FALSE, FALSE, sizeof(struct heart_battery_capacity *));
	if (!arrays) {
		_E("Failed to alloc array");
		return;
	}
	if (heart_battery_get_capacity_history(arrays, BATTERY_PREDICTION_PERIOD) != RESOURCED_ERROR_NONE) {
		_E("Failed to get battery capacity history");
		g_array_free(arrays, TRUE);
		arrays =  NULL;
		return;
	}
	total_time = 0;
	total_count = 0;
	for (i = 0; i < arrays->len; i++) {
		lbc = g_array_index(arrays, struct heart_battery_capacity *, i);
		if (!lbc)
			break;
		if (goal == CHARGING) {
			if (lbc->charger_status != CHARGING) {
				free(lbc);
				continue;
			}
			total_time += lbc->charging_time;
			total_count += lbc->diff_capacity;
		} else {
			if (lbc->charger_status != DISCHARGING) {
				free(lbc);
				continue;
			}
			total_time += lbc->used_time;
			total_count += lbc->diff_capacity;
		}
		free(lbc);
	}
	g_array_free(arrays, TRUE);
	arrays = NULL;
	if (total_time && total_count >= BATTERY_PREDICTION_DATA_MIN) {
		sec_per_cap = total_time / total_count;
		if (sec_per_cap < default_sec_per_cap[goal][DEFAULT_MIN])
			sec_per_cap = default_sec_per_cap[goal][DEFAULT_MIN];
		else if (sec_per_cap > default_sec_per_cap[goal][DEFAULT_MAX])
			sec_per_cap = default_sec_per_cap[goal][DEFAULT_MAX];
		pred_min =
			heart_battery_compute_remaining_time_in_min(capacity, sec_per_cap);
		heart_battery_set_prediction(PERIOD, goal, sec_per_cap, total_count, pred_min);

	} else
		heart_battery_set_prediction(PERIOD, goal, 0, 0, 0);

	/* Log values of all predictions calculated */
	for (i = 0; i < MAX_STRATEGY; i++) {
		_I("%s %d %ld %ld %ld",
				(goal == DISCHARGING) ? "TimeToEmpty:" : "TimeToFull:",
				batt_stat.curr_capacity,
				batt_stat.prediction[goal].sec_per_cap[i],
				batt_stat.prediction[goal].cap_counter[i],
				batt_stat.prediction[goal].time_pred_min[i]);
	}
}

static int heart_battery_add_capacity(int capacity)
{
	char info[BATTERY_DATA_MAX];
	int ret, idx, status;
	long time_diff_capacity_lvl[MAX_CHARGER_STATE];
	int diff_capacity_lvl;
	long total_time, total_count, sec_per_cap;
	time_t timestamp = time(NULL);
	time_t curr_wall_time = logging_get_time(CLOCK_BOOTTIME);

	status = batt_stat.curr_charger_status;
	/* calculate diff */
	time_diff_capacity_lvl[status] = curr_wall_time - batt_stat.last_event_wall_time;

	if (time_diff_capacity_lvl[status] < 0) {
		batt_stat.last_event_wall_time = curr_wall_time;
		return 0;
	}

	time_diff_capacity_lvl[!status] = 0;

	if (!batt_stat.curr_capacity)
		diff_capacity_lvl = 1;
	else
		diff_capacity_lvl = abs(batt_stat.curr_capacity - capacity);

	_I("%d -> %d %ld %ld", batt_stat.curr_capacity, capacity,
			timestamp, time_diff_capacity_lvl[status]);

	/* update battery current status */
	batt_stat.last_event_wall_time = curr_wall_time;
	batt_stat.curr_capacity = capacity;

	/* Full Charging status */
	if (status == CHARGING && !REMAIN_CAPACITY(capacity) && !diff_capacity_lvl)
		return 0;

	/* update run usage */
	batt_stat.curr_run_time_sec[status] += time_diff_capacity_lvl[status];
	batt_stat.curr_cap_counter[status] += diff_capacity_lvl;

	/* update batt_lvl_usage usage */
	total_time = 0;
	total_count = 0;

	if (status == CHARGING)
		idx = heart_battery_get_level_usage_index(REMAIN_CAPACITY(capacity));
	else
		idx = heart_battery_get_level_usage_index(capacity);

	total_time = heart_battery_get_usage_level_total_time(idx, status) + time_diff_capacity_lvl[status];
	if (total_time)
		total_count = heart_battery_get_usage_level_count(idx, status) + diff_capacity_lvl;

	if (total_count) {
		sec_per_cap = total_time / total_count;
		if (sec_per_cap == 0)
			sec_per_cap = default_sec_per_cap[status][DEFAULT_AVG];
		else if (sec_per_cap < default_sec_per_cap[status][DEFAULT_MIN])
			sec_per_cap = default_sec_per_cap[status][DEFAULT_MIN];
		else if (sec_per_cap > default_sec_per_cap[status][DEFAULT_MAX])
			sec_per_cap = default_sec_per_cap[status][DEFAULT_MAX];
		/*
		 * If counts reached MAXIMUM number,
		 * counts is divided by 2 to reduce previous data's effect to equation
		 */
		if (total_count >= BATTERY_HISTORY_COUNT_MAX)
			total_count = total_count >> 1;

		heart_battery_set_usage_level(idx, status, sec_per_cap, total_count);
		heart_battery_set_usage_level_stime(idx, timestamp);
	}

	/* update day usage */
	total_time = 0;
	total_count = 0;

	idx = heart_battery_get_week_day_usage_index(timestamp);
	total_time = heart_battery_get_usage_week_total_time(idx, status) + time_diff_capacity_lvl[status];
	if (total_time)
		total_count = heart_battery_get_usage_week_count(idx, status) + diff_capacity_lvl;

	if (total_count) {
		sec_per_cap = total_time / total_count;
		if (sec_per_cap == 0)
			sec_per_cap = default_sec_per_cap[status][DEFAULT_AVG];
		else if (sec_per_cap < default_sec_per_cap[status][DEFAULT_MIN])
			sec_per_cap = default_sec_per_cap[status][DEFAULT_MIN];
		else if (sec_per_cap > default_sec_per_cap[status][DEFAULT_MAX])
			sec_per_cap = default_sec_per_cap[status][DEFAULT_MAX];
		heart_battery_set_usage_week(idx, status, sec_per_cap, total_count);
		heart_battery_set_usage_week_stime(idx, CALCULATE_DAY_BASE_TIME(timestamp));
	}

	heart_battery_calculate_prediction(batt_stat.curr_charger_status);

	/* db backup */
	snprintf(info, sizeof(info), "%d %d %ld %ld %d %d ",
			capacity, diff_capacity_lvl,
			time_diff_capacity_lvl[DISCHARGING], time_diff_capacity_lvl[CHARGING],
			batt_stat.curr_charger_status, batt_stat.reset_mark);
	ret = logging_write(PID_FOR_ROOT, BATTERY_NAME, TIZEN_SYSTEM_BATTERY_APPID,
			TIZEN_SYSTEM_APPID, timestamp, info);
	if (ret != RESOURCED_ERROR_NONE)
		return ret;

	/* insert capacity history list */
	heart_battery_insert_capacity(&capacity_history_list, capacity, diff_capacity_lvl,
			timestamp, time_diff_capacity_lvl[DISCHARGING],
			time_diff_capacity_lvl[CHARGING], batt_stat.curr_charger_status,
			batt_stat.reset_mark, true);

	_D("battery_heart_capacity_write %d diff_capacity %d, used time %ld, charging time %ld, charger status %d, reset_mark %d",
			capacity, diff_capacity_lvl,
			time_diff_capacity_lvl[DISCHARGING], time_diff_capacity_lvl[CHARGING],
			batt_stat.curr_charger_status, batt_stat.reset_mark);

	resourced_notify(RESOURCED_NOTIFIER_LOGGING_START, NULL);
	return RESOURCED_ERROR_NONE;
}

/* ============================ DBUS -> DEVICED on demand ==================== */

static int heart_battery_direct_get_capacity(void)
{
	int capacity;

	capacity = d_bus_call_method_sync(DEVICED_BUS_NAME, DEVICED_PATH_BATTERY,
			DEVICED_INTERFACE_BATTERY,
			GET_BATTERY_CAPACITY,
			NULL, NULL);
	if (capacity < 0) {
		_E("Failed to sync DBUS message.");
		return RESOURCED_ERROR_FAIL;
	}

	return capacity;
}

static enum charger_status_type heart_battery_direct_get_charger_status(void)
{
	int status;

	status = d_bus_call_method_sync(DEVICED_BUS_NAME, DEVICED_PATH_BATTERY,
			DEVICED_INTERFACE_BATTERY,
			GET_CHARGER_STATUS,
			NULL, NULL);
	if (status < 0) {
		_E("Failed to sync DBUS message.");
		return RESOURCED_ERROR_FAIL;
	}

	if (status > 0)
		return CHARGING;
	return DISCHARGING;
}

/* =========================  DBUS -> DEVICED  on demand END ================= */

/* ============================ DBUS -> DEVICED handler ====================== */
static void heart_battery_capacity_status(GVariant *params)
{
	/*
	 * This handler is called when battery capacity value change in 1%
	 *
	 * The message have current percent value of capacity
	 *
	 * (This requires deviced with commit at least:
	 * "f1ae1d1f270e9 battery: add battery capacity dbus signal broadcast")
	 */

	int capacity = -1;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i)", &capacity);
	if (capacity < 0) {
		_E("Failed: dbus_message_get_args()");
		return;
	}
	heart_battery_add_capacity(capacity);
	heart_battery_update_used_time(logging_get_time(CLOCK_BOOTTIME),
			batt_stat.curr_charger_status);

	if (logic_v2) {
		/* get the device mode */
		if (vconf_get_int(VCONFKEY_HEART_BATTERY_DEVICE_MODE, &device_mode))
			_E("failed to get VCONFKEY_HEART_BATTERY_DEVICE_MODE\n");

		/* for every battery level change, calculate the battery estimation time using new logic*/
		heart_battery_cal_discharge_rem_time();

		if (batt_stat.curr_charger_status == CHARGING &&
			batt_stat.last_capacity_chg < batt_stat.curr_capacity) {
			heart_battery_cal_charging_rem_time();
			first_level_change = FALSE;
		}
	}
}

static void heart_battery_charger_status(GVariant *params)
{
	/*
	 * This handler is called when USB cable with charging capabilities
	 * is connected or disconnected from the device.
	 *
	 * The message have current status of charger connection.
	 * STATUSES:
	 * 0 - charger was disconnected
	 * 1 - charger was connected
	 */
	int charger_status = -1, cap_history_size;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i)", &charger_status);
	if (charger_status < 0) {
		_E("Failed: dbus_message_get_args()");
		return;
	}

	/* Update the statistics with capacity when charger state was changed */
	heart_battery_add_capacity(batt_stat.curr_capacity);

	cap_history_size = heart_battery_get_capacity_history_size();

	if (charger_status == DISCHARGING && batt_stat.curr_capacity >= 90) {
		/*
		 * If battery is charged over 90 and charger was disconnected.
		 * So most probably the phone was "charged".
		 * Let's reset the statistics.
		 */
		resourced_notify(RESOURCED_NOTIFIER_DATA_RESET, NULL);
	} else if (charger_status == DISCHARGING && cap_history_size >= BATTERY_CAPACITY_MAX) {
		/*
		 * Charger is not connected and the battery history is over limit.
		 * Let's reset the statistics.
		 */
		resourced_notify(RESOURCED_NOTIFIER_DATA_RESET, NULL);
	}
	/* Update current charger connection status */
	batt_stat.curr_charger_status = charger_status;
	heart_battery_update_used_time(logging_get_time(CLOCK_BOOTTIME),
			batt_stat.curr_charger_status);
	heart_battery_calculate_prediction(batt_stat.curr_charger_status);

	if (logic_v2) {
		/*
		 * If charger removed/inserted then update the battery estimation time
		 */
		heart_battery_cal_discharge_rem_time();

		if (charger_status == CHARGING) {
			/*
			 * if diff is more than 10% then flush the previous charge data
			 */
			if ((batt_stat.last_capacity_chg > 0) &&
			    (batt_stat.last_capacity_chg - batt_stat.curr_capacity > BATT_CHG_FLUSH_DATA))
				heart_battery_flush_charge_data();

			batt_stat.last_wall_time_chg = 0;
			first_level_change = TRUE;
			heart_battery_print_prev_charge_data();
			heart_battery_cal_charging_rem_time();
		}
	}
}

/* =========================  DBUS -> DEVICED handler END ==================== */

int heart_battery_get_capacity_history_latest(GArray *arrays, int charge, int max_size)
{
	int ret, size, count;
	struct heart_battery_capacity *lbc, *lbci;
	GSList *iter, *rlist;

	if (!capacity_history_list) {
		_E("empty capacity history list");
		return RESOURCED_ERROR_FAIL;
	}
	size = g_slist_length(capacity_history_list);
	if (!size) {
		_I("capacity history is empty");
		return RESOURCED_ERROR_NONE;
	}
	ret = pthread_mutex_lock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	count = 0;

	rlist = g_slist_copy(capacity_history_list);

	rlist = g_slist_reverse(rlist);

	gslist_for_each_item(iter, rlist) {
		lbc = (struct heart_battery_capacity *)iter->data;
		if (!lbc)
			break;
		if (charge < MAX_CHARGER_STATE && charge != lbc->charger_status)
			continue;
		count++;
		if (max_size < count)
			break;
		lbci = malloc(sizeof(struct heart_battery_capacity));
		if (!lbci) {
			_E("malloc failed");
			goto unlock_exit;
		}
		lbci->capacity = lbc->capacity;
		lbci->diff_capacity = lbc->diff_capacity;
		if (!lbc->diff_capacity)
			count--;
		lbci->used_time = lbc->used_time;
		lbci->charging_time = lbc->charging_time;
		lbci->charger_status = lbc->charger_status;
		g_array_prepend_val(arrays, lbci);
	}
unlock_exit:
	g_slist_free(rlist);
	ret = pthread_mutex_unlock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	return RESOURCED_ERROR_NONE;
}

int heart_battery_get_capacity_history(GArray *arrays, enum heart_data_period period)
{
	int ret, index, size;
	struct heart_battery_capacity *lbc, *lbci;
	GSList *iter;
	time_t curr = time(NULL);

	switch (period) {
	case DATA_LATEST:
		index = 0;
		break;
	case DATA_3HOUR:
		index = 3;
		break;
	case DATA_6HOUR:
		index = 6;
		break;
	case DATA_12HOUR:
		index = 12;
		break;
	case DATA_1DAY:
		index = 24;
		break;
	default:
		_E("Wrong message arguments! %d", period);
		return RESOURCED_ERROR_FAIL;
	}

	if (!capacity_history_list) {
		_E("empty capacity history list");
		return RESOURCED_ERROR_FAIL;
	}
	size = g_slist_length(capacity_history_list);
	if (!size) {
		_I("capacity history is empty");
		return RESOURCED_ERROR_NONE;
	}
	ret = pthread_mutex_lock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	gslist_for_each_item(iter, capacity_history_list) {
		lbc = (struct heart_battery_capacity *)iter->data;
		if (!lbc)
			break;
		if (index && (lbc->timestamp < curr - (index * 3600)))
			continue;
		lbci = malloc(sizeof(struct heart_battery_capacity));
		if (!lbci) {
			_E("malloc failed");
			ret = pthread_mutex_unlock(&heart_battery_mutex);
			if (ret) {
				_E("pthread_mutex_unlock() failed, %d", ret);
				return RESOURCED_ERROR_FAIL;
			}
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}
		lbci->capacity = lbc->capacity;
		lbci->diff_capacity = lbc->diff_capacity;
		lbci->used_time = lbc->used_time;
		lbci->charging_time = lbc->charging_time;
		lbci->charger_status = lbc->charger_status;
		g_array_append_val(arrays, lbci);
	}
	ret = pthread_mutex_unlock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	return RESOURCED_ERROR_NONE;
}

/* ============================ DBUS interface ====================== */

static void dbus_get_battery_capacity_history_latest(GDBusMethodInvocation *invocation, GVariant *params)
{
	int i, size, charge, max_size;
	GArray *arrays = NULL;
	GVariantBuilder builder, *sub_builder;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(ii)", &charge, &max_size);
	if (charge < 0 || max_size < 0) {
		_E("Wrong message arguments!");
		goto failure;
	}

	size = g_slist_length(capacity_history_list);
	if (!size) {
		_I("capacity history is empty");
		goto failure;
	}
	arrays = g_array_new(FALSE, FALSE, sizeof(struct heart_battery_capacity *));
	if (!arrays) {
		_E("Failed to alloc array");
		goto failure;
	}
	if (heart_battery_get_capacity_history_latest(arrays, charge, max_size) != RESOURCED_ERROR_NONE) {
		_E("Failed to get capacity history latest");
		D_BUS_REPLY_NULL(invocation);
		return;
	}

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
	sub_builder = g_variant_builder_new(G_VARIANT_TYPE("a(iii)"));
	if (!arrays->len) {
		_E("No battery capacity history data");
		goto exit;
	}

	for (i = 0; i < arrays->len; i++) {
		struct heart_battery_capacity *lbc;
		lbc = g_array_index(arrays, struct heart_battery_capacity *, i);
		if (!lbc)
			break;

		g_variant_builder_add(sub_builder, "(iii)", lbc->capacity,
							lbc->used_time, lbc->charging_time);
		free(lbc);
	}
exit:
	g_variant_builder_add_value(&builder, g_variant_new("a(iii)", sub_builder));
	g_variant_builder_unref(sub_builder);
	g_array_free(arrays, TRUE);
	g_dbus_method_invocation_return_value(invocation, g_variant_builder_end(&builder));
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_get_battery_capacity_history(GDBusMethodInvocation *invocation, GVariant *params)
{
	int ret, size, period = -1, index;
	struct heart_battery_capacity *lbc;
	GSList *iter;
	time_t curr = time(NULL);
	GVariantBuilder builder, *sub_builder;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(i)", &period);
	if (period < 0) {
		_E("Wrong message arguments!");
		goto failure;
	}
	switch (period) {
	case DATA_LATEST:
		index = 0;
		break;
	case DATA_3HOUR:
		index = 3;
		break;
	case DATA_6HOUR:
		index = 6;
		break;
	case DATA_12HOUR:
		index = 12;
		break;
	case DATA_1DAY:
		index = 24;
		break;
	default:
		_E("Wrong message arguments! %d", period);
		goto failure;
	}
	size = g_slist_length(capacity_history_list);
	if (!size) {
		_I("capacity history is empty");
		goto failure;
	}
	ret = pthread_mutex_lock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		goto failure;
	}

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
	sub_builder = g_variant_builder_new(G_VARIANT_TYPE("a(su)"));

	gslist_for_each_item(iter, capacity_history_list) {
		lbc = (struct heart_battery_capacity *)iter->data;
		if (!lbc)
			break;
		if (index && (lbc->timestamp < curr - (index * 3600)))
			continue;

		g_variant_builder_add(sub_builder, "(iii)", lbc->capacity,
							lbc->used_time, lbc->charging_time);
	}

	ret = pthread_mutex_unlock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		g_variant_builder_unref(sub_builder);
		goto failure;
	}

	g_variant_builder_add_value(&builder, g_variant_new("a(iii)", sub_builder));
	g_variant_builder_unref(sub_builder);

	g_dbus_method_invocation_return_value(invocation, g_variant_builder_end(&builder));
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_get_battery_used_time(GDBusMethodInvocation *invocation, GVariant *params)
{
	heart_battery_update_used_time(logging_get_time(CLOCK_BOOTTIME),
			batt_stat.curr_charger_status);

	g_dbus_method_invocation_return_value(invocation,
			g_variant_new("(i)", batt_used.used_time_sec));
}

static int get_battery_remaining_time(int mode, enum charger_status_type status)
{
	int i, ret, count;
	long sum, time, cumul_average, trend_average;
	double result;

	ret = count = 0;
	sum = time = 0;
	cumul_average = trend_average = 0;
	/* get prediction time of cumulative value */
	for (i = 0; i <= WEEK; i++) {
		time = heart_battery_get_prediction_time(i, status);
		if (time) {
			sum += time;
			count++;
		}
	}
	if (count)
		cumul_average = sum / count;

	count = 0;
	sum = 0;
	/* get prediction time of trend value */
	for (i = COUNT; i < MAX_STRATEGY; i++) {
		time = heart_battery_get_prediction_time(i, status);
		if (time) {
			sum += time;
			count++;
		}
	}
	if (count)
		trend_average = sum / count;

	/* failed to get prediction so return learning mode */
	if (!cumul_average && !trend_average) {
		if (batt_stat.curr_capacity != 100 && batt_stat.curr_capacity != 0)
			ret = BATTERY_USAGE_LEARNING;
	} else if (cumul_average && !trend_average) {
		/* failed to get prediction of trend average */
		ret = cumul_average;
	} else if (!cumul_average && trend_average) {
		/* failed to get prediction of cumulative average */
		ret = trend_average;
	} else
		ret = ((cumul_average * CUMUL_WEIGHT) + (trend_average * TREND_WEIGHT));

	if (status == CHARGING) {
		if (logic_v2)  /* new logic */
			return batt_stat.remaining_time_chg;
		return ret; /* old logic */
	}

	if (logic_v2) { /* new logic */
		switch (mode) {
		case POWER_SAVING_MODE:
			/* Fall through */
		case ULTRA_SAVING_MODE:
			ret = get_battery_remaining_time_ups();
			break;
		case POWER_NORMAL_MODE:
			ret = get_battery_remaining_time_new();
			break;
		default:
			break;
		}
		return ret;
	}

	if (ret <= 0)
		return BATTERY_USAGE_LEARNING;

	 /* old logic */
	switch (mode) {
	case ULTRA_SAVING_MODE:
		/* Fall through */
	case POWER_SAVING_MODE:
		result = (double)ret * default_mode_factor[mode];
		return (int)result;
	case POWER_NORMAL_MODE:
		/* Fall through */
	default:
		return ret;
	}
}

static void dbus_get_battery_remaining_time(GDBusMethodInvocation *invocation, GVariant *params)
{
	int ret, mode = -1;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(i)", &mode);
	if (mode < 0) {
		_E("Wrong message arguments!");
		goto failure;
	}

	if (!battery_learning_mode)
		battery_learning_mode = heart_battery_get_learning_mode();

	if (!battery_learning_mode) {
		_E("data is not enough to calculate prediction");
		ret = BATTERY_USAGE_LEARNING;
	} else
		ret = get_battery_remaining_time(mode, DISCHARGING);

	_I("Remaining_time %d (mode: %d)", ret, mode);
	g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", ret));
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_get_battery_charging_time(GDBusMethodInvocation *invocation, GVariant *params)
{
	int ret = get_battery_remaining_time(POWER_NORMAL_MODE, CHARGING);
	_I("Remaining_charging_time %d", ret);
	g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", ret));
}

static void dbus_get_battery_discharge_rate_level(GDBusMethodInvocation *invocation, GVariant *params)
{
	int ret = batt_stat.discharge_rate_level;
	_I("discharge_rate_level %d", ret);
	g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", ret));
}

static void dbus_battery_save_to_file(GDBusMethodInvocation *invocation, GVariant *params)
{
	heart_battery_status_save_to_db();

	int ret = heart_battery_capacity_save_to_file(HEART_BATTERY_CAPACITY_DATA_FILE);
	if (ret) {
		_E("save to file failed");
		D_BUS_REPLY_NULL(invocation)
		return;
	}

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", ret));
}

static const char dbus_methods_xml[] =
"		<method name='GetBatteryCapacityHistory'>"
"			<arg type='i' name='Period' direction='in'/>"
"			<arg type='a(iii)' name='CapacityHistory' direction='out'/>"
"		</method>"
"		<method name='GetBatteryCapacityHistoryLatest'>"
"			<arg type='i' name='Charge' direction='in'/>"
"			<arg type='i' name='MaxSize' direction='in'/>"
"			<arg type='a(iii)' name='CapacityHistory' direction='out'/>"
"		</method>"
"		<method name='GetBatteryUsedTime'>"
"			<arg type='i' name='UsedTime' direction='out'/>"
"		</method>"
"		<method name='GetBatteryRemainingTime'>"
"			<arg type='i' name='Mode' direction='in'/>"
"			<arg type='i' name='RemainingTime' direction='out'/>"
"		</method>"
"		<method name='GetBatteryChargingTime'>"
"			<arg type='i' name='ChargingTime' direction='out'/>"
"		</method>"
"		<method name='GetBatteryDischargeRateLevel'>"
"			<arg type='i' name='DischargeRateLevel' direction='out'/>"
"		</method>"
"		<method name='SaveBatteryData'>"
"			<arg type='i' name='ZeroOnSuccess' direction='out'/>"
"		</method>";

static struct d_bus_method dbus_methods[] = {
	{ "GetBatteryCapacityHistory",       dbus_get_battery_capacity_history },
	{ "GetBatteryCapacityHistoryLatest", dbus_get_battery_capacity_history_latest },
	{ "GetBatteryUsedTime",              dbus_get_battery_used_time },
	{ "GetBatteryRemainingTime",         dbus_get_battery_remaining_time },
	{ "GetBatteryChargingTime",          dbus_get_battery_charging_time },
	{ "GetBatteryDischargeRateLevel",    dbus_get_battery_discharge_rate_level },
	{ "SaveBatteryData",                 dbus_battery_save_to_file },
};

/* =========================  DBUS interface END ==================== */
static void heart_battery_used_time_init(enum charger_status_type status)
{
	batt_used.last_charger_status = status;
	batt_used.last_update_time = logging_get_time(CLOCK_BOOTTIME);
}

static void heart_battery_status_init(void)
{
	int i, ret, status, capacity;

	batt_stat.curr_capacity = 0;
	batt_stat.curr_run_time_sec[DISCHARGING] = 0;
	batt_stat.curr_run_time_sec[CHARGING] = 0;
	batt_stat.curr_cap_counter[DISCHARGING] = 0;
	batt_stat.curr_cap_counter[CHARGING] = 0;
	batt_stat.curr_charger_status = 0;
	batt_stat.reset_mark = 0;

	for (i = 0; i < BATTERY_HISTORY_RESET_MAX; i++) {
		heart_battery_set_usage_reset_stime(i, 0);
		heart_battery_set_usage_reset(i, DISCHARGING, 0, 0);
		heart_battery_set_usage_reset(i, CHARGING, 0, 0);
	}


	for (i = 0; i < BATTERY_LEVEL_MAX; i++) {
		heart_battery_set_usage_level_stime(i, 0);
		heart_battery_set_usage_level(i, DISCHARGING, default_sec_per_cap[DISCHARGING][DEFAULT_AVG], 0);
		heart_battery_set_usage_level(i, CHARGING, default_sec_per_cap[CHARGING][DEFAULT_AVG], 0);
	}

	for (i = 0; i < BATTERY_HISTORY_DAY_MAX; i++) {
		heart_battery_set_usage_week_stime(i, 0);
		heart_battery_set_usage_week(i, DISCHARGING, 0, 0);
		heart_battery_set_usage_week(i, CHARGING, 0, 0);
	}

	for (i = 0; i < MAX_STRATEGY; i++) {
		heart_battery_set_prediction(i, DISCHARGING, 0, 0, 0);
		heart_battery_set_prediction(i, CHARGING, 0, 0, 0);
	}

	ret = heart_battery_status_read_from_db();
	if (ret < 0)
		_E("Failed to read battery status data");

	if (logic_v2) {
		for (i = 0; i < INDEX_WINDOW_SIZE; i++) {
			batt_stat.last_wall_time[i] = 0;
			batt_stat.last_volt_intg[i] = 0.0;
		}
		for (i = 0; i < BATTERY_LEVEL_GAP; i++)
			batt_stat.last_wall_time_chg_diff[i] = 0;

		batt_stat.data_available = 0;
		batt_stat.last_capacity = 0;
		batt_stat.last_capacity_chg = 0;
		first_level_change = FALSE;
		batt_stat.curr_index = 0;
		batt_stat.remaining_time = 0;
		batt_stat.remaining_time_ups = 0;
		batt_stat.last_pwr_bchg = average_pwr;
		batt_stat.last_wall_time_chg = 0;
		batt_stat.index_chg = 0;
		batt_stat.remaining_time_chg = 0;
		batt_stat.last_clock_tick =  logging_get_time(CLOCK_BOOTTIME);

		/* During the device boot-up, calculate energy for 100% battery and calcuate
		pivot(norma and UPS mode) for time remaining time estimation*/
		heart_battery_calculate_pivot();
		/*get all the previousely stored data from file to program variable*/
		heart_battery_read_data_from_file();
		/*
		* check whether was device power off for long time. If yes then adjust
		* our stored time values
		*/
		heart_battery_power_off_time_adjustment();
		/*
		* register callback for any time change from user side.
		* If any time change happen then adjust whole data as per new time
		*/
		vconf_notify_key_changed(VCONFKEY_SETAPPL_TIMEZONE_INT, time_change_notify_cb, NULL);
		vconf_notify_key_changed(VCONFKEY_SYSTEM_TIME_CHANGED, time_change_notify_cb, NULL);
		vconf_notify_key_changed(VCONFKEY_REGIONFORMAT, time_change_notify_cb, NULL);
	}

	battery_learning_mode = heart_battery_get_learning_mode();

	ret = heart_battery_capacity_read_from_file(HEART_BATTERY_CAPACITY_DATA_FILE);
	if (ret < 0)
		_E("Failed to read battery capacity data");

	capacity = heart_battery_direct_get_capacity();
	if (capacity > 0)
		batt_stat.curr_capacity = capacity;
	status = heart_battery_direct_get_charger_status();
	if (status >= 0)
		batt_stat.curr_charger_status = status;
	heart_battery_used_time_init(batt_stat.curr_charger_status);
	heart_battery_calculate_prediction(batt_stat.curr_charger_status);
	batt_stat.last_event_wall_time = logging_get_time(CLOCK_BOOTTIME);
	batt_stat.discharge_rate_level = BATTERY_DISCHARGE_NONE;

	if (logic_v2) {
		heart_battery_cal_discharge_rem_time();

		if (batt_stat.curr_charger_status == CHARGING) {
			heart_battery_cal_charging_rem_time();
			first_level_change = TRUE;
		}
	}
}

static int low_battery_handler(void *data)
{
	heart_battery_save_to_file(false);
	return RESOURCED_ERROR_NONE;
}

static int heart_battery_config(struct parse_result *result, void *user_data)
{
	int val;

	if (!result)
		return -EINVAL;

	if (!strncmp(result->section, HEART_BATTERY_CONF_SECTION, sizeof(HEART_BATTERY_CONF_SECTION) + 1)) {
		if (!strncmp(result->name, "POWER_NORMAL_MODE", sizeof("POWER_NORMAL_MODE") + 1)) {
			val = atoi(result->value);
			if (val > 0)
				default_mode_spc[POWER_NORMAL_MODE] = val;
			_D("POWER_NORMAL_MODE SPC: %d", val);
		} else if (!strncmp(result->name, "POWER_SAVING_MODE", sizeof("POWER_SAVING_MODE") + 1)) {
			val = atoi(result->value);
			if (val > 0)
				default_mode_spc[POWER_SAVING_MODE] = val;
			_D("POWER_SAVING_MODE SPC: %d", val);
		} else if (!strncmp(result->name, "ULTRA_SAVING_MODE", sizeof("ULTRA_SAVING_MODE") + 1)) {
			val = atoi(result->value);
			if (val > 0)
				default_mode_spc[ULTRA_SAVING_MODE] = val;
			_D("ULTRA_POWER_SAVING_MODE SPC: %d", val);
		}
	} else if (!strncmp(result->section, battery_header, sizeof(battery_header) + 1)) {
		if (!strncmp(result->name, "LOGIC_V2", sizeof("LOGIC_V2") + 1)) {
			logic_v2 = atoi(result->value);
			_I("logic_v2 : %d", logic_v2);
		} else if (!strncmp(result->name, "DISCHARGE_FAST", sizeof("DISCHARGE_FAST") + 1)) {
			discharge_fast = atoi(result->value);
			_I("discharge_fast: %d", discharge_fast);
		} else if (!strncmp(result->name, "DISCHARGE_SLOW", sizeof("DISCHARGE_SLOW") + 1)) {
			discharge_slow = atoi(result->value);
			_I("discharge_slow: %d", discharge_slow);
		} else if (!strncmp(result->name, "AVERAGE_PWR", sizeof("AVERAGE_PWR") + 1)) {
			average_pwr = atof(result->value);
			_I("average_power: %lf", average_pwr);
		} else if (!strncmp(result->name, "OCV_SOC_POLY_COEF", sizeof("OCV_SOC_POLY_COEF") + 1)) {
			char *token, *saveptr = NULL;
			int i;
			token = strtok_r(result->value, " ", &saveptr);
			if (token) {
				ocv_degree = atoi(token);
				if (ocv_degree <= 0) {
					_E("ocv_degree should be positive if you decide to use logic_v2");
					return -EINVAL;
				}
				intg = calloc(ocv_degree, sizeof(double));
				if (!intg) {
					_E("Out of memory");
					return -ENOMEM;
				}
				_I("Degree of OCV_SOC_POLY = %d", ocv_degree);

				for (i = 0; i < ocv_degree; i++) {
					token = strtok_r(NULL, " ", &saveptr);
					if (!token) {
						_E("There is no OCV_SOC_POLY_COEF_%d", i + 1);
						return -EINVAL;
					}
					intg[i] = atof(token);
					_I("OCV_SOC_POLY_COEF_%d = %.9g", i + 1, intg[i]);
				}
			}
		}
	}
	return RESOURCED_ERROR_NONE;
}

static void heart_battery_mode_factor_init(void)
{
	double val;

	val = default_mode_spc[POWER_SAVING_MODE]/default_mode_spc[POWER_NORMAL_MODE];

	if (1.0 < val)
		default_mode_factor[POWER_SAVING_MODE] = val;
	_I("POWER_SAVING_MODE factor: %f", val);

	val = default_mode_spc[ULTRA_SAVING_MODE]/default_mode_spc[POWER_NORMAL_MODE];

	if (1.0 < val)
		default_mode_factor[ULTRA_SAVING_MODE] = val;
	_I("ULTRA_POWER_SAVING_MODE factor: %f", val);
}

/*
 * This function will read total battery capacity from sys interface
 * example - 1500mAh or 2600mAh
 */
static void heart_read_battery_total_capacity(void)
{
	u_int32_t batt_capacity;

	/*
	 * read the battery total capacity from sys interface path
	 * "/sys/class/power_supply/battery/batt_capacity"
	 * which is provided by kernel
	 */
	if (fread_uint("/sys/class/power_supply/battery/batt_capacity",
					&batt_capacity) != RESOURCED_ERROR_NONE) {
		_D("This device doesn't support battery capacity information. Disable LOGIC_V2");
		return;
	}

	snprintf(battery_header, sizeof(battery_header), "BATTERY_%d",
			batt_capacity);

	/*
	 * convert battery capacity from mAh to mAm
	 */
	total_battery_capacity = batt_capacity * 60;
	_I("battery capacity = %d, battery header = %s",
			batt_capacity, battery_header);
}

static int heart_battery_init(void *data)
{
	int ret;

	ret = logging_module_init(BATTERY_NAME, ONE_DAY, TEN_MINUTE, heart_battery_update,
							HEART_BATTERY_UPDATE_INTERVAL, SYSTEM_DEFAULT);
	if (ret != RESOURCED_ERROR_NONE) {
		_E("logging module init failed");
		return RESOURCED_ERROR_FAIL;
	}

	heart_dbus_declare_methods(dbus_methods_xml, dbus_methods, ARRAY_SIZE(dbus_methods));

	ret = d_bus_register_signal(DEVICED_PATH_BATTERY,
			DEVICED_INTERFACE_BATTERY, GET_BATTERY_CAPACITY,
			heart_battery_capacity_status, NULL);
	if (ret < 0)
		_E("Failed to add a capacity status signal handler");

	ret = d_bus_register_signal(DEVICED_PATH_BATTERY,
			DEVICED_INTERFACE_BATTERY, GET_CHARGER_STATUS,
			heart_battery_charger_status, NULL);
	if (ret < 0)
		_E("Failed to add a charger status signal handler");

	heart_read_battery_total_capacity();

	config_parse(HEART_CONF_FILE_PATH, heart_battery_config, NULL);

	heart_battery_mode_factor_init();

	heart_battery_status_init();

	register_notifier(RESOURCED_NOTIFIER_LOW_BATTERY, low_battery_handler);
	register_notifier(RESOURCED_NOTIFIER_DATA_RESET, heart_battery_reset);

	heart_battery_set_file_commit_timestamp(logging_get_time(CLOCK_BOOTTIME));
	_D("heart battery init finished");
	return RESOURCED_ERROR_NONE;
}

void heart_capacity_history_update(struct logging_table_form *data, void *user_data)
{
	int status, reset, capacity, diff;
	unsigned long discharging = 0, charging = 0;
	GSList **history_list = NULL;

	if (user_data)
		history_list = (GSList **)user_data;
	else
		history_list = &capacity_history_list;

	_D("%s %s %ld %s", data->appid, data->pkgid, data->time, data->data);
	if (sscanf(data->data, "%d %d %ld %ld %d %d ",
				&capacity, &diff,
				&discharging, &charging,
				&status, &reset) < 0) {
		_E("sscanf failed");
		return;
	}
	heart_battery_insert_capacity(history_list, capacity,
			diff, data->time, discharging, charging, status,
			reset, false);
}

static int heart_battery_dump(FILE *fp, int mode, void *data)
{
	struct heart_battery_capacity *lbc;
	GSList *iter;
	char buf[BATTERY_DATA_MAX] = {0, };
	int ret, size, len = 0;
	time_t starttime;
	char timestr[80];
	struct tm loc_tm;
	GSList *history_list = NULL;

	starttime = time(NULL);
	starttime -= mode;
	localtime_r(&starttime, &loc_tm);
	/* print timestamp */
	strftime(timestr, sizeof(timestr),
			"%Y-%m-%d %H:%M:%S%z", &loc_tm);

	logging_read_foreach(BATTERY_NAME, NULL, NULL, starttime, 0,
			heart_capacity_history_update, &history_list);

	if (!history_list) {
		_E("capacity history is NULL!");
		return RESOURCED_ERROR_NONE;
	}
	LOG_DUMP(fp, "[BATTERY CAPACITY HISTORY] since %s\n", timestr);
	LOG_DUMP(fp, "capacity diff timestamp used_time charging_time charger_status, reset_mark\n");
	ret = pthread_mutex_lock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	size = g_slist_length(history_list);
	if (!size) {
		_I("capacity history is empty");
		ret = pthread_mutex_unlock(&heart_battery_mutex);
		if (ret) {
			_E("pthread_mutex_unlock() failed, %d", ret);
			return RESOURCED_ERROR_FAIL;
		}
		return RESOURCED_ERROR_NONE;
	}
	gslist_for_each_item(iter, history_list) {
		lbc = (struct heart_battery_capacity *)iter->data;
		if (!lbc)
			break;
		len += snprintf(buf + len, BATTERY_DATA_MAX - len, "%d %d %ld %ld %ld %d %d\n",
				lbc->capacity, lbc->diff_capacity, lbc->timestamp, lbc->used_time,
				lbc->charging_time, lbc->charger_status,
				lbc->reset_mark);
		if (BATTERY_DATA_MAX < len + BATTERY_LINE_MAX) {
			LOG_DUMP(fp, "%s\n", buf);
			len = 0;
		}
	}
	LOG_DUMP(fp, "%s\n", buf);
	ret = pthread_mutex_unlock(&heart_battery_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	fflush(fp);
	if (history_list) {
		g_slist_free_full(history_list, free);
		history_list = NULL;
	}

	return RESOURCED_ERROR_NONE;
}

static int heart_battery_exit(void *data)
{
	int ret;
	GSList *iter, *next;
	struct heart_battery_capacity *lbc;

	heart_battery_save_to_file(true);
	ret = pthread_mutex_lock(&heart_battery_mutex);
	if (ret)
		_E("pthread_mutex_lock() failed, %d", ret);

	gslist_for_each_safe(capacity_history_list, iter, next, lbc) {
		capacity_history_list = g_slist_remove(capacity_history_list, lbc);
		free(lbc);
	}
	capacity_history_list = NULL;

	ret = pthread_mutex_unlock(&heart_battery_mutex);
	if (ret)
		_E("pthread_mutex_unlock() failed, %d", ret);

	unregister_notifier(RESOURCED_NOTIFIER_LOW_BATTERY, low_battery_handler);
	unregister_notifier(RESOURCED_NOTIFIER_DATA_RESET, heart_battery_reset);

	logging_module_exit();

	_D("heart battery exit");
	return RESOURCED_ERROR_NONE;
}

static const struct heart_module_ops heart_battery_ops = {
	.name           = "BATTERY",
	.init           = heart_battery_init,
	.dump           = heart_battery_dump,
	.exit           = heart_battery_exit,
};
HEART_MODULE_REGISTER(&heart_battery_ops)
