/*
 * resourced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * @file heart-memory.c
 *
 * @desc heart memory module
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <glib.h>

#include "resourced.h"
#include "trace.h"
#include "module.h"
#include "macro.h"
#include "notifier.h"
#include "proc-common.h"
#include "heart.h"
#include "logging.h"
#include "heart-common.h"
#include "dbus-handler.h"
#include "procfs.h"
#include "util.h"

#include <sqlite3.h>
#include <time.h>

struct heart_memory_info {
	unsigned int max_rss;
	unsigned int avg_rss;
};

struct heart_memory_table {
	char appid[MAX_APPID_LENGTH];
	char pkgid[MAX_PKGNAME_LENGTH];
	unsigned int total_rss;
	unsigned int max_rss;
	unsigned int latest_rss;
	unsigned int count;
	int renew;
	GArray *memory_info;
};

static GHashTable *heart_memory_app_list;
static pthread_mutex_t heart_memory_mutex = PTHREAD_MUTEX_INITIALIZER;
static time_t last_file_update_time;
static GSource *heart_memory_update_timer = NULL;

struct heart_memory_table *heart_memory_find_info(GHashTable *hashtable, char *appid)
{
	int ret;
	struct heart_memory_table *table;

	if (!hashtable) {
		_E("hashtable is NULL");
		return NULL;
	}

	ret = pthread_mutex_lock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return NULL;
	}

	table = g_hash_table_lookup(hashtable, (gconstpointer)appid);

	ret = pthread_mutex_unlock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return NULL;
	}

	return table;
}

static void heart_memory_free_value(gpointer value)
{
	struct heart_memory_table * table = (struct heart_memory_table *)value;

	if (table)
		free(table);
}

static int heart_memory_read_length(char *buf, int count)
{
	int i, find = 0;
	int len = strlen(buf);

	for (i = 0; i < len; i++) {
		if (buf[i] == ' ')
			find++;

		if (find == count)
			return i+1;
	}

	return RESOURCED_ERROR_FAIL;
}

static int heart_memory_read_from_file(GHashTable *hashtable, char *filename)
{
	int i, len, ret, result;
	unsigned int total_rss, max_rss, avg_rss;
	unsigned int count;
	FILE *fp;
	struct heart_memory_table *table;
	char appid[MAX_APPID_LENGTH] = {0, };
	char pkgid[MAX_PKGNAME_LENGTH] = {0, };
	char buf[MEM_DATA_MAX] = {0, };

	fp = fopen(filename, "r");

	if (!fp) {
		_E("%s fopen failed %d", filename, errno);
		return RESOURCED_ERROR_FAIL;
	}

	while (fgets(buf, MEM_DATA_MAX, fp)) {
		table = malloc(sizeof(struct heart_memory_table));

		if (!table) {
			_E("malloc failed");
			fclose(fp);
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}

		/* make return values */
		result = sscanf(buf, STR_FS(MAX_APPID_LENGTH_M1)" "STR_FS(MAX_PKGNAME_LENGTH_M1)" %u %u %u ",
				appid, pkgid, &total_rss, &max_rss, &count);

		if (result < 0) {
			_E("sscanf failed");
			free(table);
			fclose(fp);
			return RESOURCED_ERROR_FAIL;
		}

		if (snprintf(table->appid, MAX_APPID_LENGTH, "%s", appid) < 0) {
			_E("snprintf failed");
			free(table);
			fclose(fp);
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}

		if (snprintf(table->pkgid, MAX_PKGNAME_LENGTH, "%s", pkgid) < 0) {
			_E("snprintf failed");
			free(table);
			fclose(fp);
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}

		len = heart_memory_read_length(buf, 7);

		if (len < 0) {
			_E("7 space read length failed %s", buf);
			free(table);
			fclose(fp);
			return RESOURCED_ERROR_FAIL;
		}

		table->total_rss = total_rss;
		table->max_rss = max_rss;
		table->latest_rss = 0;
		table->count = count;
		table->memory_info =
			g_array_new(false, false, sizeof(struct heart_memory_info *));

		for (i = 0; i < MEM_ARRAY_MAX; i++) {
			struct heart_memory_info *mi;

			result = sscanf(buf + len, "%u %u ",
					&max_rss, &avg_rss);

			if (result <= 0) {
				_E("file read fail %s", buf + len);
				g_array_free(table->memory_info, true);
				free(table);
				fclose(fp);
				return RESOURCED_ERROR_FAIL;
			}

			mi = malloc(sizeof(struct heart_memory_info));

			if (!mi) {
				_E("malloc failed");
				g_array_free(table->memory_info, true);
				free(table);
				fclose(fp);
				return RESOURCED_ERROR_OUT_OF_MEMORY;
			}

			mi->max_rss = max_rss;
			mi->avg_rss = avg_rss;

			len += heart_memory_read_length(buf + len, 4);

			g_array_append_val(table->memory_info, mi);
		}

		ret = pthread_mutex_lock(&heart_memory_mutex);
		if (ret) {
			_E("pthread_mutex_lock() failed, %d", ret);
			g_array_free(table->memory_info, true);
			free(table);
			fclose(fp);
			return RESOURCED_ERROR_FAIL;
		}

		g_hash_table_insert(hashtable, (gpointer)table->appid, (gpointer)table);

		ret = pthread_mutex_unlock(&heart_memory_mutex);
		if (ret) {
			_E("pthread_mutex_unlock() failed, %d", ret);
			fclose(fp);
			return RESOURCED_ERROR_FAIL;
		}
	}

	fclose(fp);

	return RESOURCED_ERROR_NONE;
}

static int heart_memory_save_to_file(GHashTable *hashtable, char *filename)
{
	int i, len, ret, array_len;
	gpointer value;
	gpointer key;
	GHashTableIter iter;
	struct heart_memory_table *table;
	FILE *fp;
	char buf[MEM_DATA_MAX] = {0, };

	fp = fopen(filename, "w");
	if (!fp) {
		_E("%s fopen failed %d", filename, errno);
		return RESOURCED_ERROR_FAIL;
	}

	ret = pthread_mutex_lock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}

	g_hash_table_iter_init(&iter, hashtable);

	while (g_hash_table_iter_next(&iter, &key, &value)) {
		table = (struct heart_memory_table *)value;

		len = snprintf(buf, MEM_DATA_MAX, "%s %s %u %u %u ",
				table->appid, table->pkgid,
				table->total_rss, table->max_rss,
				table->count);

		array_len = table->memory_info->len;

		for (i = 0; i < MEM_ARRAY_MAX; i++) {
			struct heart_memory_info *mi;

			if (array_len <= i)
				len += snprintf(buf + len, MEM_DATA_MAX - len, "0 0 0 0 ");
			else {
				mi = g_array_index(table->memory_info, struct heart_memory_info *, i);

				len += snprintf(buf + len, MEM_DATA_MAX - len, "%u %u ",
						mi->max_rss, mi->avg_rss);
			}
		}

		snprintf(buf + len, MEM_DATA_MAX - len, "%c\n", MEM_FILE_SEPERATOR);

		fputs(buf, fp);
	}

	ret = pthread_mutex_unlock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		fclose(fp);
		return RESOURCED_ERROR_FAIL;
	}

	fclose(fp);

	return RESOURCED_ERROR_NONE;
}

void heart_memory_fill_array(struct logging_table_form *data, void *user_data)
{
	int i;
	unsigned int rss;
	struct heart_memory_data *md = NULL;
	GArray *send = NULL;

	_E("data : %s %s %d %s", data->appid, data->pkgid, (int)data->time, data->data);
	send = (GArray *)user_data;

	for (i = 0; i < send->len; i++) {
		struct heart_memory_data *loop;
		loop = g_array_index(send, struct heart_memory_data *, i);

		if (!strncmp(loop->appid, data->appid, strlen(data->appid)+1)) {
			md = loop;
			break;
		}
	}

	if (sscanf(data->data, "%u", &rss) < 0) {
		_E("sscanf failed");
		return;
	}

	if (!md) {
		md = malloc(sizeof(struct heart_memory_data));

		if (!md) {
			_E("malloc failed");
			return;
		}

		if (snprintf(md->appid, MAX_APPID_LENGTH, "%s", data->appid) < 0) {
			_E("snprintf failed");
			free(md);
			return;
		}

		if (snprintf(md->pkgid, MAX_PKGNAME_LENGTH, "%s", data->pkgid) < 0) {
			_E("snprintf failed");
			free(md);
			return;
		}

		md->max_rss = rss;
		md->avg_rss = rss;

		g_array_append_val(send, md);
	} else {
		if (md->max_rss < rss)
			md->max_rss = rss;

		md->avg_rss = (md->avg_rss + rss) / 2;
	}
}

int logging_memory_get_foreach(GArray *arrays, enum heart_data_period period)
{
	int ret;
	time_t curr_time = time(NULL);

	switch (period) {
	case DATA_LATEST:
		break;
	case DATA_3HOUR:
		curr_time -= 10800;
		break;
	case DATA_6HOUR:
		curr_time -= 21600;
		break;
	case DATA_12HOUR:
		curr_time -= 43200;
		break;
	case DATA_1DAY:
		curr_time -= 86400;
		break;
	default:
		_E("Wrong message arguments! %d", period);
		return RESOURCED_ERROR_FAIL;
	}

	ret = logging_read_foreach(MEM_NAME, NULL, NULL, curr_time, 0, heart_memory_fill_array, arrays);

	if (ret) {
		_E("failed logging_read_foreach");
		return RESOURCED_ERROR_FAIL;
	}

	return RESOURCED_ERROR_NONE;
}

int heart_memory_get_query(GArray *arrays, enum heart_data_period period)
{
	int count, result;
	time_t curr_time = time(NULL);
	sqlite3 *heart_db = NULL;
	sqlite3_stmt *stmt = NULL;
	gpointer value;
	gpointer key;
	char *data;
	unsigned int rss;
	GHashTableIter h_iter;
	struct heart_memory_data *md;
	struct heart_memory_table *table;
	char buf[MEM_DATA_MAX] = {0, };

	switch (period) {
	case DATA_LATEST:
		break;
	case DATA_3HOUR:
		curr_time -= 10800;
		break;
	case DATA_6HOUR:
		curr_time -= 21600;
		break;
	case DATA_12HOUR:
		curr_time -= 43200;
		break;
	case DATA_1DAY:
		curr_time -= 86400;
		break;
	default:
		_E("Wrong message arguments! %d", period);
		return RESOURCED_ERROR_FAIL;
	}

	if (logging_get_default_db(MEM_NAME, &heart_db) != RESOURCED_ERROR_NONE) {
		_E("Fail to get DB path");
		return RESOURCED_ERROR_FAIL;
	}

	g_hash_table_iter_init(&h_iter, heart_memory_app_list);

	while (g_hash_table_iter_next(&h_iter, &key, &value)) {
		table = (struct heart_memory_table *)value;

		md = malloc(sizeof(struct heart_memory_data));

		if (!md) {
			_E("malloc failed");
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}

		if (snprintf(md->appid, MAX_APPID_LENGTH, "%s", table->appid) < 0) {
			_E("snprintf failed");
			free(md);
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}

		if (snprintf(md->pkgid, MAX_PKGNAME_LENGTH,  "%s", table->pkgid) < 0) {
			_E("asprintf failed");
			free(md);
			return RESOURCED_ERROR_OUT_OF_MEMORY;
		}

		count = 0;
		md->max_rss = 0;
		md->avg_rss = 0;

		snprintf(buf, MEM_DATA_MAX, "select * from memory where appid = \'%s\' AND time > %d",
				table->appid, (int)curr_time);

		/* search from db */
		if (sqlite3_prepare_v2(heart_db, buf, -1, &stmt, NULL) != SQLITE_OK) {
			_E("select failed");
			free(md);
			sqlite3_finalize(stmt);
			return RESOURCED_ERROR_DB_FAILED;
		}

		do {
			result = sqlite3_step(stmt);
			switch (result) {
			case SQLITE_ROW:
				data = (char *)sqlite3_column_text(stmt, 3);

				if (sscanf(data, "%u", &rss) < 0) {
					_E("sscanf failed");
					free(md);
					sqlite3_finalize(stmt);
					return RESOURCED_ERROR_DB_FAILED;
				}

				if (md->max_rss < rss)
					md->max_rss = rss;

				md->avg_rss += rss;
				count++;
				break;
			case SQLITE_DONE:
				break;
			case SQLITE_ERROR:
				_E("select %s table failed %s",
						MEM_NAME, sqlite3_errmsg(heart_db));
				/* FALLTHROUGH */
			default:
				free(md);
				sqlite3_finalize(stmt);
				return RESOURCED_ERROR_DB_FAILED;
			}
		} while (result == SQLITE_ROW);
		if (count)
			md->avg_rss /= count;

		g_array_append_val(arrays, md);
	}

	sqlite3_finalize(stmt);

	return RESOURCED_ERROR_NONE;
}

struct heart_memory_data *heart_memory_get_data(char *appid, enum heart_data_period period)
{
	int index, i, ret, count, time;
	unsigned int max_rss = 0, avg_rss = 0, total_rss = 0;
	struct heart_memory_table *table;
	struct heart_memory_data *md;

	if (!heart_memory_app_list)
		return NULL;

	if (!appid) {
		_E("Wrong message arguments!");
		return NULL;
	}

	switch (period) {
	case DATA_LATEST:
		index = 0;
		break;
	case DATA_3HOUR:
		index = 3;
		break;
	case DATA_6HOUR:
		index = 6;
		break;
	case DATA_12HOUR:
		index = 12;
		break;
	case DATA_1DAY:
		index = 24;
		break;
	default:
		_E("Wrong message arguments! %d", period);
		return NULL;
	}

	/* read from hash and make reply hash */
	/* loop in file read and make reply */
	ret = pthread_mutex_lock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return NULL;
	}

	table = g_hash_table_lookup(heart_memory_app_list, (gconstpointer)appid);
	if (!table)
		goto unlock_exit;
	md = malloc(sizeof(struct heart_memory_data));
	if (!md) {
		_E("malloc failed");
		goto unlock_exit;
	}

	if (snprintf(md->appid, MAX_APPID_LENGTH, "%s", table->appid) < 0) {
		_E("snprintf failed");
		free(md);
		goto unlock_exit;
	}

	if (snprintf(md->pkgid, MAX_PKGNAME_LENGTH, "%s", table->pkgid) < 0) {
		_E("snprintf failed");
		free(md);
		goto unlock_exit;
	}

	count = table->count;
	total_rss = table->total_rss;
	max_rss = table->max_rss;
	avg_rss = total_rss / count;
	time = 1;

	if (period != DATA_LATEST) {
		i = table->memory_info->len;

		if (i < index)
			index = i;

		for (i = 0; i < index; i++) {
			struct heart_memory_info *mi;

			mi = g_array_index(table->memory_info, struct heart_memory_info *, i);

			if (mi->max_rss || mi->avg_rss)
				time++;

			if (max_rss < mi->max_rss)
				max_rss = mi->max_rss;

			avg_rss += mi->avg_rss;
		}

		if (time)
			avg_rss /= time;

	}

	md->max_rss = max_rss;
	md->avg_rss = avg_rss;

	ret = pthread_mutex_unlock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		free(md);
		return NULL;
	}
	return md;

unlock_exit:
	ret = pthread_mutex_unlock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return NULL;
	}
	return NULL;
}

int heart_memory_get_latest_data(char *appid, unsigned int *rss)
{
	int ret;
	char *data;
	struct heart_memory_table *table;

	if (!appid) {
		_E("Wrong message arguments!");
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	if (!heart_memory_app_list) {
		_E("hashtable heart_memory_app_list is NULL");
		return RESOURCED_ERROR_FAIL;
	}

	/* read from hash and make reply hash */
	/* loop in file read and make reply */
	ret = pthread_mutex_lock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}

	table = g_hash_table_lookup(heart_memory_app_list, (gconstpointer)appid);
	if (!table) {
		_E("NOT found in table %s", appid);

		ret = logging_get_latest_in_cache(MEM_NAME, appid, &data);

		if (ret) {
			_E("logging_get_latest_in_cache failed");

			ret = pthread_mutex_unlock(&heart_memory_mutex);
			if (ret) {
				_E("pthread_mutex_unlock() failed, %d", ret);
				return RESOURCED_ERROR_FAIL;
			}
			return RESOURCED_ERROR_FAIL;
		}

		if (sscanf(data, "%u", rss) < 0) {
			_E("sscanf failed");
			ret = pthread_mutex_unlock(&heart_memory_mutex);
			if (ret) {
				_E("pthread_mutex_unlock() failed, %d", ret);
				return RESOURCED_ERROR_FAIL;
			}
			return RESOURCED_ERROR_FAIL;
		}

		ret = pthread_mutex_unlock(&heart_memory_mutex);
		if (ret) {
			_E("pthread_mutex_unlock() failed, %d", ret);
			return RESOURCED_ERROR_FAIL;
		}
		return RESOURCED_ERROR_NONE;
	}

	if (table->latest_rss) {
		*rss = table->latest_rss;
	} else {
		ret = logging_get_latest_in_cache(MEM_NAME, appid, &data);

		if (ret) {
			_E("logging_get_latest_in_cache failed");
			ret = pthread_mutex_unlock(&heart_memory_mutex);
			if (ret) {
				_E("pthread_mutex_unlock() failed, %d", ret);
				return RESOURCED_ERROR_FAIL;
			}
			return RESOURCED_ERROR_FAIL;
		}

		if (sscanf(data, "%u", rss) < 0) {
			_E("sscanf failed");
			ret = pthread_mutex_unlock(&heart_memory_mutex);
			if (ret) {
				_E("pthread_mutex_unlock() failed, %d", ret);
				return RESOURCED_ERROR_FAIL;
			}
			return RESOURCED_ERROR_FAIL;
		}
	}

	ret = pthread_mutex_unlock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}

	return RESOURCED_ERROR_NONE;
}

int heart_memory_get_table(GArray *arrays, enum heart_data_period period)
{
	int index, i, ret, count, time;
	gpointer value;
	gpointer key;
	unsigned int max_rss = 0, avg_rss = 0, total_rss = 0;
	GHashTableIter h_iter;
	struct heart_memory_table *table;
	struct heart_memory_data *md;

	switch (period) {
	case DATA_LATEST:
		index = 0;
		break;
	case DATA_3HOUR:
		index = 3;
		break;
	case DATA_6HOUR:
		index = 6;
		break;
	case DATA_12HOUR:
		index = 12;
		break;
	case DATA_1DAY:
		index = 24;
		break;
	default:
		_E("Wrong message arguments! %d", period);
		return RESOURCED_ERROR_FAIL;
	}

	/* read from hash and make reply hash */
	/* loop in file read and make reply */
	ret = pthread_mutex_lock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_lock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}

	g_hash_table_iter_init(&h_iter, heart_memory_app_list);

	while (g_hash_table_iter_next(&h_iter, &key, &value)) {
		table = (struct heart_memory_table *)value;

		md = malloc(sizeof(struct heart_memory_data));
		if (!md) {
			_E("malloc failed");
			goto unlock_out_of_memory_exit;
		}

		if (snprintf(md->appid, MAX_APPID_LENGTH, "%s", table->appid) < 0) {
			_E("snprintf failed");
			free(md);
			goto unlock_out_of_memory_exit;
		}

		if (snprintf(md->pkgid, MAX_PKGNAME_LENGTH, "%s", table->pkgid) < 0) {
			_E("snprintf failed");
			free(md);
			goto unlock_out_of_memory_exit;
		}

		count = table->count;
		total_rss = table->total_rss;
		max_rss = table->max_rss;
		avg_rss = total_rss / count;
		time = 1;

		if (period != DATA_LATEST) {
			i = table->memory_info->len;

			if (i < index)
				index = i;

			for (i = 0; i < index; i++) {
				struct heart_memory_info *mi;

				mi = g_array_index(table->memory_info, struct heart_memory_info *, i);

				if (mi->max_rss || mi->avg_rss)
					time++;

				if (max_rss < mi->max_rss)
					max_rss = mi->max_rss;

				avg_rss += mi->avg_rss;
			}

			if (time)
				avg_rss /= time;
		}

		md->max_rss = max_rss;
		md->avg_rss = avg_rss;
		g_array_append_val(arrays, md);
	}

	ret = pthread_mutex_unlock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}

	return RESOURCED_ERROR_NONE;

unlock_out_of_memory_exit:
	ret = pthread_mutex_unlock(&heart_memory_mutex);
	if (ret) {
		_E("pthread_mutex_unlock() failed, %d", ret);
		return RESOURCED_ERROR_FAIL;
	}
	return RESOURCED_ERROR_OUT_OF_MEMORY;
}

void heart_memory_update(struct logging_table_form *data, void *user_data)
{
	int ret;
	unsigned int rss;
	time_t curr_time = logging_get_time(CLOCK_BOOTTIME);
	struct heart_memory_table *find;
	struct heart_memory_table *table;

	if (sscanf(data->data, "%u", &rss) < 0) {
		_E("sscanf failed");
		return;
	}

	find = heart_memory_find_info(heart_memory_app_list, data->appid);

	/* get last node and update it & re-insert to hash table  */
	/* update */
	if (find) {
		ret = pthread_mutex_lock(&heart_memory_mutex);
		if (ret) {
			_E("pthread_mutex_lock() failed, %d", ret);
			return;
		}

		if (find->renew) {
			find->count = 0;
			find->max_rss = 0;
			find->total_rss = 0;
			find->renew = 0;
		}

		find->total_rss += rss;

		if (find->max_rss < rss)
			find->max_rss = rss;

		find->latest_rss = rss;

		find->count++;

		table = find;

		ret = pthread_mutex_unlock(&heart_memory_mutex);
		if (ret) {
			_E("pthread_mutex_unlock() failed, %d", ret);
			return;
		}
	} else {
		table = malloc(sizeof(struct heart_memory_table));

		if (!table) {
			_E("malloc failed");
			return;
		}

		if (snprintf(table->appid, MAX_APPID_LENGTH, "%s", data->appid) < 0) {
			free(table);
			_E("snprintf failed");
			return;
		}

		if (snprintf(table->pkgid, MAX_PKGNAME_LENGTH, "%s", data->pkgid) < 0) {
			free(table);
			_E("snprintf failed");
			return;
		}

		table->memory_info =
			g_array_new(false, false, sizeof(struct heart_memory_info *));

		table->total_rss = rss;

		table->max_rss = rss;

		table->latest_rss = rss;

		table->count = 1;

		ret = pthread_mutex_lock(&heart_memory_mutex);
		if (ret) {
			free(table);
			_E("pthread_mutex_lock() failed, %d", ret);
			return;
		}

		g_hash_table_insert(heart_memory_app_list, (gpointer)table->appid, (gpointer)table);

		ret = pthread_mutex_unlock(&heart_memory_mutex);
		if (ret) {
			_E("pthread_mutex_unlock() failed, %d", ret);
			return;
		}
	}

	if (last_file_update_time + HEART_MEMORY_SAVE_INTERVAL < curr_time) {
		/* all hash table update and make new array */
		gpointer value;
		gpointer key;
		GHashTableIter iter;
		struct heart_memory_table *search;
		struct heart_memory_info *new_memory_info;

		ret = pthread_mutex_lock(&heart_memory_mutex);
		if (ret) {
			_E("pthread_mutex_lock() failed, %d", ret);
			return;
		}

		g_hash_table_iter_init(&iter, heart_memory_app_list);

		while (g_hash_table_iter_next(&iter, &key, &value)) {
			search = (struct heart_memory_table *)value;

			new_memory_info = malloc(sizeof(struct heart_memory_info));

			if (!new_memory_info) {
				ret = pthread_mutex_unlock(&heart_memory_mutex);
				if (ret) {
					_E("pthread_mutex_lock() failed, %d", ret);
					return;
				}
				_E("malloc failed");
				return;
			}

			/* make new array node */
			new_memory_info->max_rss = search->max_rss;

			if (search->count)
				new_memory_info->avg_rss = search->total_rss / search->count;
			else
				new_memory_info->avg_rss = search->total_rss;

			new_memory_info->max_rss = search->max_rss;

			search->renew = 1;

			/* hashtable sliding : remove last node and make new one */
			g_array_remove_index(search->memory_info, MEM_ARRAY_MAX-1);
			g_array_prepend_val(search->memory_info, new_memory_info);
		}

		ret = pthread_mutex_unlock(&heart_memory_mutex);
		if (ret) {
			_E("pthread_mutex_unlock() failed, %d", ret);
			return;
		}
		/* rewrite hashtable list file */
		ret = heart_memory_save_to_file(heart_memory_app_list, HEART_MEMORY_FILE);
		if (ret) {
			_E("save to file failed");
			return;
		}

		last_file_update_time = curr_time;
	}
}

static int heart_memory_write(char *appid, char *pkgid, struct proc_status *p_data)
{
	_cleanup_free_ char *info = NULL;
	unsigned int usage = 0;
	int ret;
	struct proc_app_info *pai = p_data->pai;

	/* For write to data crud during period */
	/* write memory usage in proc_list */
	ret = proc_get_mem_usage(p_data->pid, &usage);
	if (ret < 0) {
		_E("Failed to get PID(%d) rss: %m", p_data->pid);
		return ret;
	}

	if (pai && pai->childs) {
		unsigned int child_usage;
		GSList *iterchild;

		gslist_for_each_item(iterchild, pai->childs) {
			pid_t child = GPOINTER_TO_PID(iterchild->data);
			ret = proc_get_mem_usage(child, &child_usage);
			if (ret < 0) {
				_E("Failed to get PID(%d) rss: %m", child);
				continue;
			}
			usage += child_usage;
		}
	}

	_SD("memory write : %s(%d), usage %u", appid, p_data->pid, usage);
	ret = asprintf(&info, "%u", usage);
	if (ret < 0) {
		_E("Failed to allocate memory");
		return -ENOMEM;
	}

	ret = logging_write(p_data->pid, MEM_NAME, appid, pkgid, time(NULL), info);

	if (ret)
		_E("logging_write failed %d", ret);

	return ret;
}

static int heart_memory_state_cb(void *data)
{
	int ret;
	char *appid, *pkgid;
	struct proc_status *ps = (struct proc_status *)data;

	ret = proc_get_id_info(ps, &appid, &pkgid);
	if (ret != RESOURCED_ERROR_NONE) {
		_E("Failed to proc_get_id_info");
		return ret;
	}

	heart_memory_write(appid, pkgid, ps);

	return RESOURCED_ERROR_NONE;
}

static gboolean heart_memory_notify(gpointer data)
{
	resourced_notify(RESOURCED_NOTIFIER_LOGGING_START, NULL);

	return true;
}

static void dbus_get_memory_latest(GDBusMethodInvocation *invocation, GVariant *params)
{
	int ret;
	char *appid = NULL;
	unsigned int rss = 0;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(&s)", &appid);
	if (!appid) {
		_E("Wrong message arguments!");
		goto failure;
	}

	ret = heart_memory_get_latest_data(appid, &rss);

	if (ret) {
		_E("heart_memory_get_latest_data failed %d", ret);
		goto failure;
	}

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(u)", rss));
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_get_memory_data(GDBusMethodInvocation *invocation, GVariant *params)
{
	int period = -1;
	char *appid = NULL;
	struct heart_memory_data *md;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(&si)", &appid, &period);
	if (!appid || period < 0) {
		_E("Wrong message arguments!");
		goto failure;
	}

	md = heart_memory_get_data(appid, period);
	if (!md) {
		_E("heart_memory_get_data failed");
		goto failure;
	}

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(uu)", md->max_rss, md->avg_rss));

	free(md);
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_get_memory_data_list(GDBusMethodInvocation *invocation, GVariant *params)
{
	int i, ret, period = -1;
	char *appid, *pkgid;
	GArray *temp_array;
	GVariantBuilder builder, *sub_builder;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(i)", &period);
	if (period < 0) {
		_E("Wrong message arguments!");
		goto failure;
	}

	temp_array = g_array_new(false, false, sizeof(struct heart_memory_data *));

	ret = heart_memory_get_table(temp_array, period);

	if (ret) {
		_E("heart_memory_get_table failed %d", ret);
		goto failure;
	}

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
	sub_builder = g_variant_builder_new(G_VARIANT_TYPE("a(ssuu)"));

	for (i = 0; i < temp_array->len; i++) {
		struct heart_memory_data *md;

		md = g_array_index(temp_array, struct heart_memory_data *, i);

		appid = md->appid;
		pkgid = md->pkgid;

		g_variant_builder_add(sub_builder, "(ssuu)", appid, pkgid, md->max_rss, md->avg_rss);
	}
	g_variant_builder_add_value(&builder, g_variant_new("a(ssuu)", sub_builder));
	g_variant_builder_unref(sub_builder);

	g_dbus_method_invocation_return_value(invocation, g_variant_builder_end(&builder));

	g_array_free(temp_array, true);
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_get_memorydb(GDBusMethodInvocation *invocation, GVariant *params)
{
	int i, period = -1;
	char *appid, *pkgid;
	GArray *temp_array;
	GVariantBuilder builder, *sub_builder;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(i)", &period);
	if (period < 0) {
		_E("Wrong message arguments!");
		goto failure;
	}

	temp_array = g_array_new(false, false, sizeof(struct heart_memory_data *));
	_E("start get read query!!! %ld", time(NULL));
	heart_memory_get_query(temp_array, period);
	_E("end get read query!!! %ld", time(NULL));

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
	sub_builder = g_variant_builder_new(G_VARIANT_TYPE("a(ssuu)"));

	for (i = 0; i < temp_array->len; i++) {
		struct heart_memory_data *md;

		md = g_array_index(temp_array, struct heart_memory_data *, i);

		appid = md->appid;
		pkgid = md->pkgid;

		g_variant_builder_add(sub_builder, "(ssuu)", appid, pkgid, md->max_rss, md->avg_rss);
	}
	g_variant_builder_add_value(&builder, g_variant_new("a(ssuu)", sub_builder));
	g_variant_builder_unref(sub_builder);

	g_dbus_method_invocation_return_value(invocation, g_variant_builder_end(&builder));

	g_array_free(temp_array, true);
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_get_memoryforeach(GDBusMethodInvocation *invocation, GVariant *params)
{
	int i, period = -1;
	char *appid, *pkgid;
	GArray *temp_array;
	GVariantBuilder builder, *sub_builder;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(i)", &period);
	if (period < 0) {
		_E("Wrong message arguments!");
		goto failure;
	}

	/* read from query */
	temp_array = g_array_new(false, false, sizeof(struct heart_memory_data *));
	_E("start get read foreach!!! %ld", time(NULL));
	logging_memory_get_foreach(temp_array, period);
	_E("end get read foreach!!! %ld", time(NULL));

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
	sub_builder = g_variant_builder_new(G_VARIANT_TYPE("a(ssuu)"));

	for (i = 0; i < temp_array->len; i++) {
		struct heart_memory_data *md;

		md = g_array_index(temp_array, struct heart_memory_data *, i);

		appid = md->appid;
		pkgid = md->pkgid;

		g_variant_builder_add(sub_builder, "(ssuu)", appid, pkgid, md->max_rss, md->avg_rss);
	}
	g_variant_builder_add_value(&builder, g_variant_new("a(ssuu)", sub_builder));
	g_variant_builder_unref(sub_builder);

	g_dbus_method_invocation_return_value(invocation, g_variant_builder_end(&builder));

	g_array_free(temp_array, true);
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_memory_save_to_file(GDBusMethodInvocation *invocation, GVariant *params)
{
	int ret;

	ret = heart_memory_save_to_file(heart_memory_app_list, HEART_MEMORY_FILE);

	if (ret) {
		_E("save to file failed");
		D_BUS_REPLY_NULL(invocation);
		return;
	}

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", ret));
}

static const char dbus_methods_xml[] =
"		<method name='GetMemoryLatest'>"
"			<arg type='s' name='Appid' direction='in'/>"
"			<arg type='u' name='Rss' direction='out'/>"
"		</method>"
"		<method name='GetMemoryData'>"
"			<arg type='s' name='Appid' direction='in'/>"
"			<arg type='i' name='Period' direction='in'/>"
"			<arg type='u' name='MaxRss' direction='out'/>"
"			<arg type='u' name='AvgRss' direction='out'/>"
"		</method>"
"		<method name='GetMemoryDataList'>"
"			<arg type='i' name='Period' direction='in'/>"
"			<arg type='a(ssuu)' name='MemoryUsagePerApp' direction='out'/>"
"		</method>"
"		<method name='GetMemoryDB'>"
"			<arg type='i' name='Period' direction='in'/>"
"			<arg type='a(ssuu)' name='MemoryUsagePerApp' direction='out'/>"
"		</method>"
"		<method name='GetMemoryforeach'>"
"			<arg type='i' name='Period' direction='in'/>"
"			<arg type='a(ssuu)' name='MemoryUsagePerApp' direction='out'/>"
"		</method>"
"		<method name='SaveMemoryData'>"
"			<arg type='i' name='ZeroOnSuccess' direction='out'/>"
"		</method>";

static struct d_bus_method dbus_methods[] = {
	{ "GetMemoryLatest",   dbus_get_memory_latest },
	{ "GetMemoryData",     dbus_get_memory_data },
	{ "GetMemoryDataList", dbus_get_memory_data_list },
	{ "GetMemoryDB",       dbus_get_memorydb },
	{ "GetMemoryforeach",  dbus_get_memoryforeach },
	{ "SaveMemoryData",    dbus_memory_save_to_file },
};

static int heart_memory_init(void *data)
{
	int ret;

	ret = logging_module_init(MEM_NAME, ONE_DAY, TEN_MINUTE, heart_memory_update, TEN_MINUTE, USER_DEFAULT);

	if (ret != RESOURCED_ERROR_NONE) {
		_E("logging module init failed");
		return RESOURCED_ERROR_FAIL;
	}

	if (!heart_memory_app_list) {
		heart_memory_app_list = g_hash_table_new_full(
				g_str_hash,
				g_str_equal,
				NULL,
				heart_memory_free_value);

		/* make hash from file */
		ret = heart_memory_read_from_file(heart_memory_app_list, HEART_MEMORY_FILE);

		if (ret == RESOURCED_ERROR_OUT_OF_MEMORY) {
			_E("heart_memory_init failed");
			return ret;
		}
	}

	register_notifier(RESOURCED_NOTIFIER_SERVICE_LAUNCH, heart_memory_state_cb);
	register_notifier(RESOURCED_NOTIFIER_APP_FOREGRD, heart_memory_state_cb);
	register_notifier(RESOURCED_NOTIFIER_APP_BACKGRD, heart_memory_state_cb);

	heart_dbus_declare_methods(dbus_methods_xml, dbus_methods, ARRAY_SIZE(dbus_methods));

	last_file_update_time = logging_get_time(CLOCK_BOOTTIME);

	if (heart_memory_update_timer == NULL) {
		_D("logging memory update timer start");
		heart_memory_update_timer = g_timeout_source_new_seconds(TEN_MINUTE);
		g_source_set_callback(heart_memory_update_timer, heart_memory_notify, NULL, NULL);
		g_source_attach(heart_memory_update_timer, NULL);
	}
	_D("heart memory init finished");
	return RESOURCED_ERROR_NONE;
}

static int heart_memory_exit(void *data)
{
	int ret;

	/* update timer delete */
	g_source_destroy(heart_memory_update_timer);
	heart_memory_update_timer = NULL;

	unregister_notifier(RESOURCED_NOTIFIER_SERVICE_LAUNCH, heart_memory_state_cb);
	unregister_notifier(RESOURCED_NOTIFIER_APP_FOREGRD, heart_memory_state_cb);
	unregister_notifier(RESOURCED_NOTIFIER_APP_BACKGRD, heart_memory_state_cb);

	if (heart_memory_app_list) {
		ret = heart_memory_save_to_file(heart_memory_app_list, HEART_MEMORY_FILE);

		if (ret)
			_E("save file failed %d", ret);

		if (heart_memory_app_list) {
			gpointer value;
			gpointer key;
			GHashTableIter iter;

			g_hash_table_iter_init(&iter, heart_memory_app_list);

			while (g_hash_table_iter_next(&iter, &key, &value)) {
				struct heart_memory_table *table = (struct heart_memory_table *)value;

				if (table->memory_info)
					g_array_free(table->memory_info, true);
			}

			g_hash_table_destroy(heart_memory_app_list);
		}
	}

	logging_module_exit();
	_D("heart memory exit");
	return RESOURCED_ERROR_NONE;
}

static const struct heart_module_ops heart_memory_ops = {
	.name           = "MEMORY",
	.init           = heart_memory_init,
	.exit           = heart_memory_exit,
};
HEART_MODULE_REGISTER(&heart_memory_ops)
