/*
 * resourced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * @file heart-storage.c
 *
 * @desc heart storage module
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <stdbool.h>
#include <time.h>

#include "resourced.h"
#include "trace.h"
#include "module.h"
#include "macro.h"
#include "notifier.h"
#include "heart.h"
#include "logging.h"
#include "heart-common.h"
#include "dbus-handler.h"
#include "proc-common.h"

#include <sqlite3.h>

#define STORAGE_NAME		"storage"

static bool heart_storage_initailized = false;

static void dbus_insert_log(GDBusMethodInvocation *invocation, GVariant *params)
{
	int ret;
	int pid = 0;
	char *pkgid = NULL;
	char *data = NULL;
	struct proc_app_info *pai;

	do_expr_unless_g_variant_get_typechecked(goto finish, params, "(&s&s)", &pkgid, &data);
	if (!pkgid) {
		_E("Wrong message arguments!");
		goto finish;
	}

	pai = find_app_info_by_appid(pkgid);
	if (!pai) {
		_E("Failed to find the app info of %s", pkgid);
		goto finish;
	}

	pid = pai->main_pid;

	_SD("Insert record (%s, %s)", pkgid, data);
	ret = logging_write(pid, STORAGE_NAME, pkgid, pkgid, time(NULL), data);
	if (ret != RESOURCED_ERROR_NONE)
		_E("Write request failed");

finish:
	D_BUS_REPLY_NULL(invocation);
}

static const char dbus_methods_xml[] =
"		<method name='Insert'>"
"			<arg type='s' name='Pkgid' direction='in'/>"
"			<arg type='s' name='UserData' direction='in'/>"
"		</method>";

static struct d_bus_method dbus_methods[] = {
	{ "Insert", dbus_insert_log },
};

static bool is_storage_logging(void)
{
	static const struct module_ops *block;

	if (!block) {
		block = find_module("block");
		if (block)
			heart_storage_initailized = true;
	}

	return heart_storage_initailized;
}

static int heart_storage_write(void *data)
{
	int ret;
	struct logging_data *ld = (struct logging_data *)data;

	ret = logging_write(ld->pid, STORAGE_NAME, ld->appid, ld->pkgid, time(NULL), ld->data);
	return ret;
}

static int heart_storage_init(void *data)
{
	int ret;

	if (!is_storage_logging())
		return RESOURCED_ERROR_UNINITIALIZED;

	ret = logging_module_init(STORAGE_NAME, FOUR_MONTH, FIVE_MINUTE, NULL, 0, USER_OWN);
	if (ret != RESOURCED_ERROR_NONE) {
		_E("logging module init failed");
		return RESOURCED_ERROR_FAIL;
	}

	heart_dbus_declare_methods(dbus_methods_xml, dbus_methods, ARRAY_SIZE(dbus_methods));

	register_notifier(RESOURCED_NOTIFIER_LOGGING_WRITE, heart_storage_write);
	return RESOURCED_ERROR_NONE;
}

static int heart_storage_exit(void *data)
{
	if (!heart_storage_initailized)
		return RESOURCED_ERROR_NONE;

	_D("heart_storage exit");
	unregister_notifier(RESOURCED_NOTIFIER_LOGGING_WRITE, heart_storage_write);
	logging_module_exit();
	return RESOURCED_ERROR_NONE;
}

static const struct heart_module_ops heart_storage_ops = {
	.name           = "STORAGE",
	.init           = heart_storage_init,
	.exit           = heart_storage_exit,
};
HEART_MODULE_REGISTER(&heart_storage_ops)
