/*
 * resourced
 *
 * Copyright (c) 2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/**
 * @file lowmem_handler.h
 * @desc handler function for setting memcgroup memory controller and
 *	receiving event fd.
 **/

#ifndef __LOWMEM_HANDLER_H__
#define __LOWMEM_HANDLER_H__

#include <proc-common.h>
#include <memory-common.h>

#define MAX_MEMORY_CGROUP_VICTIMS	10

enum lmk_type {
	LMK_MEMORY,	/* Kill all range of apps OOMADJ_INIT ~ OOMADJ_APP_MAX */
	LMK_FOREGROUND,	/* Kill foreground apps OOMADJ_INIT ~ OOMADJ_BACKGRD_PERCEPTIBLE */
	LMK_ACTIVE,	/* Kill active apps OOMADJ_PREVIOUS_DEFAULT ~ OOMADJ_BACKGRD_LOCKED */
	LMK_RECENTLY_USE, /* Kill recently use apps OOMADJ_FAVORITE ~ OOMADJ_RECENTLY_USED */
	LMK_OLDEST/* Kill only oldest inactive processes OOMADJ_BACKGRD_OLD ~ OOMADJ_APP_MAX */
};

struct task_info {
	/*
	 * Mostly, there are not multiple processes with the same pgid.
	 * So, for the frequent case, we use pid variable to avoid
	 * allocating arrays.
	 */
	pid_t pid;
	GArray *pids;
	pid_t pgid;
	/* oom_score_adj is smae as /proc/<pid>/oom_score_adj */
	int oom_score_adj;
	/*
	 * oom_score_lru is same as oom_score_adj or adjusted by
	 * proc_app_info lru_state for apps that are marked as favourite.
	 *
	 * oom_score_lru is the main value used in comparison for LMK.
	 */
	int oom_score_lru;
	int size;
};

void lowmem_dbus_init(void);
int lowmem_trigger_reclaim(int flags, int victims, enum lmk_type type, int threshold);
void lowmem_trigger_swap_reclaim(enum lmk_type type, int swap_size);
void lowmem_change_memory_state(int state, int force);
void lowmem_memcg_set_threshold(int idx, int level, int value);
void lowmem_memcg_set_leave_threshold(int idx, int value);
unsigned long lowmem_get_ktotalram(void);
void lowmem_trigger_swap(pid_t pid, int memcg_idx);
void lowmem_limit_init(void);
void lowmem_limit_exit(void);
void lowmem_limit_set(pid_t pid, unsigned int limit);
int lowmem_limit_move_cgroup(struct proc_app_info *pai);
void lowmem_reassign_limit(const char *dir, unsigned int limit);
unsigned int lowmem_get_task_mem_usage_rss(const struct task_info *tsk);
bool lowmem_fragmentated(void);
unsigned int lowmem_get_proactive_thres(void);
void lowmem_system_init();
void lowmem_system_exit();


/**
 * @desc restore memory cgroup from pid when resourced is restarted
 */
void lowmem_restore_memcg(struct proc_app_info *pai);

/*
 * Return memcg pointer to selected cgroup.
 */
int lowmem_get_memcg(enum memcg_type type, struct memcg **memcg_ptr);

enum oom_killer_cb_flags {
	OOM_NONE		= 0x0,		/* for main oom killer thread */
	OOM_FORCE		= (0x1 << 0),	/* for forced kill */
	OOM_NOMEMORY_CHECK	= (0x1 << 1),	/* check victims' memory */
	/*------------------------------------------------------------------*/
	OOM_IN_DEPTH		= (0x1 << 2),	/* consider all possible cgroups */
	OOM_SINGLE_SHOT		= (0x1 << 3),	/* do not retry if failed to reclaim memory */
	OOM_REVISE		= (0x1 << 4),   /* Run another attemp case failed for the first time*/
	OOM_DROP		= (0x1 << 5),	/* deactivate the worker */
};

/**
 * @brief  Low memory killer status
 *
 * LOWMEM_RECLAIM_NONE: no potential candidates for memory reclaim
 * LOWMEM_RECLAIM_DONE: requested size of memory has been successfully
 *			reclaimed through terminating number of processes
 * LOWMEM_RECLAIM_DROP: the whole reclaim procedure should be dropped
 * LOWMEM_RECLAIM_CONT: selected process might be considered as a potential
 *			memory reclaim source - green light for terminating
 *			the process
 * LOWMEM_RECLAIM_WAIT : check again after some seconds
 *			because killing processes will take some time.
 */
enum {
	LOWMEM_RECLAIM_NONE,
	LOWMEM_RECLAIM_DONE,
	LOWMEM_RECLAIM_DROP,
	LOWMEM_RECLAIM_CONT,
	LOWMEM_RECLAIM_WAIT,
};

#endif /*__LOWMEM_HANDLER_H__*/
