/*

 * resourced
 *
 * Copyright (c) 2000 - 2019 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/*
 * @file proc-monitor.c
 *
 * Copyright (c) 2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 */

#include <sys/types.h>
#include <unistd.h>
#include <sys/mount.h>

#include "proc-main.h"
#include "proc-monitor.h"
#include "resourced.h"
#include "macro.h"
#include "trace.h"
#include "dbus-handler.h"
#include "proc-process.h"
#include "proc-usage-stats.h"
#include "proc-usage-stats-helper.h"
#include "procfs.h"
#include "lowmem-handler.h"
#include "notifier.h"
#include "init.h"
#include "module.h"
#include "util.h"
#include "file-helper.h"

#define WATCHDOG_LAUNCHING_PARAM "PopupLaunch"
#define WATCHDOG_KEY1			"_SYSPOPUP_CONTENT_"
#define WATCHDOG_KEY2			"_APP_NAME_"
#define WATCHDOG_VALUE_1			"watchdog"

#define TIZEN_DEBUG_MODE_FILE	RD_SYS_ETC"/.debugmode"
#define DEFAULT_APPID	"NULL"

#define INIT_PID	1
#define POWER_OFF_DIRECT  2

static int current_lcd_state;

static GSource *watchdog_check_timer;
#define WATCHDOG_TIMER_INTERVAL		10

static struct proc_watchdog_info {
	pid_t pid;
	int signum;
} proc_watchdog = { -1, -1 };

int proc_debug_enabled(void)
{
	if (access(TIZEN_DEBUG_MODE_FILE, F_OK) == 0)
		return 1;
	else
		return 0;
}

static void dbus_get_meminfo(GDBusMethodInvocation *invocation, GVariant *params)
{
	unsigned int mem_total, mem_free, mem_available, cached, used;
	unsigned int swap_total, swap_free, swap;
	struct meminfo mi;
	int r;

	r = proc_get_meminfo(&mi,
			     MEMINFO_MASK_MEM_TOTAL |
			     MEMINFO_MASK_MEM_FREE |
			     MEMINFO_MASK_MEM_AVAILABLE |
			     MEMINFO_MASK_CACHED |
			     MEMINFO_MASK_SWAP_TOTAL |
			     MEMINFO_MASK_SWAP_FREE);
	if (r < 0) {
		_E("Failed to get meminfo: %m");
		D_BUS_REPLY_NULL(invocation);
		return;
	}

	mem_total = mi.value[MEMINFO_ID_MEM_TOTAL];
	mem_free = mi.value[MEMINFO_ID_MEM_FREE];
	mem_available = mi.value[MEMINFO_ID_MEM_AVAILABLE];
	cached = mi.value[MEMINFO_ID_CACHED];
	swap_total = mi.value[MEMINFO_ID_SWAP_TOTAL];
	swap_free = mi.value[MEMINFO_ID_SWAP_FREE];

	used = mem_total - mem_available;
	swap = swap_total - swap_free;

	_D("memory info total = %u, free = %u, cache = %u, used = %u, swap = %u",
		mem_total, mem_free, cached, used, swap);

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(uuuuu)",
				mem_total, mem_free, cached, used, swap));
}

static void dbus_reclaim_memory(GDBusMethodInvocation *invocation, GVariant *params)
{
	_D("reclaiming memory!");

	bool ok = true;
	ok &= (RESOURCED_ERROR_NONE == proc_sys_node_trigger(SYS_VM_SHRINK_MEMORY));
	ok &= (RESOURCED_ERROR_NONE == proc_sys_node_trigger(SYS_VM_COMPACT_MEMORY));

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", ok ? RESOURCED_ERROR_NONE : RESOURCED_ERROR_FAIL));
}

static void turn_off_vip_handler()
{
	/* Set the agent to something harmless. This destroys the previous
	 * state by losing information about the previous agent, though it's
	 * fine because the system is shutting down anyway and the "proper"
	 * way of doing this (iterate over all subdirectories and set their
	 * `notify_on_release` to 0) is bound to be slower than this. */
	static const char AGENT_REPLACEMENT_PATH[] = "/bin/true";
	if (fwrite_str("/sys/fs/cgroup/vip/release_agent", AGENT_REPLACEMENT_PATH) == RESOURCED_ERROR_NONE) {
		_I("disabled VIP release agent by setting to %s", AGENT_REPLACEMENT_PATH);
		return;
	}

	/* Fallback: subvert the agent itself. This is somewhat dangerous
	 * as now the program does something else than expected but the
	 * system is shutting down and nobody else should use the agent
	 * anyway so there isn't much opportunity to cause any surprises. */
	static const char VIP_RELEASE_AGENT_PATH[] = "/usr/bin/vip-release-agent";
	if (!mount(AGENT_REPLACEMENT_PATH, VIP_RELEASE_AGENT_PATH, NULL, MS_BIND | MS_RDONLY, NULL)) {
		_I("disabled VIP release agent by bind mounting %s to %s", AGENT_REPLACEMENT_PATH, VIP_RELEASE_AGENT_PATH);
		return;
	}

	_E("disabling VIP release agent by bind mounting %s to %s failed; VIP release agent is potentially still alive", AGENT_REPLACEMENT_PATH, VIP_RELEASE_AGENT_PATH);
}

static void turn_off_vip()
{
	/* Turn off the "reboot on VIP death" behaviour.
	 * It is normal for VIPs to die throughout a poweroff, while it
	 * is undesirable for the poweroff to be replaced by a reboot. */
	turn_off_vip_handler();

	/* Unmounting the VIP cgroup does not make the above redundant because
	 * the unmounting can be done lazily and is not guaranteed to take place
	 * immediately. */
	umount2("/sys/fs/cgroup/vip", MNT_FORCE | MNT_DETACH);
	_I("disabled VIP by unmounting VIP cgroup");
}

static void proc_poweroff()
{
	_D("pre power off: unmounting cgroup fs");

	proc_sweep_memory(PROC_SWEEP_EXCLUDE_ACTIVE, INIT_PID);
	resourced_notify(RESOURCED_NOTIFIER_POWER_OFF, NULL);
	turn_off_vip();
}

static void dbus_pre_poweroff(GDBusMethodInvocation *invocation, GVariant *params)
{
	proc_poweroff();
	g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", -1));
}

static void dbus_get_checkappstatus(GDBusMethodInvocation *invocation, GVariant *params)
{
	int len;
	int detect = 0;
	struct proc_app_info *pai;
	char *appid = DEFAULT_APPID;
	char *type = NULL;
	pid_t pid = -1;

	do_expr_unless_g_variant_get_typechecked(goto null_reply, params, "(i&s)", &pid, &type);
	if (pid < 0 || type == NULL) {
		_D("there is no message");
		goto null_reply;
	}

	pai = find_app_info(pid);
	if (!pai) {
		_D("There is no appid %d", pid);
		goto response;
	}

	/*
	 * allowed strings: normal, lcddim, lcdoff
	 * normal & lcddim : allow only UI application
	 * lcdoff : allow if it sends the excluding dbus signal
	 */
	len = strlen(type);

	if (len != 6) {
		_D("wrong lock type : %s", type);
		goto response;
	}

	appid = pai->appid;
	if (pai->watchdog_exclude || pai->runtime_exclude) {
		_E("application (%d, %s) used %s lock longly but it is excluded", pid, appid, type);
		goto response;
	}

	if (!strncmp(type, "norm", 4) || !strncmp(type, "lcdd", 4)) {
		if (pai->type == PROC_TYPE_SERVICE || pai->type == PROC_TYPE_READY ||
		    pai->type == PROC_TYPE_NONE)
			detect = 1;
	} else if (!strncmp(type, "lcdo", 4)) {
		detect = 1;
	}

	if (detect)
		_E("detect application locking %s lock longly : %d, %s", type, pid, appid);
	else
		_D("application locking %s lock longly but it seems normal : %d, %s", type, pid, appid);

response:
	g_dbus_method_invocation_return_value(invocation, g_variant_new("(iis)",
				pid, detect, appid));
	return;

null_reply:
	D_BUS_REPLY_NULL(invocation);
}

static void proc_dbus_active_signal_handler(GVariant *params)
{
	int type;
	char *str = NULL;
	pid_t pid = 0;

	do_expr_unless_g_variant_get_typechecked(return, params, "(&si)", &str, &pid);
	if (!str || pid == 0) {
		_D("there is no message");
		return;
	}

	if (!strncmp(str, "active", strlen("active")+1))
		type = PROC_CGROUP_SET_ACTIVE;
	else if (!strncmp(str, "inactive", strlen("inactive")+1))
		type = PROC_CGROUP_SET_INACTIVE;
	else
		return;

	_D("received %s signal for pid %d", str, pid);
	resourced_proc_status_change(type, pid, NULL, NULL, PROC_TYPE_NONE);
}

static void dbus_get_app_cpu(GDBusMethodInvocation *invocation, GVariant *params)
{
	char *appid = NULL;
	unsigned long total, utime, stime, starttime;
	struct proc_app_info *pai = NULL;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(&s)", &appid);
	if (!appid) {
		_E("Wrong message arguments!");
		goto failure;
	}

	pai = find_app_info_by_appid(appid);
	if (!pai) {
		_E("There is no appid %s", appid);
		goto failure;
	}

	if (proc_get_cpu_time(pai->main_pid, &utime, &stime, &starttime) != RESOURCED_ERROR_NONE) {
		_E("proc_get_cpu_time = %s (%d)", appid, pai->main_pid);
		goto failure;
	}

	_D("cpu usage of %s (%d), utime = %lu, stime = %lu", appid, pai->main_pid, utime, stime);
	total = utime + stime;

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(u)", total));
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_get_app_memory(GDBusMethodInvocation *invocation, GVariant *params)
{
	char *appid;
	unsigned int usage = 0;
	struct proc_app_info *pai = NULL;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(&s)", &appid);
	if (!appid) {
		_E("Wrong message arguments!");
		goto failure;
	}

	pai = find_app_info_by_appid(appid);
	if (!pai || !pai->main_pid) {
		_E("There is no appid %s", appid);
		goto failure;
	}

	if (proc_get_mem_usage(pai->main_pid, &usage) < 0) {
		_E("lowmem_get_proc_mem_usage failed for appid = %s (%d)",
			appid, pai->main_pid);
		goto failure;
	}

	_D("memory usage of %s (%d), rss = %u", appid, pai->main_pid, usage);

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(u)", usage));
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_get_memory_list(GDBusMethodInvocation *invocation, GVariant *params)
{
	_cleanup_app_list_close_ GSList *proc_app_list = PAL_INIT_VALUE;
	GSList *giter;
	char *appid;
	struct proc_app_info *pai;
	unsigned int usage = 0;
	GVariantBuilder builder, *sub_builder;

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
	sub_builder = g_variant_builder_new(G_VARIANT_TYPE("a(su)"));

	proc_app_list = proc_app_list_open();
	gslist_for_each_item(giter, proc_app_list) {
		pai = (struct proc_app_info *)giter->data;
		if (!pai || !pai->main_pid)
			continue;
		if (proc_get_mem_usage(pai->main_pid, &usage) < 0)
			continue;

		appid = pai->appid;

		g_variant_builder_add(sub_builder, "(su)", appid, usage);
	}
	g_variant_builder_add_value(&builder, g_variant_new("a(su)", sub_builder));
	g_variant_builder_unref(sub_builder);

	g_dbus_method_invocation_return_value(invocation, g_variant_builder_end(&builder));
}

static void dbus_get_cpu_list(GDBusMethodInvocation *invocation, GVariant *params)
{
	_cleanup_app_list_close_ GSList *proc_app_list = PAL_INIT_VALUE;
	GSList *giter;
	char *appid;
	struct proc_app_info *pai;
	unsigned long uptime, utime, stime, deltatime, starttime;
	int ret;
	GVariantBuilder builder, *sub_builder;

	proc_app_list = proc_app_list_open();

	ret = proc_get_uptime(&uptime);
	if (ret) {
		_E("Failed to get uptime");
		D_BUS_REPLY_NULL(invocation);
		return;
	}

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
	sub_builder = g_variant_builder_new(G_VARIANT_TYPE("a(su)"));

	gslist_for_each_item(giter, proc_app_list) {
		unsigned long percent;
		pai = (struct proc_app_info *)giter->data;
		if (!pai->main_pid)
			continue;
		if (proc_get_cpu_time(pai->main_pid, &utime, &stime, &starttime) != RESOURCED_ERROR_NONE)
			continue;
		appid = pai->appid;

		deltatime = uptime - pai->starttime;
		if (deltatime == 0)
			percent = 0;
		else
			percent = (utime + stime) / deltatime;

		g_variant_builder_add(sub_builder, "(su)", appid, percent);
	}
	g_variant_builder_add_value(&builder, g_variant_new("a(su)", sub_builder));
	g_variant_builder_unref(sub_builder);

	g_dbus_method_invocation_return_value(invocation, g_variant_builder_end(&builder));
}

static void dbus_get_memory_lists(GDBusMethodInvocation *invocation, GVariant *params)
{
	GSList *proc_app_list = PAL_INIT_VALUE;
	GSList *giter;
	char *appid;
	int type;
	struct proc_app_info *pai;
	unsigned int usage = 0;
	GVariantBuilder builder, *sub_builder;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(i)", &type);
	if (type < 0 || type >= PROC_TYPE_MAX) {
		_E("Wrong message arguments!");
		goto failure;
	}

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
	sub_builder = g_variant_builder_new(G_VARIANT_TYPE("a(su)"));

	proc_app_list = proc_app_list_open();
	gslist_for_each_item(giter, proc_app_list) {
		pai = (struct proc_app_info *)giter->data;
		if (!pai || !pai->main_pid)
			continue;
		if (type != PROC_TYPE_MAX && pai->type != type)
			continue;
		if (proc_get_mem_usage(pai->main_pid, &usage) < 0)
			continue;

		appid = pai->appid;

		g_variant_builder_add(sub_builder, "(su)", appid, usage);
	}
	g_variant_builder_add_value(&builder, g_variant_new("a(su)", sub_builder));
	g_variant_builder_unref(sub_builder);

	g_dbus_method_invocation_return_value(invocation, g_variant_builder_end(&builder));
	proc_app_list_close();
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void dbus_get_cpu_lists(GDBusMethodInvocation *invocation, GVariant *params)
{
	GSList *proc_app_list = PAL_INIT_VALUE;
	GSList *giter;
	int ret, type;
	char *appid;
	struct proc_app_info *pai;
	unsigned long uptime, utime, stime, deltatime, starttime;
	GVariantBuilder builder, *sub_builder;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(i)", &type);
	if (type < 0 || type >= PROC_TYPE_MAX) {
		_E("Wrong message arguments!");
		goto failure;
	}

	ret = proc_get_uptime(&uptime);
	if (ret) {
		_E("Failed to get uptime");
		goto failure;
	}

	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
	sub_builder = g_variant_builder_new(G_VARIANT_TYPE("a(su)"));

	proc_app_list = proc_app_list_open();
	gslist_for_each_item(giter, proc_app_list) {
		unsigned long percent;
		pai = (struct proc_app_info *)giter->data;
		if (!pai->main_pid)
			continue;
		if (type != PROC_TYPE_MAX && pai->type != type)
			continue;
		if (proc_get_cpu_time(pai->main_pid, &utime, &stime, &starttime) != RESOURCED_ERROR_NONE)
			continue;
		appid = pai->appid;

		/*
		 * cpu usage can be calculated using total usage of stime and utime out of delta time.
		 * cpu usage = (utime + stime ticks) / (current uptime - process start time) * 100 percentage
		 *           = (utime + stime seconds) / (current uptime - process start time)
		 *
		 * Meanwhile, delta time between uptime and starttime should be checked
		 * whether it is zero or not.
		 * If delta time is zero, it can cause "divided by zero" fatal error.
		 */
		deltatime = uptime - pai->starttime;
		if (deltatime == 0)
			percent = 0;
		else
			percent = (utime + stime) / deltatime;

		g_variant_builder_add(sub_builder, "(su)", appid, percent);
	}
	g_variant_builder_add_value(&builder, g_variant_new("a(su)", sub_builder));
	g_variant_builder_unref(sub_builder);

	g_dbus_method_invocation_return_value(invocation, g_variant_builder_end(&builder));
	proc_app_list_close();
	return;

failure:
	D_BUS_REPLY_NULL(invocation);
}

static void proc_dbus_exclude_signal_handler(GVariant *params)
{
	const char *str = NULL;
	pid_t pid = 0;
	struct proc_exclude pe;
	int len;

	do_expr_unless_g_variant_get_typechecked(return, params, "(&si)", &str, &pid);
	if (!str || pid == 0) {
		_D("there is no message");
		return;
	}

	/*
	 * allowed strings: wakeup, exclude, include
	 */
	len = strlen(str);

	_D("%s signal on pid = %d", str, pid);
	if (len == 6 && !strncmp(str, "wa", 2)) {
		struct proc_status ps = {0};

		ps.pai = find_app_info(pid);
		ps.pid = pid;
		resourced_notify(RESOURCED_NOTIFIER_APP_WAKEUP, &ps);
	} else if (len == 7) {
		if (!strncmp(str, "ex", 2)) {
			pe.pid = pid;
			pe.type = PROC_EXCLUDE;
			resourced_notify(RESOURCED_NOTIFIER_CONTROL_EXCLUDE, &pe);
			proc_set_runtime_exclude_list(pe.pid, pe.type);
		} else if (!strncmp(str, "in", 2)) {
			pe.pid = pid;
			pe.type = PROC_INCLUDE;
			resourced_notify(RESOURCED_NOTIFIER_CONTROL_EXCLUDE, &pe);
			proc_set_runtime_exclude_list(pe.pid, pe.type);
		}
	} else
		return;
}

static void proc_dbus_exclude_method_handler(GDBusMethodInvocation *invocation, GVariant *params)
{
	const char *str = NULL;
	pid_t pid = 0;

	do_expr_unless_g_variant_get_typechecked(goto failure, params, "(&si)", &str, &pid);
	if (!str || pid == 0) {
		_E("invalid parameters");
		goto failure;
	}
	proc_dbus_exclude_signal_handler(params);
	D_BUS_REPLY_NULL(invocation);
	return;

failure:
	g_dbus_method_invocation_return_error(invocation, G_DBUS_ERROR,
		G_DBUS_ERROR_INVALID_ARGS, "invalid arguments");
}

static void proc_dbus_exclude_appid_signal_handler(GVariant *params)
{
	char *str = NULL;
	char *appid = NULL;
	struct proc_exclude pe;
	int len;
	struct proc_status ps = {0};

	do_expr_unless_g_variant_get_typechecked(return, params, "(&s&s)", &str, &appid);
	if (!str || !appid) {
		_D("there is no message");
		return;
	}

	ps.pai = find_app_info_by_appid(appid);
	if (!ps.pai) {
		_E("no entry of %s in app list", appid);
		return;
	}
	ps.pid = ps.pai->main_pid;

	/*
	 * allowed strings: wakeup, exclude, include
	 */
	len = strlen(str);

	_D("%s signal on app = %s", str, appid);
	if (len == 6 && !strncmp(str, "wa", 2)) {
		resourced_notify(RESOURCED_NOTIFIER_APP_WAKEUP, &ps);
	} else if (len == 7) {
		if (!strncmp(str, "ex", 2)) {
			pe.pid = ps.pid;
			pe.type = PROC_EXCLUDE;
			resourced_notify(RESOURCED_NOTIFIER_CONTROL_EXCLUDE, &pe);
			proc_set_runtime_exclude_list(pe.pid, pe.type);
		} else if (!strncmp(str, "in", 2)) {
			pe.pid = ps.pid;
			pe.type = PROC_INCLUDE;
			resourced_notify(RESOURCED_NOTIFIER_CONTROL_EXCLUDE, &pe);
			proc_set_runtime_exclude_list(pe.pid, pe.type);
		}
	} else
		return;
}


static void proc_dbus_prelaunch_signal_handler(GVariant *params)
{
	char *appid = NULL;
	char *pkgid = NULL;
	int flags = -1;
	int categories = -1;
	struct proc_status ps;
	struct proc_app_info *pai;

	do_expr_unless_g_variant_get_typechecked(return, params, "(&s&sii)", &appid, &pkgid, &flags, &categories);
	if (!appid || !pkgid || flags < 0 || categories < 0) {
		_D("there is no message");
		return;
	}

	_D("call proc_dbus_prelaunch_handler: appid = %s, pkgid = %s, flags = %d,"
		" categories = %X\n", appid, pkgid, flags, categories);

	pai = proc_add_app_info(appid, pkgid, 0, flags, categories, PROC_TYPE_READY, PROC_STATE_DEFAULT);
	ps.pai = pai;
	resourced_notify(RESOURCED_NOTIFIER_APP_PRELAUNCH, &ps);
}

static void proc_dbus_sweep_signal_handler(GVariant *params)
{
	proc_sweep_memory(PROC_SWEEP_INCLUDE_ACTIVE, INIT_PID);
}

static gboolean check_watchdog_cb(gpointer data)
{
	int oom_score_adj = 0, ret;
	pid_t pid = proc_watchdog.pid;

	ret = proc_get_oom_score_adj(pid, &oom_score_adj);
	if (!ret) {
		_E("watchdog pid %d not terminated, kill again\n", pid);
		kill(pid, SIGKILL);
	}
	watchdog_check_timer = NULL;
	proc_watchdog.pid = -1;
	proc_watchdog.signum = -1;

	return false;
}

static int proc_dbus_show_popup(const char *value)
{
	char str_val[32];	// The original max size of represented string is 32

	snprintf(str_val, sizeof(str_val), "%s", value);

	GVariantBuilder *const gv_builder = g_variant_builder_new(G_VARIANT_TYPE("a{ss}"));
	g_variant_builder_add(gv_builder, "{ss}", WATCHDOG_KEY1, WATCHDOG_VALUE_1);
	g_variant_builder_add(gv_builder, "{ss}", WATCHDOG_KEY2, str_val);

	GVariant *const params = g_variant_new("(a{ss})", gv_builder);
	g_variant_builder_unref(gv_builder);

	int ret_val = d_bus_call_method_sync_gvariant(SYSTEM_POPUP_BUS_NAME, SYSTEM_POPUP_PATH_SYSTEM,
		SYSTEM_POPUP_IFACE_SYSTEM, WATCHDOG_LAUNCHING_PARAM, params);

	g_variant_unref(params);

	return ret_val;
}

static void proc_dbus_watchdog_handler(GVariant *params)
{
	int ret;
	int pid = -1;
	int command = -1;
	char appname[PROC_NAME_MAX];
	struct proc_status ps;

	do_expr_unless_g_variant_get_typechecked(return, params, "(ii)", &pid, &command);
	if (pid < 0 || command < 0) {
		_D("there is no message");
		return;
	}

	ret = proc_get_cmdline(pid, appname);
	if (ret != RESOURCED_ERROR_NONE) {
		_E("ERROR : invalid pid(%d)", pid);
		return;
	}

	ret = resourced_watchdog_excluded(appname);
	if (ret == RESOURCED_ERROR_NONMONITOR)
		return;

	if (current_lcd_state == LCD_STATE_OFF) {
		_E("Receive watchdog signal to pid: %d(%s) but don't show ANR popup in LCD off state\n", pid, appname);
		return;
	}

	_E("Receive watchdog signal to app %s, pid %d\n", appname, pid);
	ps.pai = find_app_info(pid);
	ps.pid = pid;
	resourced_notify(RESOURCED_NOTIFIER_APP_ANR, &ps);

	if (proc_dbus_show_popup(appname) < 0)
		_E("Failed to show ANR popup");

	if (watchdog_check_timer) {
		if (proc_watchdog.pid == pid) {
			_E("app %s, pid %d has already received watchdog siganl but not terminated", appname, pid);
			kill(pid, SIGKILL);
			proc_watchdog.pid = -1;
			proc_watchdog.signum = -1;
			return;
		}
	}

	resourced_proc_status_change(PROC_CGROUP_SET_TERMINATE_REQUEST,
		    pid, NULL, NULL, PROC_TYPE_NONE);
	kill(pid, SIGABRT);
	if (!watchdog_check_timer) {
		watchdog_check_timer = g_timeout_source_new_seconds(WATCHDOG_TIMER_INTERVAL);
		g_source_set_callback(watchdog_check_timer, check_watchdog_cb, NULL, NULL);
		g_source_attach(watchdog_check_timer, NULL);

		proc_watchdog.pid = pid;
		proc_watchdog.signum = command;
	}
}

static void send_dump_signal(char *signal)
{
	char *arg[1];
	char tmpc;
	int pid_len = 0; // should get the length of the PID decimal representation

	pid_t pid = getpid();
	pid_len = snprintf(&tmpc, 1, "%u", pid);
	arg[0] = (char *)malloc(pid_len+1);
	g_assert(arg[0]);

	snprintf(arg[0], pid_len+1, "%u", pid);

	d_bus_broadcast_signal(DUMP_SERVICE_OBJECT_PATH,
	    DUMP_SERVICE_INTERFACE_NAME, signal, "i", arg);

	free(arg[0]);
}

static void proc_dbus_dump_handler(GVariant *params)
{
	char *path = NULL;
	int mode = -1;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i&s)", &mode, &path);
	if (mode < 0 || !path) {
		_D("there is no message");
		return;
	}

	_D("received dump request: path %s", path);
	send_dump_signal(SIGNAL_DUMP_START);
	resourced_proc_dump(mode, path);
	send_dump_signal(SIGNAL_DUMP_FINISH);
}

static void proc_dbus_systemservice_handler(GVariant *params)
{
	pid_t pid = 0;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i)", &pid);
	if (pid == 0) {
		_D("there is no message");
		return;
	}

	resourced_proc_status_change(PROC_CGROUP_SET_SYSTEM_SERVICE, pid,
		    NULL, NULL, PROC_TYPE_NONE);
}

static void proc_dbus_lcd_on(GVariant *params)
{
	current_lcd_state = LCD_STATE_ON;
	resourced_notify(RESOURCED_NOTIFIER_LCD_ON, NULL);
	/* nothing */
}

static void proc_dbus_lcd_off(GVariant *params)
{
	current_lcd_state = LCD_STATE_OFF;
	resourced_notify(RESOURCED_NOTIFIER_LCD_OFF, NULL);
}

static void booting_done_signal_handler(GVariant *params)
{
	_I("booting done");
	resourced_notify(RESOURCED_NOTIFIER_BOOTING_DONE, NULL);
}

static void low_battery_signal_handler(GVariant *params)
{
	resourced_notify(RESOURCED_NOTIFIER_LOW_BATTERY, NULL);
}

static void poweroff_signal_handler(GVariant *params)
{
	int response;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i)", &response);

	/* This code does not check for values which could be too high.
	 * This is for legacy reasons (we inherited the code without such
	 * checks so potentially somebody might be using it that way). */
	if (response < POWER_OFF_DIRECT) {
		_D("Invalid power off status: %d", response);
		return;
	}

	proc_poweroff();
}

static void systemtime_changed_signal_handler(GVariant *params)
{
	resourced_notify(RESOURCED_NOTIFIER_SYSTEMTIME_CHANGED, NULL);
}

static void proc_dbus_aul_launch(GVariant *params)
{
	pid_t pid = 0;
	int status, apptype;
	char *appid = NULL;
	char *pkgid = NULL;
	char *pkgtype = NULL;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i&s&s&s)", &pid, &appid, &pkgid, &pkgtype);
	if (pid == 0 || !appid || !pkgid || !pkgtype) {
		_D("there is no message");
		return;
	}

#ifdef PROC_DEBUG
	_D("aul_launch: app %s, pkgd %s, pid %d, pkgtype %s",
			appid, pkgid, pid, pkgtype);
#endif

	if (!strncmp(pkgtype, "svc", 3)) {
		apptype = PROC_TYPE_SERVICE;
		status = PROC_CGROUP_SET_SERVICE_REQUEST;
	} else if (!strncmp(pkgtype, "ui", 2)) {
		apptype = PROC_TYPE_GUI;
		status = PROC_CGROUP_SET_LAUNCH_REQUEST;
	} else if (!strncmp(pkgtype, "widget", 6)) {
		apptype = PROC_TYPE_WIDGET;
		status = PROC_CGROUP_SET_LAUNCH_REQUEST;
	} else if (!strncmp(pkgtype, "watch", 5)) {
		apptype = PROC_TYPE_WATCH;
		status = PROC_CGROUP_SET_LAUNCH_REQUEST;
	} else
		return;

	resourced_proc_status_change(status, pid, appid, pkgid, apptype);
}

static void proc_dbus_aul_resume(GVariant *params)
{
	pid_t pid = 0;
	int status = PROC_CGROUP_SET_RESUME_REQUEST, apptype;
	char *appid = NULL;
	char *pkgid = NULL;
	char *pkgtype = NULL;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i&s&s&s)", &pid, &appid, &pkgid, &pkgtype);
	if (pid == 0 || !appid || !pkgid || !pkgtype) {
		_D("there is no message");
		return;
	}

	if (!strncmp(pkgtype, "svc", 3))
		apptype = PROC_TYPE_SERVICE;
	else if (!strncmp(pkgtype, "widget", 6))
		apptype = PROC_TYPE_WIDGET;
	else if (!strncmp(pkgtype, "watch", 5))
		apptype = PROC_TYPE_WATCH;
	else
		apptype = PROC_TYPE_GUI;

	resourced_proc_status_change(status, pid, appid, pkgid, apptype);
}

static void proc_dbus_aul_terminate(GVariant *params)
{
	pid_t pid = 0;
	int status = PROC_CGROUP_SET_TERMINATE_REQUEST;
	char *appid = NULL;
	char *pkgid = NULL;
	char *pkgtype = NULL;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i&s&s&s)", &pid, &appid, &pkgid, &pkgtype);
	if (pid == 0 || !appid || !pkgid || !pkgtype) {
		_D("there is no message");
		return;
	}

	resourced_proc_status_change(status, pid, appid, pkgid, PROC_TYPE_NONE);
}

static void proc_dbus_aul_changestate(GVariant *params)
{
	pid_t pid = 0;
	int status, apptype;
	char *appid = NULL;
	char *pkgid = NULL;
	char *statstr = NULL;
	char *pkgtype = NULL;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i&s&s&s&s)", &pid, &appid, &pkgid, &statstr, &pkgtype);
	if (pid == 0 || !appid || !pkgid || !statstr || !pkgtype) {
		_D("there is no message");
		return;
	}

	if (!strncmp(statstr, "fg", 2))
		status = PROC_CGROUP_SET_FOREGRD;
	else if (!strncmp(statstr, "bg", 2))
		status = PROC_CGROUP_SET_BACKGRD;
	else
		return;

	if (!strncmp(pkgtype, "svc", 3))
		apptype = PROC_TYPE_SERVICE;
	else if (!strncmp(pkgtype, "widget", 6))
		apptype = PROC_TYPE_WIDGET;
	else if (!strncmp(pkgtype, "watch", 5))
		apptype = PROC_TYPE_WATCH;
	else
		apptype = PROC_TYPE_GUI;

	resourced_proc_status_change(status, pid, appid, pkgid, apptype);
}

static void proc_dbus_aul_group(GVariant *params)
{
	pid_t ownerpid = 0;
	pid_t childpid = 0;
	char *appid = NULL;

	do_expr_unless_g_variant_get_typechecked(return, params, "(ii&s)", &ownerpid, &childpid, &appid);
	if (ownerpid == 0 || childpid == 0 || !appid) {
		_D("there is no message");
		return;
	}

	if (ownerpid == childpid) {
		_I("specific case, one application %d makes multiple window. skip to merge", ownerpid);
		return;
	}

	_D("received process grouping : owner %d, child %d, previous appid %s",
		    ownerpid, childpid, appid);
	proc_set_group(ownerpid, childpid, appid);
}

static void proc_dbus_aul_terminated(GVariant *params)
{
	pid_t pid = 0;
	int status = PROC_CGROUP_SET_TERMINATED;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i)", &pid);
	if (pid == 0) {
		_D("there is no message");
		return;
	}

	resourced_proc_status_change(status, pid, NULL, NULL, PROC_TYPE_NONE);
}

static void proc_dbus_suspend_hint(GVariant *params)
{
	pid_t pid = 0;
	struct proc_app_info *pai = NULL;
	struct proc_status ps = {0};
	enum proc_state state;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i)", &pid);
	if (pid == 0) {
		_D("there is no message");
		return;
	}

	pai = find_app_info(pid);
	if (!pai)
		return;

	if (pai->appid)
		_D("received suspend hint : app %s, pid %d",
				pai->appid, pid);
	else
		_D("received suspend hint : pid %d", pid);

	state = proc_check_suspend_state(pai);
	if (state == PROC_STATE_SUSPEND) {
		ps.pid = pid;
		ps.pai = pai;
		resourced_notify(RESOURCED_NOTIFIER_APP_SUSPEND,
			    &ps);
	}
}

static const char dbus_methods_xml[] =
"<node>"
"	<interface name='"RESOURCED_INTERFACE_PROCESS"'>"
"		<method name='GetAppCpu'>"
"			<arg type='s' name='Appid' direction='in'/>"
"			<arg type='u' name='CpuUsage' direction='out'/>"
"		</method>"
"		<method name='GetCpuList'>"
"			<arg type='a(su)' name='PidWithCpuUsagePercent' direction='out'/>"
"		</method>"
"		<method name='GetCpuLists'>"
"			<arg type='i' name='AppType' direction='in'/>"
"			<arg type='a(su)' name='UsagePercent' direction='out'/>"
"		</method>"
"		<method name='GetAppMemory'>"
"			<arg type='s' name='Appid' direction='in'/>"
"			<arg type='u' name='MemoryUsage' direction='out'/>"
"		</method>"
"		<method name='GetMemoryList'>"
"			<arg type='a(su)' name='PidWithMemoryUsage' direction='out'/>"
"		</method>"
"		<method name='GetMemoryLists'>"
"			<arg type='i' name='AppType' direction='in'/>"
"			<arg type='a(su)' name='PidWithMemoryUsage' direction='out'/>"
"		</method>"
"		<method name='GetMemInfo'>"
"			<arg type='u' name='MemTotal' direction='out'/>"
"			<arg type='u' name='MemFree' direction='out'/>"
"			<arg type='u' name='MemCached' direction='out'/>"
"			<arg type='u' name='MemUsed' direction='out'/>"
"			<arg type='u' name='MemSwap' direction='out'/>"
"		</method>"
"		<method name='ReclaimMemory'>"
"			<arg type='i' name='ZeroOnSuccess' direction='out'/>"
"		</method>"
"		<method name='PrePoweroff'>"
"			<arg type='i' name='MinusOne' direction='out'/>"
"		</method>"
"		<method name='ProcExclude'>"
"			<arg type='s' name='Action' direction='in'/>"
"			<arg type='i' name='PID' direction='in'/>"
"		</method>"
/* Following functions are defined in proc-usage-stats.c */
"		<method name='ProcMemoryUsage'>"
"			<arg type='ai' name='ProcList' direction='in'/>"
"			<arg type='a(iiiiiii)' name='MemoryUsage' direction='out'/>"
"		</method>"
"		<method name='ProcCpuUsage'>"
"			<arg type='ai' name='ProcList' direction='in'/>"
"			<arg type='a(ii)' name='CpuUsage' direction='out'/>"
"		</method>"
"		<method name='CheckAppStatus'>"
"			<arg type='i' name='Pid' direction='in'/>"
"			<arg type='s' name='Type' direction='in'/>"
"			<arg type='i' name='Pid' direction='out'/>"
"			<arg type='i' name='Detect' direction='out'/>"
"			<arg type='s' name='Appid' direction='out'/>"
"		</method>"
"	</interface>"
"</node>";

static struct d_bus_method dbus_methods[] = {
	{ "GetAppCpu", dbus_get_app_cpu },
	{ "GetCpuList", dbus_get_cpu_list },
	{ "GetCpuLists", dbus_get_cpu_lists },
	{ "GetAppMemory", dbus_get_app_memory },
	{ "GetMemoryList", dbus_get_memory_list },
	{ "GetMemoryLists", dbus_get_memory_lists },
	{ "GetMemInfo", dbus_get_meminfo },
	{ "ReclaimMemory", dbus_reclaim_memory },
	{ "PrePoweroff", dbus_pre_poweroff },
	{ "ProcExclude", proc_dbus_exclude_method_handler},
/* Following functions are defined in proc-usage-stats.c */
	{ "ProcMemoryUsage", dbus_proc_memory_usage },
	{ "ProcCpuUsage", dbus_proc_cpu_usage },
	{ "CheckAppStatus", dbus_get_checkappstatus },
	/* Add methods here */
};

static const struct d_bus_signal dbus_signals[] = {
	/* RESOURCED DBUS */
	{RESOURCED_PATH_PROCESS, RESOURCED_INTERFACE_PROCESS,
	    SIGNAL_PROC_ACTIVE, proc_dbus_active_signal_handler, NULL},
	{RESOURCED_PATH_PROCESS, RESOURCED_INTERFACE_PROCESS,
	    SIGNAL_PROC_EXCLUDE, proc_dbus_exclude_signal_handler, NULL},
	{RESOURCED_PATH_PROCESS, RESOURCED_INTERFACE_PROCESS,
	    SIGNAL_PROC_PRELAUNCH, proc_dbus_prelaunch_signal_handler, NULL},
	{RESOURCED_PATH_PROCESS, RESOURCED_INTERFACE_PROCESS,
	    SIGNAL_PROC_SWEEP, proc_dbus_sweep_signal_handler, NULL},
	{RESOURCED_PATH_PROCESS, RESOURCED_INTERFACE_PROCESS,
	    SIGNAL_PROC_WATCHDOG, proc_dbus_watchdog_handler, NULL},
	{RESOURCED_PATH_PROCESS, RESOURCED_INTERFACE_PROCESS,
	    SIGNAL_PROC_SYSTEMSERVICE, proc_dbus_systemservice_handler, NULL},
	{RESOURCED_PATH_PROCESS, RESOURCED_INTERFACE_PROCESS,
	    SIGNAL_PROC_EXCLUDEAPPID, proc_dbus_exclude_appid_signal_handler, NULL},

	/* DEVICED DBUS */
	{DEVICED_PATH_DISPLAY, DEVICED_INTERFACE_DISPLAY,
	    SIGNAL_DEVICED_LCDON, proc_dbus_lcd_on, NULL},
	{DEVICED_PATH_DISPLAY, DEVICED_INTERFACE_DISPLAY,
	    SIGNAL_DEVICED_LCDOFF, proc_dbus_lcd_off, NULL},
	/* DEVICED_INTERFACE_POWEROFF / SIGNAL_DEVICED_CHANGESTATE left unhandled
	   on purpose - DeviceD sends the signal earlier than supposed to. Nothing
	   is lost though because at some point DeviceD also sends a SIGTERM,
	   which is handled the same way (i.e. quits the main loop). */
	{DEVICED_PATH_POWEROFF, DEVICED_INTERFACE_POWEROFF,
	    SIGNAL_DEVICED_POWEROFF_STATE, poweroff_signal_handler, NULL},
	{DEVICED_PATH_BATTERY, DEVICED_INTERFACE_BATTERY,
	    SIGNAL_DEVICED_LOW_BATTERY, low_battery_signal_handler, NULL},
	{DUMP_SERVICE_OBJECT_PATH, DUMP_SERVICE_INTERFACE_NAME,
	    SIGNAL_DUMP, proc_dbus_dump_handler, NULL},
	{DEVICED_PATH_TIME, DEVICED_INTERFACE_TIME,
	    SIGNAL_DEVICED_SYSTEMTIME_CHANGED,
	    systemtime_changed_signal_handler, NULL},

	/* AMD DBUS */
	{AUL_APPSTATUS_OBJECT_PATH, AUL_APPSTATUS_INTERFACE_NAME,
	    SIGNAL_AMD_LAUNCH, proc_dbus_aul_launch, NULL},
	{AUL_APPSTATUS_OBJECT_PATH, AUL_APPSTATUS_INTERFACE_NAME,
	    SIGNAL_AMD_RESUME, proc_dbus_aul_resume, NULL},
	{AUL_APPSTATUS_OBJECT_PATH, AUL_APPSTATUS_INTERFACE_NAME,
	    SIGNAL_AMD_TERMINATE, proc_dbus_aul_terminate, NULL},
	{AUL_APPSTATUS_OBJECT_PATH, AUL_APPSTATUS_INTERFACE_NAME,
	    SIGNAL_AMD_STATE, proc_dbus_aul_changestate, NULL},
	{AUL_APPSTATUS_OBJECT_PATH, AUL_APPSTATUS_INTERFACE_NAME,
	    SIGNAL_AMD_GROUP, proc_dbus_aul_group, NULL},
	{AUL_APPSTATUS_OBJECT_PATH, AUL_APPSTATUS_INTERFACE_NAME,
	    SIGNAL_AMD_TERMINATED, proc_dbus_aul_terminated, NULL},
	{AUL_SUSPEND_OBJECT_PATH, AUL_SUSPEND_INTERFACE_NAME,
	    SIGNAL_AMD_SUSPNED, proc_dbus_suspend_hint, NULL},

	/* BOOTING DBUS */
	{BOOTING_DONE_PATH, BOOTING_DONE_INTERFACE,
	    SIGNAL_BOOTING_DONE, booting_done_signal_handler, NULL},
};

static int proc_dbus_init(void *data)
{
	resourced_ret_c ret;

	/* start watchdog check timer for preveting ANR during booting */
	watchdog_check_timer = g_timeout_source_new_seconds(WATCHDOG_TIMER_INTERVAL);
	g_source_set_callback(watchdog_check_timer, check_watchdog_cb, NULL, NULL);
	g_source_attach(watchdog_check_timer, NULL);

	ret = d_bus_register_signals(dbus_signals, ARRAY_SIZE(dbus_signals));
	if (ret != RESOURCED_ERROR_NONE)
		goto out;

	ret = d_bus_register_methods(RESOURCED_PATH_PROCESS, dbus_methods_xml,
			dbus_methods, ARRAY_SIZE(dbus_methods));
out:
	return ret;
}

static int proc_dbus_exit(void *data)
{
	if (watchdog_check_timer)
		g_source_destroy(watchdog_check_timer);
	return RESOURCED_ERROR_NONE;
}

static const struct proc_module_ops proc_dbus_ops = {
	.name		= "PROC_DBUS",
	.init		= proc_dbus_init,
	.exit		= proc_dbus_exit,
};
PROC_MODULE_REGISTER(&proc_dbus_ops)
