/*
 * resourced
 *
 * Copyright (c) 2000 - 2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/

#include <ctype.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/types.h>

#include "freezer.h"
#include "resourced.h"
#include "trace.h"
#include "proc-main.h"
#include "cgroup.h"
#include "proc-process.h"
#include "procfs.h"
#include "memory-common.h"
#include "macro.h"
#include "notifier.h"
#include "proc-appusage.h"

#define PROC_SWEEP_TIMER	3
static GHashTable *proc_sweep_list;
static GSource *proc_sweep_timer;

enum proc_background_type {
	PROC_BACKGROUND_INACTIVE,
	PROC_BACKGROUND_ACTIVE,
};

int proc_set_service_oomscore(const pid_t pid, const int oom_score)
{
	int service_oom;
	if (oom_score > 0 && oom_score != OOMADJ_SERVICE_DEFAULT)
		service_oom = oom_score - OOMADJ_SERVICE_GAP;
	else
		service_oom = OOMADJ_SERVICE_DEFAULT;
	return proc_set_oom_score_adj(pid, service_oom);
}

static void proc_set_oom_score_childs(GSList *childs, int oom_score_adj)
{
	GSList *iter;

	if (!childs)
		return;

	gslist_for_each_item(iter, childs) {
		proc_set_oom_score_adj(GPOINTER_TO_PID(iter->data), oom_score_adj);
	}
}

static void proc_set_oom_score_services(int state, GSList *svcs,
		int oom_score_adj)
{
	GSList *iter;

	if (!svcs)
		return;

	gslist_for_each_item(iter, svcs) {
		struct proc_app_info *svc = (struct proc_app_info *)(iter->data);
		svc->state = state;
		proc_set_service_oomscore(svc->main_pid, oom_score_adj);
	}
}

static int proc_get_lowest_oom_score(GSList *uiapps)
{
	GSList *iter;
	int min_oom_score = OOMADJ_APP_MAX;
	int oom_score_adj = 0, ret;

	if (!uiapps)
		return OOMADJ_SU;

	gslist_for_each_item(iter, uiapps) {
		struct proc_app_info *pai = (struct proc_app_info *)(iter->data);

		ret = proc_get_oom_score_adj(pai->main_pid, &oom_score_adj);
		if (ret)
			continue;
		min_oom_score = (oom_score_adj < min_oom_score) ? oom_score_adj : min_oom_score;
	}

	return min_oom_score;
}

static int proc_backgrd_manage(int currentpid, int active, int oom_score_adj)
{
	pid_t pid = -1;
	int flag = RESOURCED_NOTIFIER_APP_BACKGRD;
	struct proc_status ps;
	FILE *fp;
	char buf[sizeof(PROC_OOM_SCORE_ADJ_PATH) + MAX_DEC_SIZE(int)] = {0};
	int cur_oom = -1;
	static int checkprevpid;
	int freeze_val = 0;
	GSList *iter;
	struct proc_program_info *ppi;
	struct proc_app_info *pai = find_app_info(currentpid);
	int is_favorite = 0;

	if (proc_get_freezer_status() != CGROUP_FREEZER_DISABLED)
		freeze_val = resourced_freezer_proc_late_control();

	if (!pai) {
		_E("can't find app info about pid %d", currentpid);
		return RESOURCED_ERROR_INVALID_PARAMETER;
	}

	/*
	 * About groupd process with multiple applications,
	 * all application with same group could be went to background state.
	 * If one application has already managed to background application
	 * it skipped to go to the background again.
	 */
	if (pai->lru_state >= PROC_BACKGROUND) {
		_D("pid %d already in background", currentpid);
		return RESOURCED_ERROR_NONE;
	}

	ps.pid = currentpid;
	ps.pai = pai;
	if (active)
		flag = RESOURCED_NOTIFIER_APP_BACKGRD_ACTIVE;
	resourced_notify(flag, &ps);

	if (active)
		goto set_oom;

	pai->lru_state = PROC_BACKGROUND;
	if (proc_check_favorite_app(pai->appid)) {
		_D("detect favorite application : %s", pai->appid);
		oom_score_adj = OOMADJ_FAVORITE;
		pai->lru_state = PROC_FAVORITE;
		is_favorite = 1;
	}

	if (checkprevpid != currentpid) {
		_cleanup_app_list_close_ GSList *proc_app_list = PAL_INIT_VALUE;

		proc_app_list = proc_app_list_open();
		gslist_for_each_item(iter, proc_app_list) {
			struct proc_app_info *spi = (struct proc_app_info *)iter->data;
			int new_oom;
			int lru_offset = freeze_val;

			if (!spi->main_pid || spi->type != PROC_TYPE_GUI)
				continue;

			/*
			 * It checks all applications from proc_app_list
			 * So, current background process can be checked again.
			 * It is because lru value of this application was already changed,
			 * it should be skipped
			 */

			if (spi == pai)
				continue;

			pid = spi->main_pid;
			snprintf(buf, sizeof(buf), PROC_OOM_SCORE_ADJ_PATH, pid);
			fp = fopen(buf, "r+");
			if (fp == NULL) {
				spi->main_pid = 0;
				continue;
			}
			if (fgets(buf, sizeof(buf), fp) == NULL) {
				fclose(fp);
				spi->main_pid = 0;
				continue;
			}
			cur_oom = atoi(buf);

			if (spi->lru_state == PROC_BACKGROUND ||
					spi->lru_state == PROC_FAVORITE) {
				memset(&ps, 0, sizeof(struct proc_status));
				ps.pai = spi;
				ps.pid = pid;
				resourced_notify(
					    RESOURCED_NOTIFIER_APP_SUSPEND_READY,
					    &ps);
			}
			/*
			 * clear lru offset if platform controls background application
			 */
			if (spi->flags & PROC_BGCTRL_PLATFORM)
				lru_offset = 0;

			if (proc_check_lru_suspend(lru_offset, spi->lru_state) &&
			    (proc_check_suspend_state(spi) == PROC_STATE_SUSPEND)) {
				memset(&ps, 0, sizeof(struct proc_status));
				ps.pai = spi;
				ps.pid = pid;
				resourced_notify(
					    RESOURCED_NOTIFIER_APP_SUSPEND,
					    &ps);
			}

			if (is_favorite) {
				if (spi->lru_state >= PROC_FAVORITE &&
						spi->lru_state < PROC_FAVORITE_LRU_MAX) {
					spi->lru_state++;
					_D("FAVORITE : process %d increase lru %d",
							pid, spi->lru_state - PROC_FAVORITE + 1);
				}

				if (cur_oom >= OOMADJ_FAVORITE &&
						cur_oom < OOMADJ_FAVORITE_APP_MAX) {
					new_oom = cur_oom + OOMADJ_FAVORITE_APP_INCREASE;
					_D("FAVORITE : process %d set score %d (beford %d)",
							pid, new_oom, cur_oom);
					proc_set_oom_score_adj(pid, new_oom);
					proc_set_oom_score_childs(spi->childs, new_oom);
				}
			} else {
				if (spi->lru_state >= PROC_BACKGROUND &&
						spi->lru_state < PROC_LRU_MAX) {
					spi->lru_state++;
					_D("BACKGRD : process %d increase lru %d", pid, spi->lru_state);
				}

				if (cur_oom >= OOMADJ_APP_MAX) {
					fclose(fp);
					continue;
				} else if (cur_oom >= OOMADJ_BACKGRD_UNLOCKED) {
					new_oom = cur_oom + OOMADJ_APP_INCREASE;
					_D("BACKGRD : process %d set score %d (before %d)",
							pid, new_oom, cur_oom);
					proc_set_oom_score_adj(pid, new_oom);
					proc_set_oom_score_childs(spi->childs, new_oom);
				}
			}
			fclose(fp);
		}
	}

set_oom:
	proc_set_oom_score_adj(pai->main_pid, oom_score_adj);

	/* change oom score about child pids */
	proc_set_oom_score_childs(pai->childs, oom_score_adj);

	/* change oom score about grouped service processes */
	ppi = pai->program;
	if (ppi && proc_get_svc_state(ppi) == PROC_STATE_BACKGROUND)
		proc_set_oom_score_services(PROC_STATE_BACKGROUND, ppi->svc_list,
			(proc_get_lowest_oom_score(ppi->app_list) >= oom_score_adj) ? OOMADJ_BACKGRD_UNLOCKED : OOMADJ_SERVICE_DEFAULT);

	checkprevpid = currentpid;
	return RESOURCED_ERROR_NONE;
}

static int proc_foregrd_manage(int pid, int oom_score_adj)
{
	int ret = 0;
	struct proc_program_info *ppi;
	struct proc_app_info *pai;

	pai = find_app_info(pid);
	if (!pai) {
		proc_set_oom_score_adj(pid, oom_score_adj);
		return RESOURCED_ERROR_NO_DATA;
	}

	proc_set_oom_score_adj(pai->main_pid, oom_score_adj);

	/* change oom score about child pids */
	proc_set_oom_score_childs(pai->childs, oom_score_adj);

	pai->lru_state = PROC_FOREGROUND;

	/* change oom score about grouped service processes */
	ppi = pai->program;
	if (ppi)
		proc_set_oom_score_services(PROC_STATE_FOREGROUND, ppi->svc_list,
		    oom_score_adj);

	return ret;
}

void proc_kill_victiom(gpointer key, gpointer value, gpointer user_data)
{
	int pid = *(gint *)key;
	int cur_oom = -1;
	char buf[sizeof(PROC_OOM_SCORE_ADJ_PATH) + MAX_DEC_SIZE(int)] = {0};
	FILE *fp;

	snprintf(buf, sizeof(buf), PROC_OOM_SCORE_ADJ_PATH, pid);
	fp = fopen(buf, "r+");
	if (fp == NULL) {
		_D("sweep proc_kill_victim : background process %d already terminated", pid);
		return;
	}
	if (fgets(buf, sizeof(buf), fp) == NULL) {
		fclose(fp);
		return;
	}
	cur_oom = atoi(buf);
	if (cur_oom >= OOMADJ_FAVORITE) {
		kill(pid, SIGKILL);
		_D("sweep memory : background process %d killed by sigkill", pid);
	}
	fclose(fp);
}

static gboolean proc_check_sweep_cb(gpointer data)
{
	GHashTable *List = (GHashTable *)data;
	g_hash_table_foreach(List, proc_kill_victiom, NULL);
	g_hash_table_destroy(List);
	proc_sweep_list = NULL;

	return false;
}

int proc_sweep_memory(enum proc_sweep_type type, pid_t callpid)
{
	_cleanup_app_list_close_ GSList *proc_app_list = PAL_INIT_VALUE;
	pid_t pid = -1;
	int count = 0, ret;
	FILE *fp;
	gint *piddata;
	char buf[sizeof(PROC_OOM_SCORE_ADJ_PATH) + MAX_DEC_SIZE(int)] = {0};
	char appname[PROC_NAME_MAX];

	int cur_oom = -1;
	int select_sweep_limit;
	GSList *iter;
	struct proc_app_info *pai;

	if (proc_sweep_timer)
		g_source_destroy(proc_sweep_timer);
	if (proc_sweep_list)
		g_hash_table_destroy(proc_sweep_list);
	proc_sweep_list = g_hash_table_new(g_int_hash, g_int_equal);

	if (type == PROC_SWEEP_EXCLUDE_ACTIVE)
		select_sweep_limit = OOMADJ_FAVORITE;
	else
		select_sweep_limit = OOMADJ_BACKGRD_LOCKED;

	proc_app_list = proc_app_list_open();
	gslist_for_each_item(iter, proc_app_list) {
		pai = (struct proc_app_info *)iter->data;
		if (!pai->main_pid || pai->type != PROC_TYPE_GUI)
			continue;

		pid = pai->main_pid;

		snprintf(buf, sizeof(buf), PROC_OOM_SCORE_ADJ_PATH, pid);
		fp = fopen(buf, "r+");
		if (fp == NULL)
			continue;
		if (fgets(buf, sizeof(buf), fp) == NULL) {
			fclose(fp);
			continue;
		}
		cur_oom = atoi(buf);
		if (cur_oom >= select_sweep_limit) {
			ret = proc_get_cmdline(pid, appname);
			if (ret != 0) {
				fclose(fp);
				continue;
			}
			resourced_proc_status_change(PROC_CGROUP_SET_TERMINATE_REQUEST,
				    pid, NULL, NULL, PROC_TYPE_NONE);
			piddata = g_new(gint, 1);
			*piddata = pid;
			g_hash_table_insert(proc_sweep_list, piddata, NULL);
			kill(pid, SIGTERM);
			_D("sweep memory : background process %d(%s) killed",
					pid, appname);
			count++;
		}
		fclose(fp);
	}
	if (count > 0) {
		proc_sweep_timer = g_timeout_source_new_seconds(PROC_SWEEP_TIMER);
		g_source_set_callback(proc_sweep_timer, proc_check_sweep_cb, (gpointer)proc_sweep_list, NULL);
		g_source_attach(proc_sweep_timer, NULL);
	}

	return count;
}

int proc_set_foregrd(pid_t pid, int oom_score_adj)
{
	int ret = 0;

	switch (oom_score_adj) {
	case OOMADJ_FOREGRD_UNLOCKED:
	case OOMADJ_SU:
		ret = 0;
		break;
	case OOMADJ_FOREGRD_LOCKED:
	case OOMADJ_BACKGRD_LOCKED:
		ret = proc_foregrd_manage(pid, OOMADJ_FOREGRD_LOCKED);
		break;
	case OOMADJ_BACKGRD_UNLOCKED:
	case OOMADJ_INIT:
		ret = proc_foregrd_manage(pid, OOMADJ_FOREGRD_UNLOCKED);
		break;
	default:
		if (oom_score_adj > OOMADJ_BACKGRD_UNLOCKED ||
				(oom_score_adj >= OOMADJ_FAVORITE &&
				 oom_score_adj <= OOMADJ_FAVORITE_APP_MAX))
			ret = proc_foregrd_manage(pid, OOMADJ_FOREGRD_UNLOCKED);
		else
			ret = -1;
		break;

	}
	return ret;
}

int proc_set_backgrd(int pid, int oom_score_adj)
{
	int ret = 0;

	switch (oom_score_adj) {
	case OOMADJ_BACKGRD_LOCKED:
	case OOMADJ_BACKGRD_UNLOCKED:
	case OOMADJ_SU:
		ret = -1;
		break;
	case OOMADJ_FOREGRD_LOCKED:
		ret = proc_backgrd_manage(pid, PROC_BACKGROUND_ACTIVE, OOMADJ_BACKGRD_LOCKED);
		break;
	case OOMADJ_FOREGRD_UNLOCKED:
	case OOMADJ_INIT:
		ret = proc_backgrd_manage(pid, PROC_BACKGROUND_INACTIVE, OOMADJ_BACKGRD_UNLOCKED);
		break;
	default:
		if (oom_score_adj > OOMADJ_BACKGRD_UNLOCKED)
			ret = 0;
		else
			ret = -1;
		break;
	}
	return ret;
}

int proc_set_active(int pid, int oom_score_adj)
{
	int ret = 0;
	struct proc_app_info *pai;

	switch (oom_score_adj) {
	case OOMADJ_FOREGRD_LOCKED:
	case OOMADJ_BACKGRD_LOCKED:
	case OOMADJ_SU:
		/* don't change oom value pid */
		ret = -1;
		break;
	case OOMADJ_INIT:
	case OOMADJ_FOREGRD_UNLOCKED:
		ret = proc_set_oom_score_adj(pid, OOMADJ_FOREGRD_LOCKED);
		break;
	case OOMADJ_BACKGRD_UNLOCKED:
		pai = find_app_info(pid);
		if (pai)
			pai->lru_state = PROC_ACTIVE;
		ret = proc_set_oom_score_adj(pid, OOMADJ_BACKGRD_LOCKED);
		break;
	case OOMADJ_PREVIOUS_BACKGRD:
		ret = proc_set_oom_score_adj(pid, OOMADJ_PREVIOUS_DEFAULT);
		break;
	default:
		if (oom_score_adj > OOMADJ_BACKGRD_UNLOCKED) {
			pai = find_app_info(pid);
			if (pai)
				pai->lru_state = PROC_ACTIVE;
			ret = proc_set_oom_score_adj(pid, OOMADJ_BACKGRD_LOCKED);
		} else
			ret = -1;
		break;
	}
	return ret;
}

int proc_set_inactive(int pid, int oom_score_adj)
{
	int ret = 0;
	struct proc_app_info *pai;
	switch (oom_score_adj) {
	case OOMADJ_FOREGRD_UNLOCKED:
	case OOMADJ_BACKGRD_UNLOCKED:
	case OOMADJ_SU:
	case OOMADJ_INIT:
		/* don't change oom value pid */
		ret = -1;
		break;
	case OOMADJ_FOREGRD_LOCKED:
		ret = proc_set_oom_score_adj(pid, OOMADJ_FOREGRD_UNLOCKED);
		break;
	case OOMADJ_BACKGRD_LOCKED:
		pai = find_app_info(pid);
		if (pai) {
			struct proc_status ps = {0};
			ps.pid = pid;
			ps.pai = pai;
			pai->lru_state = PROC_BACKGROUND;
			ret = proc_set_oom_score_adj(pid, OOMADJ_BACKGRD_UNLOCKED);
			resourced_notify(RESOURCED_NOTIFIER_APP_BACKGRD, &ps);
			struct proc_program_info *const ppi = pai->program;
			if (ppi && proc_get_svc_state(ppi) == PROC_STATE_BACKGROUND)
				proc_set_oom_score_services(PROC_STATE_BACKGROUND, ppi->svc_list,
					(proc_get_lowest_oom_score(ppi->app_list) >= oom_score_adj) ? OOMADJ_BACKGRD_UNLOCKED : OOMADJ_SERVICE_DEFAULT);
		}
		break;
	default:
		if (oom_score_adj > OOMADJ_BACKGRD_UNLOCKED)
			ret = 0;
		else
			ret = -1;
		break;

	}
	return ret;
}
