/*
 * resourced
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "macro.h"
#include "module.h"
#include "module-data.h"
#include "trace.h"
#include "util.h"
#include "file-helper.h"
#include "procfs.h"
#include "swap-common.h"
#include "memory-common.h"
#include "config-parser.h"
#include "lowmem-handler.h"
#include "losetup.h"

#define FILESWAP_FULLNESS_RATIO	0.8
#define DEFAULT_SWAP_FILE_SIZE MBYTE_TO_BYTE(30)

struct swap_file_control {
	char crypt_type[MAX_TYPE_LENGTH];
	char swapfile[64];
	unsigned long swap_file_size;
	unsigned long swap_reclaim_bytes;
};

static struct swap_file_control file_control = {
	.crypt_type = "aes",
	.swapfile = SWAP_FILE_NAME,
	.swap_file_size = DEFAULT_SWAP_FILE_SIZE,
	.swap_reclaim_bytes = 0,
};

static int swap_file_activate(void *data)
{
	struct swap_module_ops *swap = (struct swap_module_ops *)data;

	file_control.swap_reclaim_bytes =
			file_control.swap_file_size * FILESWAP_FULLNESS_RATIO;

	return swap_set_file(file_control.swapfile, swap, file_control.crypt_type);
}

static int swap_file_reclaim(void *data)
{
	struct swap_module_ops *swap = (struct swap_module_ops *)data;
	_cleanup_proc_swaps_free_ struct proc_swaps **swaps = NULL;
	unsigned int swap_size = 0;
	int n, i;

	n = proc_get_swaps(&swaps);
	if (n <= 0)
		return -ENOENT;

	for (i = 0; i < n; i++) {
		if (streq(swaps[i]->filename, swap->path)) {
			swap_size = swaps[i]->used;
			break;
		}
	}

	swap_size = KBYTE_TO_BYTE(swap_size);
	if (swap_size < file_control.swap_reclaim_bytes)
		return 0;

	/*
	 * if current swap has multiple devices,
	 * it is no necessary to trigger reclaiming
	 * because there is no impact to save memory
	 * in spite of killing victims
	 */
	if (n > 1)
		return 0;

	/*
	 * In this case, swap storage is almost full
	 * when using only storage for swap.
	 * It means that there are many background processes or
	 * some process makes memory leak.
	 * So, it requires to trigger proactive oom killer
	 * with LMK_MEMORY type.
	 */
	lowmem_trigger_swap_reclaim(LMK_MEMORY, swap_size);
	return -ENOSPC;
}

static int swap_file_parse_config_file(void)
{
	ConfigTableItem items[] = {
		{ "FILE",		"CryptType",		config_parse_string,	0,	NULL	},
		{ "FILE",		"FileSize",			config_parse_bytes,	0,	NULL	},
		{ NULL,		NULL,			NULL,				0,	NULL	}
	};

	int r;
	_cleanup_free_ char *crypt_type = NULL;

	items[0].data = &crypt_type;
	items[1].data = &file_control.swap_file_size;

	r = config_parse_new(SWAP_CONF_FILE, (void*) items);
	if (r < 0) {
		_E("Failed to parse configuration file: %s", strerror(-r));
		return r;
	}

	if (check_valid_compressor(crypt_type) == RESOURCED_ERROR_NONE)
		strncpy(file_control.crypt_type, crypt_type, MAX_TYPE_LENGTH-1);
	else
		memset(file_control.crypt_type, 0, MAX_TYPE_LENGTH);

	return 0;
}

static int swap_file_init(void *data)
{
	struct swap_module_ops *swap = (struct swap_module_ops *)data;
	int r;

	r = swap_file_parse_config_file();
	if (r < 0) {
		_E("Failed to parse SwapFile config: %s", strerror(-r));
		return r;
	}

	swap->k_size = BYTE_TO_KBYTE(file_control.swap_file_size);
	return 0;
}

static struct swap_module_ops swap_file_ops = {
	.name = "FILESWAP",
	.type = SWAP_TYPE_FILE,
	.path = "",
	.priority = SWAP_PRI_LOW,
	.k_size = 0,
	.init = swap_file_init,
	.activate = swap_file_activate,
	.reclaim = swap_file_reclaim,
};
SWAP_MODULE_REGISTER(&swap_file_ops)
