/*
 * resourced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/

/*
 * @file swap.c
 * @desc swap process
 */
#include <trace.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <pthread.h>
#include <sys/sysinfo.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <memory-common.h>
#include <sys/swap.h>
#include <linux/loop.h>
#include <fcntl.h>

#include "macro.h"
#include "module.h"
#include "module-data.h"
#include "dbus-handler.h"
#include "swap-common.h"
#include "config-parser.h"
#include "lowmem-handler.h"
#include "notifier.h"
#include "procfs.h"
#include "cgroup.h"
#include "const.h"
#include "file-helper.h"
#include "proc-common.h"
#include "proc-main.h"
#include "util.h"
#include "losetup.h"
#include "init.h"

#define MEMCG_PATH			"/sys/fs/cgroup/memory/"
#define MEMCG_SIZE_LIMIT		"memory.limit_in_bytes"
#define SWAPCG_RECLAIM			"memory.force_reclaim"
#define MOVE_CHARGE			"memory.move_charge_at_immigrate"

#define SWAP_PRIORITY			20
#define SWAP_HARD_LIMIT_DEFAULT		0.5
#define SWAP_FORCE_RECLAIM_NUM_MAX	5
#define SWAP_RECLAIM_PAGES_MAX		2560
#define SWAP_RECLAIM_PAGES_MIN		128
#define SWAP_MEMCG_SWAPPINESS		60
#define SWAP_MIN_SWAPPINESS		0
#define SWAP_EARLYRECLAIM_TIME_DEFAULT	60

#define EARLYRECLAIM_WITH_AN_EXPLANATION_FOR_LAYMEN "early memory reclaim (done to retrieve resources used by daemons during system start-up)"

enum swap_thread_op {
	SWAP_OP_ACTIVATE,
	SWAP_OP_RECLAIM,
	SWAP_OP_COMPACT,
	SWAP_OP_MOVE_TO_SWAP_AND_RECLAIM,
	SWAP_OP_END,
};

enum swap_reclaim_node {
	SWAP_NODE_HARD_LIMIT,
	SWAP_NODE_FORCE_RECLAIM,
	SWAP_NODE_END,
};

struct swap_task {
	struct proc_app_info *pai;
	int size;
};

struct swap_safe_queue {
	GQueue *queue;
	pthread_mutex_t lock;
};

struct swap_thread_bundle {
	struct swap_status_msg msg;
	enum swap_thread_op op;
};

static pthread_mutex_t swap_mutex;
static pthread_cond_t swap_cond;
static struct swap_safe_queue swap_thread_queue;
static struct module_ops swap_modules_ops;
static float swap_hard_limit_fraction = SWAP_HARD_LIMIT_DEFAULT;
static enum swap_reclaim_node swap_node;
static enum swap_state swap_state;

static bool arg_swap_enable = false;
static bool arg_swap_at_boot = false;
static int arg_timer_swap_at_boot = SWAP_EARLYRECLAIM_TIME_DEFAULT;
static enum swap_type arg_swap_type = SWAP_TYPE_ZRAM;
static int current_swappiness = SWAP_MEMCG_SWAPPINESS;

static GSList *swap_module;  /* module list */
static GSource *swap_activating_timer = NULL;

static int swap_sort_func(const struct swap_module_ops *a,
					    const struct swap_module_ops *b)
{
	return (a->priority - b->priority);
}

void swap_add(const struct swap_module_ops *ops)
{
	swap_module = g_slist_insert_sorted(swap_module,
					(gpointer)ops,
					(GCompareFunc) swap_sort_func);
}

void swap_remove(const struct swap_module_ops *ops)
{
	swap_module = g_slist_remove(swap_module, (gpointer)ops);
}

int do_mkswap(const char *device)
{
	const char *argv[3] = { "/sbin/mkswap", NULL, NULL };

	argv[1] = (char *) device;

	return exec_cmd(ARRAY_SIZE(argv), argv);
}

int do_dd(char *input, char *output, unsigned int size, unsigned int count)
{
	const char *argv[6] = { "/bin/dd", NULL, NULL, NULL, NULL, NULL };
	_cleanup_free_ char *if_arg = NULL, *of_arg = NULL;
	_cleanup_free_ char *bs_arg = NULL, *count_arg = NULL;

	if (asprintf(&if_arg, "if=%s", input) < 0 ||
	    asprintf(&of_arg, "of=%s", output) < 0 ||
	    asprintf(&bs_arg, "bs=%u", size) < 0 ||
	    asprintf(&count_arg, "count=%u", count) < 0)
		return -ENOMEM;

	argv[1] = if_arg;
	argv[2] = of_arg;
	argv[3] = bs_arg;
	argv[4] = count_arg;

	return exec_cmd(ARRAY_SIZE(argv), argv);
}

static const char *compact_reason_to_str(enum swap_compact_reason reason)
{
	static const char *reasons_table[] = {"lowmem: low", "lowmem: medium",
			"swap: zram full"};
	if (reason >= SWAP_COMPACT_LOWMEM_LOW && reason < SWAP_COMPACT_RESASON_MAX)
		return reasons_table[reason];
	return "";
}

enum swap_state swap_get_state(void)
{
	return swap_state;
}

static void swap_set_state(enum swap_state state)
{
	if ((state != SWAP_ON) && (state != SWAP_OFF))
		return;

	swap_state = state;
}

static unsigned int swap_calculate_hard_limit_in_bytes(unsigned int mem_subcg_usage)
{
	return (unsigned int)((float)mem_subcg_usage * swap_hard_limit_fraction);
}

static inline void swap_add_bundle(struct swap_thread_bundle *bundle)
{
	pthread_mutex_lock(&swap_thread_queue.lock);
	g_queue_push_tail(swap_thread_queue.queue, bundle);
	pthread_mutex_unlock(&swap_thread_queue.lock);
}

static int swap_move_to_cgroup_by_pid(enum memcg_type type, pid_t pid)
{
	int ret;
	struct memcg *memcg_swap = NULL;
	struct memcg_info *mi;
	struct proc_app_info *pai = find_app_info(pid);
	GSList *iter_child = NULL;

	ret = lowmem_get_memcg(type, &memcg_swap);
	if (ret != RESOURCED_ERROR_NONE)
		return RESOURCED_ERROR_FAIL;

	mi = memcg_swap->info;
	if (!pai)
		return cgroup_write_pid_fullpath(mi->name, pid);

	ret = cgroup_write_pid_fullpath(mi->name, pai->main_pid);
	gslist_for_each_item(iter_child, pai->childs) {
		pid_t child = GPOINTER_TO_PID(iter_child->data);
		ret = cgroup_write_pid_fullpath(mi->name, child);
	}
	pai->memory.memcg_idx = MEMCG_SWAP;
	pai->memory.memcg_info = mi;
	return ret;
}

static int swap_move_to_cgroup(struct memcg_info *info, GArray *candidates)
{
	int index;
	struct swap_task tsk;
	struct proc_app_info *pai = NULL;
	GSList *iter_child = NULL;

	if (!candidates)
		return RESOURCED_ERROR_NO_DATA;

	for (index = 0; index < candidates->len; index++) {
		tsk = g_array_index(candidates, struct swap_task, index);
		pai = tsk.pai;
		cgroup_write_pid_fullpath(info->name, pai->main_pid);
		gslist_for_each_item(iter_child, pai->childs) {
			pid_t child = GPOINTER_TO_PID(iter_child->data);
			cgroup_write_pid_fullpath(info->name, child);
		}
		pai->memory.memcg_idx = MEMCG_SWAP;
		pai->memory.memcg_info = info;
	}
	return RESOURCED_ERROR_NONE;
}

static int swap_sort_by_oom(const struct swap_task *ta,
		const struct swap_task *tb)
{
	/* sort by oom score adj */
	assert(ta != NULL);
	assert(tb != NULL);

	return ((int)(tb->pai->memory.oom_score_adj) -
		(int)(ta->pai->memory.oom_score_adj));
}

static int swap_sort_by_vmrss(const struct swap_task *ta,
		const struct swap_task *tb)
{
	/* sort by task memory usage */
	assert(ta != NULL);
	assert(tb != NULL);

	return ((int)(tb->size) - (int)(ta->size));
}

static int swap_reduce_victims(GArray *candidates, int max)
{
	int index;
	struct swap_task tsk;
	struct proc_app_info *pai = NULL;
	unsigned int usage = 0;

	if (!candidates)
		return RESOURCED_ERROR_NO_DATA;

	for (index = 0; index < candidates->len; index++) {
		tsk = g_array_index(candidates, struct swap_task, index);
		pai = tsk.pai;

		/* Measuring VmRSS is OK as it's anonymous + swapcache */
		if (proc_get_approx_mem_usage(pai->main_pid, &usage) < 0)
			continue;

		tsk.size += usage;

		if (pai->childs) {
			GSList *iter_child = NULL;

			gslist_for_each_item(iter_child, pai->childs) {
				pid_t child = GPOINTER_TO_PID(iter_child->data);
				if (proc_get_approx_mem_usage(child, &usage) < 0)
					continue;
				tsk.size += usage;
			}
		}
	}
	/* sort by oom_score_adj value, older are better candidates */
	g_array_sort(candidates, (GCompareFunc)swap_sort_by_oom);

	/* sort by memory usage, swapping bigger will free more memory */
	g_array_sort(candidates, (GCompareFunc)swap_sort_by_vmrss);

	/* limit the number of potential candidates, after sort by oom */
	g_array_remove_range(candidates, max, candidates->len - max);

	return RESOURCED_ERROR_NONE;
}

static int swap_use_hard_limit(char *memcg)
{
	int ret;
	unsigned int usage, memcg_limit;

	ret = cgroup_read_node_uint32(memcg, MEMCG_USAGE, &usage);
	if (ret != RESOURCED_ERROR_NONE)
		usage = 0;

	memcg_limit = swap_calculate_hard_limit_in_bytes(usage);
	_D("Swap request: %s cgroup usage is %u, hard limit set to %u (hard limit fraction %f)",
			memcg, usage, memcg_limit, swap_hard_limit_fraction);
	if (memcg_limit != 0)
		ret = cgroup_write_node_uint32(memcg, MEMCG_SIZE_LIMIT, memcg_limit);
	else
		/* If the group is empty don't set the limit to enable adding processes. */
		ret = cgroup_write_node_int32(memcg, MEMCG_SIZE_LIMIT, -1);

	if (ret != RESOURCED_ERROR_NONE)
		_E("Not able to set hard limit of %s memory cgroup", memcg);

	return ret;
}

static int swap_use_force_reclaim(char *memcg)
{
	int ret, len;
	int try = SWAP_FORCE_RECLAIM_NUM_MAX;
	unsigned int usage, nr_to_reclaim;
	unsigned int total_reclaim = 0;
	bool root_memcg = false;

	/*
	 * In case of LOWMEM_ROOT_CGROUP,
	 * it is no necessary to check anon usage of memcg
	 * because it doesn't include all memory usage.
	 * But instead, try to swap memory as much as possible.
	 * It will be happend only once after booting done.
	 */
	len = strlen(memcg);
	if (!strncmp(memcg, LOWMEM_ROOT_CGROUP, len))
		root_memcg = true;

	do {
		/*
		 * Currently, we only move only anonymous pages to swap memcg by
		 * setting move_charge_at_immigrate as 0. However, there might
		 * be a little of inactive file pages in swap memcg.
		 * For this reason it's better to use '.stat' and calculate only
		 * anoynymous memory usage.
		 */
		if (root_memcg)
			ret = cgroup_read_node_uint32(memcg, MEMCG_USAGE, &usage);
		else
			ret = memcg_get_anon_usage(memcg, &usage);

		if (ret != RESOURCED_ERROR_NONE)
			usage = 0;

		nr_to_reclaim = BYTE_TO_PAGE(usage);
		if (nr_to_reclaim <= SWAP_RECLAIM_PAGES_MIN)
			break; /* don't reclaim if little gain */
		if (nr_to_reclaim > SWAP_RECLAIM_PAGES_MAX)
			nr_to_reclaim = SWAP_RECLAIM_PAGES_MAX;

		total_reclaim += nr_to_reclaim;
		ret = cgroup_write_node_uint32(memcg, SWAPCG_RECLAIM,
			nr_to_reclaim);
		if (ret != RESOURCED_ERROR_NONE)
			break; /* if we can't reclaim don't continue */

		try -= 1;
	} while (try > 0);

	_D("FORCE_RECLAIM tried %u pages from %s", total_reclaim, memcg);
	return ret;
}

static int swap_start_reclaim(char *memcg)
{
	int r;
	int swappiness;

	r = cgroup_read_node_int32(memcg, MEMCG_SWAPNESS_PATH,
			&swappiness);
	if (r)
		return r;

	/*
	 * Resourced sometimes changes swappiness value.
	 * But, this function means to reclaim memory by force as much as possible
	 * If current swappiness value is smaller than default swappiness,
	 * it is better to set default value before reclaiming memory.
	 * And current value will be set again after reclaim.
	 */
	if (swappiness != SWAP_MEMCG_SWAPPINESS) {
		r = cgroup_write_node_uint32(memcg,
			MEMCG_SWAPNESS_PATH, SWAP_MEMCG_SWAPPINESS);
		if (r) {
			_I("failed to write %s %d to %s the",
				MEMCG_SWAPNESS_PATH, SWAP_MEMCG_SWAPPINESS, memcg);
			return -errno;
		}
	}
	if (swap_node == SWAP_NODE_FORCE_RECLAIM)
		swap_use_force_reclaim(memcg);
	else
		swap_use_hard_limit(memcg);

	/*
	 * restore original swappiness value
	 */
	if (current_swappiness != SWAP_MEMCG_SWAPPINESS) {
		r = cgroup_write_node_uint32(memcg,
			MEMCG_SWAPNESS_PATH, current_swappiness);
		if (r) {
			_I("failed to write %s %d to %s the",
				MEMCG_SWAPNESS_PATH, SWAP_MEMCG_SWAPPINESS, memcg);
			return -errno;
		}
	}
	return 0;
}

static int swap_compact_in_module(void)
{
	int r, fail = 0;
	GSList *iter;
	struct swap_module_ops *swaps;

	gslist_for_each_item(iter, swap_module) {
		swaps = (struct swap_module_ops *)iter->data;

		if (!swaps->reclaim)
			continue;

		r = swaps->reclaim(swaps);
		if (r)
			fail = r;
	}
	return fail ? fail : 0;
}

static int swap_reclaim_memcg(struct swap_status_msg msg)
{
	int r;
	struct memcg_info *info = msg.info;

	if (!info)
		return -EINVAL;

	r = swap_compact_in_module();
	if (r) {
		if (r == -ENOSPC) {
			/*
			 * If swap space is not enough, swap module will trigger the proactive LMK.
			 * While saving swap space again,
			 * swap will be blocked with minimum swappiness.
			 */
			r = cgroup_read_node_int32(info->name,
				MEMCG_SWAPNESS_PATH, &current_swappiness);
	if (r)
		return r;

			r = cgroup_write_node_uint32(info->name,
				MEMCG_SWAPNESS_PATH, SWAP_MIN_SWAPPINESS);
		}
		return r;
	}

	return swap_start_reclaim(info->name);
}

static int swap_move_inactive_to_swap(struct swap_status_msg *msg)
{
	GSList *proc_app_list = NULL;
	GSList *iter;
	int ret, max_victims;
	struct swap_task victim;
	GArray *candidates = NULL;
	struct memcg *memcg_swap = NULL;
	struct proc_app_info *pai = NULL;

	candidates = g_array_new(false, false, sizeof(struct swap_task));
	if (!candidates) {
		_E("failed to allocate memory");
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	}
	memset(&victim, 0, sizeof(struct swap_task));

	proc_app_list = proc_app_list_open();
	gslist_for_each_item(iter, proc_app_list) {
		pai = (struct proc_app_info *)iter->data;
		if ((!pai->main_pid) ||
		    (pai->memory.memcg_info != msg->info) ||
		    (pai->memory.oom_score_adj < OOMADJ_BACKGRD_UNLOCKED) ||
		    (pai->lru_state <= PROC_BACKGROUND))
			continue;
		victim.pai = pai;
		g_array_append_val(candidates, victim);
	}
	/*
	 * Let's consider 50% of inactive apps to be swappable at once.
	 */
	max_victims = candidates->len >> 1;
	if (max_victims == 0) {
		ret = RESOURCED_ERROR_NO_DATA;
		goto out;
	}
	swap_reduce_victims(candidates, max_victims);
	ret = lowmem_get_memcg(MEMCG_SWAP, &memcg_swap);
	if (ret != RESOURCED_ERROR_NONE)
		goto out;
	/*
	 * change swap info from inactive cgroup to swap group
	 * for using same structure to move and swap it
	 */
	msg->info = memcg_swap->info;
	msg->type = MEMCG_SWAP;
	ret = swap_move_to_cgroup(msg->info, candidates);
out:
	proc_app_list_close();
	g_array_free(candidates, TRUE);

	return ret;
}

static int gen_urandom_string(char *buf, size_t len)
{
	_cleanup_close_ int fd = -1;
	ssize_t l;

	fd = open("/dev/urandom", O_RDONLY);
	if (fd < 0)
		return -errno;

	l = read(fd, buf, len);
	if (l < 0)
		return -errno;

	return 0;
}

static int swap_losetup(struct swap_module_ops *swap,
				char *crypt_type,
				char *swapfile)
{
	char swapdevice[MAX_NAME_LENGTH];
	char passwd[LO_KEY_SIZE] = "";
	int r;

	if (!is_empty(crypt_type)) {
		r = gen_urandom_string(passwd, LO_KEY_SIZE);
		if (r < 0) {
			_E("Failed to generate random passwd: %m");
			return r;
		}
	}

	r = find_loop_device(swapfile, swapdevice);
	if (r < 0) {
		if (r == -EEXIST) {
			_I("'%s' is already setuped to '%s'", swapfile, swapdevice);
			goto finish;
		} else
			_E("Failed to find available loop device.");
		return r;
	}

	r = losetup(swapdevice, swapfile, 0, 0,
		    crypt_type, passwd, false);

	if (r < 0) {
		_E("Failed to setup loop: %s", strerror(-r));
		return r;
	}

finish:
	swap->path = strdup(swapdevice);
	if (!swap->path)
		return -ENOMEM;
	return r;
}

int swap_set_file(char *filename, struct swap_module_ops *swap, char *crypt_type)
{
	int r;

	r = access(filename, F_OK);
	if (r < 0 && errno == ENOENT) {
		r = do_dd("/dev/zero", filename, PAGE_SIZE,
			  KBYTE_TO_BYTE(swap->k_size) >> PAGE_SHIFT);
		if (r > 0) {
			_E("Failed to create swap file(%s), exit code: %d",
					filename, r);
			return -1;
		} else if (r < 0) {
			_E("Failed to create swap file(%s): %s",
					filename, strerror(-r));
			return r;
		}
		_I("SwapFile is created: %s", filename);
	} else if (r != 0) {
		_E("Failed to access swap file(%s): %m", filename);
		return -errno;
	}

	r = swap_losetup(swap, crypt_type, filename);
	if (r < 0)
		return r;

	if (!swap->path)
		return -ENOENT;

	r = do_mkswap(swap->path);
	if (r > 0) {
		_E("Failed to make swap device(%s), exit code: %d", swap->path, r);
		return -EBUSY;
	} else if (r < 0) {
		_E("Failed to make swap device(%s): %s", swap->path, strerror(-r));
	}
	return r;
}

bool swap_is_on(const char *name)
{
	_cleanup_proc_swaps_free_ struct proc_swaps **swaps = NULL;
	int n, i;

	n = proc_get_swaps(&swaps);
	if (n <= 0 || !swaps)
		return false;

	for (i = 0; i < n; i++) {
		if (streq(swaps[i]->filename, name))
			return true;
	}

	return false;
}

static void swap_activate_in_module(void)
{
	int r;
	GSList *iter;
	struct swap_module_ops *swaps;
	static bool early_reclaim;

	r = cgroup_read_node_int32(LOWMEM_ROOT_CGROUP,
		MEMCG_SWAPNESS_PATH, &current_swappiness);
	if (r)
		return;

	if (swap_get_state() == SWAP_ON)
		return;

	gslist_for_each_item(iter, swap_module) {
		int flags = SWAP_FLAG_DISCARD;

		swaps = (struct swap_module_ops *)iter->data;

		if (!swaps->activate)
			continue;

		r = swaps->activate(swaps);
		if (r != 0) {
			if (r != -EEXIST) {
				_E("%s activate failed", swaps->name);
				swap_remove(swaps);
			}
			swap_set_state(SWAP_ON);
			continue;
		}

#ifdef SWAP_FLAG_PREFER
		if (swaps->priority >= 0) {
			flags |= SWAP_FLAG_PREFER |
				(swaps->priority & SWAP_FLAG_PRIO_MASK) << SWAP_FLAG_PRIO_SHIFT;
		}
#endif
		r = swapon(swaps->path, flags);
		if (r < 0) {
			_E("Failed to turn on swap device(%s): %m", swaps->path);
			continue;
		}
		_I("swap device(%s) was turned on: %s", swaps->name, swaps->path);

		/*
		 * Most system daemons use much memory for intializing own process.
		 * but there is little change to reuse it while running.
		 * So early reclaiming moves unused system memory to swap device.
		 * It is enough to trigger reclaiming about root memcg only once.
		 * If multiple swap devices are available,
		 * first swap with the lowest priority will start reclaiming.
		 */
		if (!early_reclaim) {
			swap_start_reclaim(LOWMEM_ROOT_CGROUP);
			early_reclaim = true;
		}

		swap_set_state(SWAP_ON);
	}
}

static void *swap_thread_main(void * data)
{
	int is_empty, ret;
	struct swap_thread_bundle *bundle;

	setpriority(PRIO_PROCESS, 0, SWAP_PRIORITY);

	while (1) {
		pthread_mutex_lock(&swap_mutex);
		/* THREAD: WAIT FOR START */
		pthread_mutex_lock(&swap_thread_queue.lock);
		is_empty = g_queue_is_empty(swap_thread_queue.queue);
		pthread_mutex_unlock(&swap_thread_queue.lock);
		if (is_empty) {
			/* The queue is empty, wait for thread signal */
			pthread_cond_wait(&swap_cond, &swap_mutex);
		}

		/* We're in swap thread, now it's time to dispatch bundles */
		pthread_mutex_lock(&swap_thread_queue.lock);
		bundle = g_queue_pop_head(swap_thread_queue.queue);
		pthread_mutex_unlock(&swap_thread_queue.lock);

		if (!bundle)
			goto unlock_out;

		switch (bundle->op) {
		/* Swap activation operttion: mkswap, swapon etc. */
		case SWAP_OP_ACTIVATE:
			swap_activate_in_module();
			break;
		/* Swap reclaim opertation: move to swap, force_reclaim */
		case SWAP_OP_RECLAIM:
			swap_reclaim_memcg(bundle->msg);
			break;
		/* Swap compact operation of zsmalloc. */
		case SWAP_OP_COMPACT:
			swap_compact_in_module();
			break;
		/* Move inactive procesess to swap, and reclaim after that. */
		case SWAP_OP_MOVE_TO_SWAP_AND_RECLAIM:
			ret = swap_move_inactive_to_swap(&(bundle->msg));
			/* Check if any process was moved to swap. */
			if (ret == RESOURCED_ERROR_NONE)
				swap_reclaim_memcg(bundle->msg);
			break;
		case SWAP_OP_END:
		default:
			_D("wrong swap thread operation selected");
		}

		free(bundle);
unlock_out:
		pthread_mutex_unlock(&swap_mutex);
	}
	return NULL;
}

static int swap_communicate_thread(struct swap_thread_bundle *bundle)
{
	int ret;

	if (!bundle)
		return RESOURCED_ERROR_NO_DATA;

	swap_add_bundle(bundle);
	/* Try to signal swap thread, that there is some work to do */
	ret = pthread_mutex_trylock(&swap_mutex);
	if (ret == 0) {
		pthread_cond_signal(&swap_cond);
		pthread_mutex_unlock(&swap_mutex);
		_I("send signal to swap thread");
		return RESOURCED_ERROR_NONE;
	} else if (ret == EBUSY) {
		_D("swap thread already active");
		return RESOURCED_ERROR_NONE;
	}

	_E("pthread_mutex_trylock fail: %d, errno: %d", ret, errno);
	return RESOURCED_ERROR_FAIL;
}

static int swap_start_handler(void *data)
{
	int ret;
	struct swap_thread_bundle *bundle;

	if (!data)
		return RESOURCED_ERROR_NO_DATA;

	bundle = malloc(sizeof(struct swap_thread_bundle));
	if (!bundle)
		return RESOURCED_ERROR_OUT_OF_MEMORY;

	bundle->op = SWAP_OP_RECLAIM;
	memcpy(&(bundle->msg), data, sizeof(struct swap_status_msg));

	if (bundle->msg.type == MEMCG_APPS) {
		/*
		 * Background tasks are concerned special way, we select
		 * tasks and move them to Swap cgroup. They are not there already.
		 */
		bundle->op = SWAP_OP_MOVE_TO_SWAP_AND_RECLAIM;
	}
	ret = swap_communicate_thread(bundle);
	return ret;
}

static int swap_internal_bundle_sender(enum swap_thread_op operation)
{
	int ret;
	struct swap_thread_bundle *bundle;

	bundle = malloc(sizeof(struct swap_thread_bundle));
	if (!bundle)
		return RESOURCED_ERROR_OUT_OF_MEMORY;
	bundle->op = operation;
	ret = swap_communicate_thread(bundle);
	return ret;
}

static int swap_activate_handler(void *data)
{
	return swap_internal_bundle_sender(SWAP_OP_ACTIVATE);
}

static gboolean swap_activate_timer_cb(gpointer data)
{
	_I("Starting an " EARLYRECLAIM_WITH_AN_EXPLANATION_FOR_LAYMEN);
	swap_activating_timer = NULL;
	swap_internal_bundle_sender(SWAP_OP_ACTIVATE);
	return false;
}

static int swap_booting_done(void *data)
{
	if (!arg_swap_at_boot) {
		_D(EARLYRECLAIM_WITH_AN_EXPLANATION_FOR_LAYMEN " is disabled");
		return RESOURCED_ERROR_NONE;
	}

	/* No need to involve the timer mechanism when the delay is 0,
	 * partially to keep things simple but primarily because it can
	 * introduce an artificial delay since the timer is ran async
	 * in another thread. */
	if (arg_timer_swap_at_boot == 0) {
		swap_activate_timer_cb(NULL);
		return RESOURCED_ERROR_NONE;
	}

	_D("booting done; starting up a timer to perform an " EARLYRECLAIM_WITH_AN_EXPLANATION_FOR_LAYMEN " %ds from now", arg_timer_swap_at_boot);
	swap_activating_timer = g_timeout_source_new_seconds((guint) arg_timer_swap_at_boot);
	g_source_set_callback(swap_activating_timer, swap_activate_timer_cb, NULL, NULL);
	g_source_attach(swap_activating_timer, NULL);

	return RESOURCED_ERROR_NONE;
}

static int swap_compact_handler(void *data)
{
	_I("compaction request. Reason: %s",
			compact_reason_to_str((enum swap_compact_reason)data));
	return swap_internal_bundle_sender(SWAP_OP_COMPACT);
}

/* This function is callback function for the notifier RESOURCED_NOTIFIER_SWAP_UNSET_LIMIT.
 * This notifier is notified from normal_act function of vmpressure module whenever the
 * memory state changes to normal.
 * This function resets the hard limit of the swap subcgroup to -1 (unlimited) */
static int swap_cgroup_reset_limit(void *data)
{
	int ret, limit;
	struct swap_status_msg *msg = data;

	if (swap_node == SWAP_NODE_FORCE_RECLAIM)
		return RESOURCED_ERROR_NONE;

	limit = -1;
	ret = cgroup_write_node_int32(msg->info->name, MEMCG_SIZE_LIMIT, limit);
	if (ret != RESOURCED_ERROR_NONE)
		_E("Failed to change hard limit of %s cgroup to -1", msg->info->name);
	else
		_D("changed hard limit of %s cgroup to -1", msg->info->name);

	return ret;
}

static void swap_start_pid_dbus_signal_handler(GVariant *params)
{
	int ret;
	pid_t pid;
	struct memcg *memcg_swap;
	struct swap_status_msg ss_msg;

	do_expr_unless_g_variant_get_typechecked(return, params, "(i)", &pid);
	if (pid <= 0) {
		_D("Invalid parameter");
		return;
	}

	ret = lowmem_get_memcg(MEMCG_SWAP, &memcg_swap);
	if (ret != RESOURCED_ERROR_NONE)
		return;
	swap_move_to_cgroup_by_pid(MEMCG_SWAP, pid);
	ss_msg.pid = pid;
	ss_msg.type = MEMCG_SWAP;
	ss_msg.info = memcg_swap->info;
	swap_start_handler(&ss_msg);
	_I("swap cgroup entered : pid : %d", (int)pid);
}


static const struct d_bus_signal dbus_signals[] = {
	/* RESOURCED DBUS */
	{RESOURCED_PATH_SWAP, RESOURCED_INTERFACE_SWAP,
	    SIGNAL_NAME_SWAP_START_PID, swap_start_pid_dbus_signal_handler, NULL},
};

static void swap_dbus_init(void)
{
	d_bus_register_signals(dbus_signals, ARRAY_SIZE(dbus_signals));
}

static int config_parse_swap_types(const char *filename,
				  unsigned line,
				  const char *section,
				  const char *lvalue,
				  int ltype,
				  const char *rvalue,
				  void *data)
{
	enum swap_type *type = data;
	char *word, *state;
	size_t l;

	if (is_empty(rvalue))
		return 0;

	*type = 0;

	FOREACH_WORD_SEPARATOR(word, l, rvalue, "+|", state) {
		if (strneq(word, "zram", l))
			*type |= SWAP_TYPE_ZRAM;
		else if (strneq(word, "file", l))
			*type |= SWAP_TYPE_FILE;
		else if (strneq(word, "zswap", l))
			*type |= SWAP_TYPE_ZSWAP;
		else
			return -EINVAL;
	}

	return 0;
}

static int swap_parse_config_file(void)
{
	const ConfigTableItem items[] = {
		{ "SWAP",	"Enable",		config_parse_bool,
			0,	&arg_swap_enable	},
		{ "SWAP",	"Type",			config_parse_swap_types,
			0,	&arg_swap_type	},
		{ "SWAP",	"ReclaimAtBoot",	config_parse_bool,
			0,	&arg_swap_at_boot	},
		{ "SWAP",	"TimerReclaimAtBoot",	config_parse_int,
			0,	&arg_timer_swap_at_boot	},
		{ NULL,		NULL,			NULL,
			0,	NULL	}
	};

	int r;

	r = config_parse_new(SWAP_CONF_FILE, (void*) items);
	if (r < 0) {
		_E("Failed to parse configuration file: %s", strerror(-r));
		return r;
	}

	/* `g_timeout_source_new_seconds` wants an unsigned value,
	 * but there is no `config_parse_uint` to produce one. */
	if (arg_timer_swap_at_boot < 0) {
		_E("The `TimerReclaimAtBoot` field in the `" SWAP_CONF_FILE "` config file cannot be negative because it represents a time period for " EARLYRECLAIM_WITH_AN_EXPLANATION_FOR_LAYMEN);
		return -EINVAL;
	}

	return 0;
}

static int swap_thread_create(void)
{
	int ret = 0;
	pthread_t pth;

	pthread_mutex_init(&swap_mutex, NULL);
	pthread_cond_init(&swap_cond, NULL);
	pthread_mutex_init(&(swap_thread_queue.lock), NULL);
	swap_thread_queue.queue = g_queue_new();

	if (!swap_thread_queue.queue) {
		_E("fail to allocate swap thread queue");
		return RESOURCED_ERROR_FAIL;
	}

	ret = pthread_create(&pth, NULL, &swap_thread_main, (void *)NULL);
	if (ret) {
		_E("pthread creation for swap_thread failed\n");
		return ret;
	} else {
		pthread_detach(pth);
	}

	return RESOURCED_ERROR_NONE;
}

static int swap_init(void)
{
	int r;
	GSList *iter, *next;
	struct swap_module_ops *swaps = NULL;
	int swap_pri = SWAP_PRI_DEFAULT;

	r = swap_parse_config_file();
	if (r < 0)
		return r;

	if (!arg_swap_enable)
		return -ENODEV;

	gslist_for_each_safe(swap_module, iter, next, swaps) {
		swaps = (struct swap_module_ops *)iter->data;
		if (!CHECK_BIT(arg_swap_type, swaps->type)) {
			swap_remove(swaps);
			continue;
		}

		if (!swaps->init)
			continue;

		/*
		 * swap_pri has a priority of previous swaps.
		 * If swaps has SWAP_PRI_DISABLE,
		 * swap will be supported with only one module.
		 * After that, other modules should be removed.
		 */
		if (swap_pri == SWAP_PRI_DISABLE) {
			swap_remove(swaps);
			_E("don't allow other %s module", swaps->name);
			continue;
		}
		swap_pri = swaps->priority;

		r = swaps->init(swaps);
		if (r != 0) {
			_E("%s module init failed", swaps->name);
			swap_remove(swaps);
		}
	}

	r = swap_thread_create();
	if (r) {
		_E("swap thread create failed");
		return r;
	}
	swap_dbus_init();

	return r;
}

static int swap_check_node(char *path)
{
	_cleanup_fclose_ FILE *fp = NULL;

	fp = fopen(path, "w");
	if (fp == NULL) {
		_E("%s open failed", path);
		return RESOURCED_ERROR_NO_DATA;
	}
	return RESOURCED_ERROR_NONE;
}

static int resourced_swap_check_runtime_support(void *data)
{
	int r;
	uint32_t usage;

	/*
	 * Check whether CONFIG_SWAP is enabled in kernel.
	 */
	if (access("/proc/swaps", R_OK) != 0)
		return -ENOENT;

	/*
	 * Check whether kernel is supporting MEMCG_SWAP.
	 */
	r = cgroup_read_node_uint32(LOWMEM_ROOT_CGROUP,
				MEMCG_SWAP_USAGE, &usage);
	if (r)
		return -ENOENT;

	return r;
}

/*
 * Quote from: kernel Documentation/cgroups/memory.txt
 *
 * Each bit in move_charge_at_immigrate has its own meaning about what type of
 * charges should be moved. But in any case, it must be noted that an account of
 * a page or a swap can be moved only when it is charged to the task's current
 * (old) memory cgroup.
 *
 *  bit | what type of charges would be moved ?
 * -----+------------------------------------------------------------------------
 *   0  | A charge of an anonymous page (or swap of it) used by the target task.
 *      | You must enable Swap Extension (see 2.4) to enable move of swap charges.
 * -----+------------------------------------------------------------------------
 *   1  | A charge of file pages (normal file, tmpfs file (e.g. ipc shared memory)
 *      | and swaps of tmpfs file) mmapped by the target task. Unlike the case of
 *      | anonymous pages, file pages (and swaps) in the range mmapped by the task
 *      | will be moved even if the task hasn't done page fault, i.e. they might
 *      | not be the task's "RSS", but other task's "RSS" that maps the same file.
 *      | And mapcount of the page is ignored (the page can be moved even if
 *      | page_mapcount(page) > 1). You must enable Swap Extension (see 2.4) to
 *      | enable move of swap charges.
 * quote end.
 *
 * In our case it's better to set only the bit number 0 to charge only
 * anon pages. Therefore file pages etc. will be managed directly by
 * kernel reclaim mechanisms.
 * That will help focus us only on swapping the memory that we actually
 * can swap - anonymous pages.
 * This will prevent from flushing file pages from memory - causing
 * slowdown when re-launching applications.
 */
static void resourced_swap_change_memcg_settings(enum memcg_type type)
{
	int ret;
	struct memcg *memcg_swap = NULL;
	char buf[MAX_PATH_LENGTH];

	ret = lowmem_get_memcg(type, &memcg_swap);
	if (ret != RESOURCED_ERROR_NONE)
		return;

	cgroup_write_node_uint32(memcg_swap->info->name, MOVE_CHARGE, 1);
	snprintf(buf, sizeof(buf), "%s/%s", MEMCG_PATH, SWAPCG_RECLAIM);
	ret = swap_check_node(buf);
	if (ret == RESOURCED_ERROR_NONE) {
		swap_node = SWAP_NODE_FORCE_RECLAIM;
		_I("use %s node for swapping memory", SWAPCG_RECLAIM);
	} else {
		swap_node = SWAP_NODE_HARD_LIMIT;
		_I("use %s node for swapping memory", MEMCG_SIZE_LIMIT);
	}
}

static int resourced_swap_init(void *data)
{
	int ret;

	resourced_swap_change_memcg_settings(MEMCG_SWAP);
	swap_set_state(SWAP_OFF);

	ret = swap_init();
	if (ret != RESOURCED_ERROR_NONE)
		return ret;

	if (!resourced_restarted()) {
		register_notifier(RESOURCED_NOTIFIER_BOOTING_DONE, swap_booting_done);
	} else {
		_I("Assuming swap is already set up. To reconfigure, reboot the device.");
	}

	register_notifier(RESOURCED_NOTIFIER_SWAP_START, swap_start_handler);
	register_notifier(RESOURCED_NOTIFIER_SWAP_ACTIVATE, swap_activate_handler);
	register_notifier(RESOURCED_NOTIFIER_SWAP_COMPACT, swap_compact_handler);
	register_notifier(RESOURCED_NOTIFIER_SWAP_UNSET_LIMIT, swap_cgroup_reset_limit);

	return ret;
}

static int resourced_swap_finalize(void *data)
{
	unregister_notifier(RESOURCED_NOTIFIER_SWAP_START, swap_start_handler);
	unregister_notifier(RESOURCED_NOTIFIER_SWAP_ACTIVATE, swap_activate_handler);
	unregister_notifier(RESOURCED_NOTIFIER_BOOTING_DONE, swap_booting_done);
	unregister_notifier(RESOURCED_NOTIFIER_SWAP_COMPACT, swap_compact_handler);
	unregister_notifier(RESOURCED_NOTIFIER_SWAP_UNSET_LIMIT, swap_cgroup_reset_limit);

	g_queue_free(swap_thread_queue.queue);

	return RESOURCED_ERROR_NONE;
}

static struct module_ops swap_modules_ops = {
	.priority = MODULE_PRIORITY_NORMAL,
	.name = "swap",
	.init = resourced_swap_init,
	.exit = resourced_swap_finalize,
	.check_runtime_support = resourced_swap_check_runtime_support,
};

MODULE_REGISTER(&swap_modules_ops)
