/*
 * watchdog-test
 *
 * Copyright (c) 2019 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>
#include <gio/gio.h>

#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <system/system_info.h>

#define EXIT_OK		0	/* test passed */
#define EXIT_ERROR	1	/* test failed */
#define EXIT_FAIL	2	/* testing framework failed */

#define DESTINATION	"org.tizen.resourced"
#define PATH		"/Org/Tizen/ResourceD/Watchdog"
#define INTERFACE	"org.tizen.resourced.watchdog"
#define METHOD		"Start"

#define CORE_PATTERN_PATH "/proc/sys/kernel/core_pattern"
char core_pattern[4096];

int send_watchdog_start(GDBusConnection *conn, pid_t pid, unsigned int t)
{
	GVariant *params = NULL;
	GError *err = NULL;
	int ret = EXIT_FAIL;

	params = g_variant_new("(iu)", pid, t);
	g_dbus_connection_call_sync(conn,
								DESTINATION,
								PATH,
								INTERFACE,
								METHOD,
								params,
								NULL,
								G_DBUS_CALL_FLAGS_NONE,
								-1,
								NULL,
								&err);
	if (err == NULL) {
		ret = EXIT_OK;
	} else {
		fprintf(stderr, "%s:%d:%s", __func__, __LINE__, err->message);
	}

	g_clear_error(&err);
	return ret;
}

int watched()
{
	int ret = EXIT_FAIL;
	GDBusConnection *conn;
	int i;

	conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, NULL);
	if (conn == NULL)
		return ret;

	for (i = 1; i <= 3; i++) {
		ret = send_watchdog_start(conn, getpid(), 5);
		if (ret != EXIT_OK) {
			goto out;
		}
		sleep(2*i);
	}

	ret = EXIT_ERROR;
out:
	g_object_unref(conn);
	return ret;
}

int milliseconds(struct timeval *t)
{
	return t->tv_sec * 1000 + t->tv_usec / 1000;
}

void save_core_pattern()
{
	int fd;
	int ret;
	memset(core_pattern, 0, sizeof(core_pattern));
	fd = open(CORE_PATTERN_PATH, O_RDONLY);
	if (fd < 0)
		return;
	ret = read(fd, core_pattern, sizeof(core_pattern));
	if (ret < 0) {
		// do nothing
	}
	close(fd);
}

void set_core_pattern(const char *t)
{
	int fd;
	int ret;
	fd = open(CORE_PATTERN_PATH, O_WRONLY);
	if (fd < 0)
		return;
	ret = write(fd, t, strlen(t));
	if (ret < 0) {
		// do nothing
	}
	close(fd);
}

int main(int argc, char* argv[])
{
	int ret = EXIT_ERROR;
	int wstatus;
	pid_t child;

	child = fork();
	if (child == 0) {
		ret = watched();
		return ret;
	} else if (child == -1) {
		ret = EXIT_FAIL;
		goto out;
	}

	save_core_pattern();
	set_core_pattern("/dev/null");

	if (waitpid(child, &wstatus, 0) != child) {
		ret = EXIT_FAIL;
		goto out;
	}

	if (WTERMSIG(wstatus) == SIGABRT)
		ret = EXIT_OK;
out:
	if (core_pattern[0] != '\0')
		set_core_pattern(core_pattern);
	return ret;
}
